MARKER TERMINAL

\ allocate buufer to read position from terminal
VARIABLE at_buff #12 ALLOT 

\ default number of rows and columns
#80 VALUE at_rows
#25 VALUE at_cols

\ send a 7 bit ANSI CSI sequence ( Esc[ ) 
: at-csi ( -- ) 
  #27 EMIT $5B EMIT ;

\ write a decimal byte value with no space chars  
: at-byte ( c -- )
  $FF AND 1 U.R ;

\ send a semi-colon ";" char to the terminal
: at-semi ( -- ) 
  $3B EMIT ;

\ set the terminal x Y cursor position
: at-xy! ( X Y -- )
  \ Convert 0 based X Y to 1 based X Y for Ansi terminal, 
  SWAP 1+ SWAP 1+
  \ send CSI row ; coloum H 
  at-csi 
  at-byte at-semi at-byte
  \ send "H" character 
  $48 EMIT ;

\ return number of columns and rows
: at-max-xy@ ( -- X Y) 
  at_cols at_rows ;
  
\ clear the terminal screen
: at-page ( -- )  
  at-csi ." 2J" ;

\ clear to end of current row
: at-eol ( -- ) 
  at-csi ." 0K" ;

\ show or hide terminal cursor
: at-cursor ( fShowCursor -- )
  at-csi 
  ." ?25" 
  IF 
    ." h" 
  ELSE 
    ." l" 
  THEN ;

\ read terminal cursor position responce string
: at-rxst ( -- c-adr uLength )
  \ wait for [ character
  BEGIN
    \ see if we have a left bracket char
    KEY $5B =
  UNTIL
  \ read responce string params
  at_buff 
  ( c-adr )
  BEGIN 
    ( c-adr )
    \ see if we read entire responce, end with char "R"
    KEY [CHAR] R OVER <> IF
      \ not end of responce string, add char to buffer
      ( c-adr cKey )
      \ convert semi-colons to space char
      DUP $3B = IF
        DROP BL
      THEN
      \ save char to string buffer  
      ( c-adr cKey )
      OVER C!
      \ point to next buffer address
      1+ 
      ( c-adr' ) 
      \ indicate we should loop for more data
      FALSE
    ELSE
      ( c-adr cKey )
      \ end of responce string found, calculate length
      DROP 
      ( c-adr )
      at_buff -
      ( uLength )
      at_buff SWAP 
      ( c-adr uLength )
      \ indicate we should exit read loop
      TRUE
    THEN
  UNTIL ; 
  
\ Read the terminals current X Y cursor position  
: at-xy@ ( -- X Y ) 
  at-csi ." 6n" 
  EMPTY-KEYS
  at-rxst
  \ convert the return data string to X Y values
  EVALUATE
  \ convert to 0 based values 
  1- SWAP 1- ;

\ calc max X Y terminal position
: at-intmax ( -- ) 
  \ read current terminal cursor position
  at-xy@
  ( uCurX uCurY )
  \ set terminal cursor past max position 
  #160 #160 at-xy!
  \ read actual terminal position (should be at Max X Y) 
  at-xy@ 
  \ save values for later use
  TO at_rows 
  TO at_cols
  \ restore prior terminal position 
  at-xy! ;

\ Use at>term to assign terminal functions to ansi driver words.
: at>term ( -- ) 
  ['] at-xy! IS AT-XY 
  ['] at-xy@ IS XY
  ['] at-page IS PAGE 
  ['] at-eol IS CLR-EOL
  ['] at-max-xy@ IS MAX-XY 
  ['] at-cursor IS CURSOR
  at-intmax ;
  
\ Use tt>term when ansi terminal is windows telnet client.
: tt>term ( -- )
  \ instructs Windows telnet client not to echo chars
  $FF EMIT $84 EMIT
  at>term ;
  
at>term  
