\ This code is a forth based screen editor that allows you to
\ edit a forth screen saved in a block device block.
\ This editor requires terminal support (PAGE, AT-XY...).
\ ForthTerm and AnsiTerm are supplied either will do
\ or you can create your own terminal driver.

\ The screen editor requires a display at least 66 columns x 23 rows.

\ To run the editor type: "xx edit" 
\ where xx is a valid block device block number.

\ I did not create this screen editor. However 
\ I did alot of changes to the original version.

\ There is an external link to the orginal source
\ code for the editor on my site www.shaels.net

MARKER unload-edit

( 1 Editor Constants )
CHAR i CONSTANT 'i   \ Insert mode
CHAR o CONSTANT 'o   \ Overwrite mode
CHAR c CONSTANT 'c   \ Command mode
CHAR 0 CONSTANT '0   
CHAR y CONSTANT 'y
CHAR n CONSTANT 'n
CHAR A CONSTANT 'A
CHAR Z CONSTANT 'Z
CHAR $ CONSTANT '$

( Editor State )
 2 CREATE scr ,   \ Current block
 0 CREATE x ,     \ Cursor X position 0..63
 0 CREATE y ,     \ Cursor Y position 0..15
'c CREATE mode ,  \ Change to bitmap later.
CREATE wordname #5 C, '$ C, '$ C, 0 C, 0 C, 0 C,

( 2 Editor Undo buffer )
#16 CONSTANT ub-lines
#15 CONSTANT ub-line-msk
VARIABLE ub-head
VARIABLE ub-count
VARIABLE ub-buffer ub-lines #64 * ALLOT

FALSE VALUE blk_updated

\ will need original UPDATE word
: BLK-UPDATE UPDATE ;

: UPDATE ( -- )
  \ indicate block must be flushed
  TRUE TO blk_updated ;

( Editor State Control )
: insert 'i mode ! ;
: overwrite 'o mode ! ;
: cmd 'c mode ! ;
: inserting? mode @ 'i = ;

\ toggle between insert and replace mode
: toggleinsmode 
  mode @ DUP
  ( cMode cMode ) 
  'c <> IF
    ( cMode )
    'o = IF 
      insert 
    ELSE 
      overwrite 
    THEN 
  THEN ;  

: cmd? ( -- fIsCommandMode ) 
  mode @ 
  'c = ;

: edit? ( -- fInEditMode ) 
  mode @ 'i = 
  mode @ 'o = 
  OR ;

: mode!
  edit? 'i AND
  ( FALSE | cI ) 
  cmd? 'c AND 
  ( FALSE | cI  FALSE | cC ) 
  OR 
  ( cI | cC | FALSE )
  wordname  
  #3 + C! ;

( 3 Editor Display 1 )
\ mult by chars per line
: 64* #6 LSHIFT ;

: where ( x y -- adr )
  \ calc block offset
  64* +
  \ calc block buffer source address
  scr @ BLOCK + ;
  
\  scr @ BLOCK
\  SWAP 64* + 
\  SWAP + ;
  
: wh 
  x @ y @ where ;

\ display current editor mode
: mode. ( -- )
  #63 0 AT-XY
  edit? CURSOR 
  mode @ EMIT ;

: scr. 0 0 AT-XY ." Block: " scr @ . ."      " ;
: header scr. mode. ;
: 64-s #64 [CHAR] - (chars) ;

DEFER border

\ draw a border with a space followed by 64 dashes for a wide screen
: borwide 
  SPACE 64-s CR ;

\ draw a border with 64 dashes for a narrow screen
: borthin ( -- )
  64-s ;

\ display row starting at address c-adr, return start of next row
: row ( c-adr -- c-adr' )
  DUP #64 TYPE #64 + ;

\ display a row of text on a wide screen ( > 66 columns )
: linewide ( c-adr -- c-adr' )
  ." |"
  row 
  ." |" 
  CR ;

\ draw a row of text on the screen
DEFER line ( c-adr -- c-adr' )

\ returns a flag to indicate if we have a wide screen  
: wide? ( -- fWide )
  MAX-XY 
  DROP #66 > ;

DEFER cursor

\ Position the cursor to our current position
: curwide ( -- )
  x @ 1+ 
  y @ 2+ 
  AT-XY ;

\ Position the cursor to our current position
: curthin ( -- )
  x @ 
  y @ 2+ 
  AT-XY ;

: adjwide
  wide? IF
    ['] borwide IS border
    ['] linewide IS line
    ['] curwide IS cursor
  ELSE
    ['] borthin IS border
    ['] row IS line
    ['] curthin IS cursor
  THEN ;
  
( 4 draw help stuff )
: clrhelp 0 #19 AT-XY #64 SPACES 0 #20 AT-XY #64 SPACES ;
: hlp1a ." Char: ^G-del ^S-left ^D-right ^X-down ^E-up   " ;
: hlp1b ." ^A-home ^F-end " ;
: hlp2a ." Line: ^Y-cut ^C-copy ^U-paste ^T-join ^V-mode " ;
: hlp2b ." ESC-command" ;
: .edithelp clrhelp 1 #19 AT-XY hlp1a hlp1b
  1 #20 AT-XY hlp2a hlp2b ;
: cmdhelp1 ."  Mode: (i)nsert (o)verwrite (q)uit" ;
: cmdhelp2 ." Block: (e)rase (n)ext (p)rior (s)ave (g)oto" ;
: .cmdhelp clrhelp 1 #19 AT-XY cmdhelp1 1 #20 AT-XY cmdhelp2 ;

: .help ( -- )
    cmd? IF 
      .cmdhelp 
    ELSE 
      .edithelp
    THEN ;

( Editor Display 2 )
\ : 4lines line line line line ;
\ : 16lines scr @ BLOCK 4lines 4lines 4lines 4lines DROP ;
\ : card 0 1 AT-XY border 16lines border ;
: card ( -- )
  0 1 AT-XY 
  border 
  scr @ BLOCK
  #16 0 DO
    line
  LOOP   
  DROP 
  border ;

\ draw entire edit screen
: screen ( -- ) 
  FALSE CURSOR 
  header 
  card 
  .help 
  cursor
  cmd? NOT CURSOR ;

: draw-line FALSE CURSOR 0 y @ #2 + AT-XY
  scr @ BLOCK y @ 64* + line DROP
  TRUE CURSOR cursor ;
: ref-line inserting? IF draw-line ELSE
  wh C@ EMIT THEN cursor ;
\ : bounds scr @ 0 MAX blk-count 1- MIN scr ! ;
: home 0 x ! 0 y ! cursor ;
\ : prevblock -1 scr +! bounds home ;
\ : nextblock  1 scr +! bounds home ;


( block selection functions )

( Editor Cursor Control )
: flushleft 0 x ! ;
: boundX x @ 0 MAX #63 MIN x ! ;
: boundY y @ 0 MAX #15 MIN y ! ;
: bounds boundX boundY ;
: left -1 x +! bounds ;
: right 1 x +! bounds ;
: up -1 y +! bounds ;
: down 1 y +! bounds ;

\ beep speaker for invalid key
: invalidkey ( cKey -- ) 
  DROP 
  #7 EMIT ;

: nextline y @ #15 < IF flushleft down THEN ;
: next x @ #63 = IF nextline EXIT THEN right ;
: y>adr ( y -- y-adr) 0 SWAP where ;
: eol #63 y @ where ;
: -eol? x @ #63 < ;

( Editor Insert/Replace Text )
: place wh C! UPDATE ;
: openr wh DUP 1+ #63 x @ - MOVE ;
: openRight -eol? IF openr THEN ;

: printable 
  #31 #128 WITHIN ;
  
: canins? eol C@ BL = ;
: inschr canins? IF openRight place ref-line next THEN ;

\ write the char to the edit buffer
: chr ( cKey -- )
  DUP printable IF 
    inserting? IF 
      inschr 
    ELSE
      DUP place EMIT next 
    THEN 
  ELSE 
    DROP 
  THEN ;
  
: eotx ( -- x' ) 0 0 eol DUP 1+ SWAP #63 -
  DO 
    I C@ BL <> IF 
      NIP DUP 
    THEN 
    1+ 
  LOOP
  DROP DUP y @ where C@ BL <> OVER #63 <
  AND IF 
    1+ 
  THEN ;
  
: goeot eotx x ! boundX ;


( Editor Get number )
CREATE input_string 10 ALLOT
: #in-digit? ( c - f ) $30 $40 WITHIN ;
: #in-push ( MaxLen Len c-adr c -- MaxLen n1 c-adr1 ) 2>R 2DUP >
  2R> ROT IF DUP EMIT OVER C! 1+ SWAP 1+ SWAP ELSE DROP THEN ;
: #in-pop ( MaxLen Len c-adr -- MaxLen n1 c-adr1 )
  OVER 1 > IF BACKSPACE SWAP 1- SWAP 1- THEN ;
: #in-quit ( c -- f ) DUP $1B = SWAP $0D = OR ;
: #in ( max-len - d f ) 2+ DUP #10 > IF DROP #9 THEN 1-
  1 input_string 1+ [CHAR] # OVER C! 1+
  BEGIN KEY DUP #in-quit NOT WHILE
  DUP BS = IF DROP #in-pop ELSE DUP #in-digit? IF #in-push ELSE
  DROP THEN THEN REPEAT $1B = IF 2DROP DROP 0. 0 ELSE
  [CHAR] . SWAP C! 1+ input_string C! NIP
  input_string ?DNUMBER 0> THEN ;
: get-block# #5 #in IF D>S TRUE ELSE DROP FALSE THEN ;

( Editor Undo buffer )
: ub-init ( -- ) ub-buffer ub-lines 64* BL FILL
  0 ub-head ! 0 ub-count ! ;
: ub-empty? ( -- f ) ub-count @ 0= ;
: ub-lineadr ( u -- c-adr ) 64* ub-buffer + ;
: ub-next ( u -- u-next ) 1+ ub-line-msk AND ;
: ub-prior ( u -- u-prior ) 1- ub-line-msk AND ;
: ub-count+ ( -- ) ub-count @ 1+ ub-lines MIN ub-count ! ;
: ub-count- ( -- ) ub-count @ 1- 0 MAX ub-count ! ;
: ub-save ( y -- ) y>adr ub-head @ ub-lineadr
  64 MOVE ub-head @ ub-next ub-head ! ub-count+ ;
: ub-restore ( y -- ) ub-empty? IF DROP ELSE
  y>adr ub-head @ ub-prior DUP ub-head ! ub-lineadr
  SWAP #64 MOVE ub-count- THEN ;


( Editor Delete )
: padding #32 eol C! UPDATE ;
: del wh DUP 1+ SWAP #63 x @ - MOVE ;
: delete -eol? IF del THEN padding ;
: copylinesup ( y -- ) DUP >R #15 < IF
  0 R@ 1+ where 0 R@ where OVER
  #63 #15 where 1+ SWAP - CMOVE
  THEN 0 #15 where #64 BL FILL UPDATE R> DROP ;
: deleteline y @ ub-save y @ copylinesup ;

( Editor Join lines/Backspace )
: blanky? ( y -- ) TRUE SWAP y>adr DUP #64 + SWAP
  DO I C@ BL <> IF DROP FALSE LEAVE THEN LOOP ;
: joinlines? y @ #15 < eotx x @ <= AND ;
: joinlines ( -- ) joinlines? IF y @ 1+
  y>adr wh #64 MOVE wh #64 + #64 x @ - BL
  FILL y @ 1+ blanky? IF y @ 1+ copylinesup THEN
  UPDATE card THEN ;
: insline? 15 blanky? ;
: bs left delete ;
: joinback y @ 0> IF y @ 1- y ! goeot joinlines THEN ;
: backspace x @ 0> IF bs ELSE joinback THEN ;

( Editor Carriage Return )
: nextln eol 1+ ;
: #chrs scr @ BLOCK #1024 + nextln - #64 - ;
: copydown y @ #14 < IF nextln DUP #64 + #chrs MOVE THEN ;
: blankdown nextln #64 #32 FILL UPDATE ;
: splitdown wh nextln 2DUP SWAP - MOVE ;
: blankrest wh nextln OVER - #32 FILL ;
: opendown copydown blankdown ;
: splitline insline? IF opendown splitdown blankrest THEN ;
: retrn inserting? IF splitline THEN flushleft nextline ;
: return y @ #15 < IF retrn inserting?
  IF card ELSE cursor THEN THEN ;
: movedown y @ #15 <
  IF y @ y>adr DUP #64 + #chrs #64 + MOVE THEN ;
: restoreline insline? IF ub-empty? NOT IF movedown y @
  ub-restore UPDATE card THEN THEN ;

( promt user for yes or no )
: usr-keys S" nyNY" ;

: usr-ask? ( -- cChar )
  BEGIN
    KEY? IF
      KEY
      \ char to match 
      usr-keys 
      ( cChar adrChars uCharCnt )
      0 ?DO
        ( cChar adrChars )
        2DUP
        ( cChar adrChars cChar adrChars )
        C@
        ( cChar adrChars cChar cNextChar )
        = IF
          \ we have a match
          ( cChar adrChars )
          DROP
          UNLOOP
          #32 OR
          [CHAR] y =
          EXIT
        ELSE
          ( cChar adrChars )
          1+
          ( cChar adrChars' )
        THEN
      LOOP
      ( cChar adrChars )
      2DROP
    THEN
  AGAIN ;  

\ : valid? DUP 'n = SWAP 'y = OR ;
\ : uppercase? 'A 'Z 1+ WITHIN ;
\ : lowercase DUP uppercase? IF #32 XOR THEN ;
\ : validkey BEGIN KEY lowercase DUP valid?
\   IF TRUE ELSE DROP FALSE THEN UNTIL ;

: usr-prep ( adrChars uCharCnt -- )
  clrhelp 
  0 #20 AT-XY TYPE ; 

: usr-clr ( -- )
  .help
  cmd? NOT CURSOR ;

: usr-prmt? ( adrChars uCharCnt -- fWasYes )
  usr-prep 
  ." (Y/N)?" 
  TRUE CURSOR
  usr-ask?
  usr-clr ;

: usr-wait ( adrChars uCharCnt -- )
  usr-prep
  ." (Press Any Key)" 
  TRUE CURSOR
  KEY DROP
  usr-clr ;

( Editor Wipe Block )
: wipe
  S"   Erase block " usr-prmt? IF
    scr @ BLOCK 
    #1024 #32 FILL
    UPDATE 
    0 x ! 
    0 y !
  THEN ;

( Editor Wipe Block )
\ : msg clrhelp 0 #20 AT-XY ."   Erase block? (Y/N) " ;
\ : clrmsg .help ;
\ : no? msg validkey clrmsg 'n = ;
\ : ?confirm no? IF R> DROP THEN ;
\ : wipe ?confirm scr @ BLOCK #1024 #32 FILL
\   UPDATE 0 x ! 0 y ! ;


: blk-save ( -- )
  blk_updated IF
    BLK-UPDATE FLUSH
    FALSE TO blk_updated
  THEN ;      


: blk-changing ( -- )
  blk_updated IF
    \ prompt to see if user wants to save block
    S"   Save changes " usr-prmt? IF
      blk-save
    ELSE
      EMPTY-BUFFERS
    THEN
    FALSE TO blk_updated
  THEN ;      

: blk-prompt S"   Goto block number: " usr-prep ;

: blk-error S"  Invalid block " usr-wait  ;

: block# TRUE CURSOR blk-prompt get-block# ;

: blk-valid ( blk -- f) 
  blk-count U< ;

: goto-block 
  block# IF 
    DUP blk-valid IF 
      blk-changing scr ! bounds 
    ELSE
      blk-error DROP 
    THEN 
  ELSE 
    DROP 
  THEN 
  .help home ;
  
: prevblock 
  scr @ DUP 0<> IF
    blk-changing
    1- scr !
    home
  ELSE
    DROP  
  THEN ;

: nextblock 
  scr @ DUP blk-count 1- U< IF
    blk-changing
    1+ scr !
    home
  ELSE
    DROP  
  THEN ;

: blk-save ( -- )
  blk_updated IF
    S"   Save changes " usr-prmt? IF
      blk-save
\      S"  Block saved " usr-wait
    THEN
  THEN ;

( Command Mode: Quit, block, et. al. )
: $$c71 DROP blk-changing 0 #23 AT-XY R> R> DROP >R TRUE CURSOR ; \ q - quit
: $$c69 DROP insert mode. .help ;        \ i
: $$c6F DROP overwrite mode. .help ;     \ o
: $$c73 DROP blk-save ;                  \ s - save
: $$i1B DROP cmd mode. .help ;           \ (escape)
: $$c70 DROP prevblock screen ;          \ p
: $$c6E DROP nextblock screen ;          \ n
: $$c65 DROP wipe card ;                 \ e
: $$c67 DROP goto-block screen ;         \ g

( Editor: backspace, cursor, delete, et. al. )
: $$i07 DROP delete draw-line ;            \ CTRL-G
: $$i18 DROP down cursor ;                 \ CTRL-X
: $$i05 DROP up cursor ;                   \ CTRL-E
: $$i13 DROP left cursor ;                 \ CRTL-S
: $$i04 DROP right cursor ;                \ CRTL-D
: $$i01 DROP flushleft cursor ;            \ CRTL-A
: $$i06 DROP goeot cursor ;                \ CRTL-F
: $$i19 DROP deleteline card  ;            \ CRTL-Y
: $$i15 DROP restoreline ;                 \ CRTL-U
: $$i14 DROP joinlines ;                   \ CRTL-T
: $$i08 DROP backspace draw-line ;         \ (bs)
: $$i7F DROP backspace draw-line ;         \ DEL
: $$i0D DROP return ;                      \ (cr)
: $$i16 DROP toggleinsmode mode. ;         \ CRTL-V
: $$i03 DROP y @ ub-save ;                 \ CRTL-C


( Editor Keyboard Handler )

: >hex DUP #9 > IF #7 + THEN '0 + ;

: h! DUP #240 AND #4 RSHIFT >hex wordname 4 + C! ;

: l! #15 AND >hex wordname #5 + C! ;

: name! mode! h! l! ;

: nomapping ( cKey -- )
  DROP 
  ['] invalidkey cmd? AND
  ['] chr edit? AND 
  OR ;

: handlerword name! wordname FIND IF ELSE nomapping THEN ;

: handler ( cKey -- ) 
  DUP handlerword EXECUTE ;

\ main edit loop
: editloop ( -- ) 
  BEGIN 
    KEY 
    handler 
    cursor 
  AGAIN ;

\ edit startup word
: edit ( uBlkNum -- )
  DEPTH 0= IF
    scr @
  THEN   
  \ set correct border and line style
  adjwide 
  ub-init scr ! 
  PAGE screen 
  FALSE CURSOR 
  home 
  editloop
  TRUE CURSOR 
  PAGE ;

