''Forth words to control I/O pins implemented in propeller assembly
''For MV4th16 project
''Version 1.0
''Nov 1, 2012
''www.shaels.net
''
CONST
  c_ForthCmdCount = 17
  
PUB Start

DAT

Entry                   jmp     #F_Do_Cmds

'*********************************[ FORTH system command support routines]*********************************
'
' Variables used by forth SYSTEMEX interface
ForthTmp                long    0
ForthCmd                long    0
ForthStackPtr           long    0
ForthScratch            long    0   
' Constants
ForthWordMask           long    $FFFF


' Remove top 16 bit value from forth stack.T1 = 16 bit value on return
                        'make sure we are in a command and stack pointer is valid
F_PopStack              tjz     ForthStackPtr, #F_PopStack_Ret
                        'We are in a command OK to pop value from stack
                        rdword  ForthTmp, ForthStackPtr
                        add     ForthStackPtr, #2
F_PopStack_Ret          ret             
'
' Add 16 bit value to top of forth stack.T1 = 16 bit value to write.
                        'make sure we are in a command and stack pointer is valid
F_PushStack             tjz     ForthStackPtr, #F_PushStack_Ret
                        'We are in a command OK to push a value to the stack
                        sub     ForthStackPtr, #2
                        wrword  ForthTmp, ForthStackPtr
F_PushStack_Ret         ret             
'
'
'Pop a 32 bit value from the forth stack in Low word high word order.
                       'Read High 16 bits from the stack
F_PopStack32            call    #F_PopStack
                        mov     ForthScratch, ForthTmp
                        shl     ForthScratch, #16
                        'write high 16 bits to stack
                        call    #F_PopStack
                        or      ForthTmp, ForthScratch
F_PopStack32_ret        ret
'
'Push a 32 bit value to the stack in Low word high word order.
                       'write low 16 bits to stack
F_PushStack32           call    #F_PushStack
                        'write high 16 bits to stack
                        shr     ForthTmp, #16
                        call    #F_PushStack
F_PushStack32_ret       ret
'
'
'Wait for a system command to be requested by forth engine
F_WaitForCmd            rdlong  ForthCmd, par
                        test    ForthCmd, ForthWordMask  wz
        if_z            jmp     #F_WaitForCmd
                        'Seperate stack address from command number
                        mov     ForthStackPtr, ForthCmd
                        shr     ForthStackPtr, #16
                        and     ForthCmd, ForthWordMask
                        'see if we have a valid command
                        cmp     ForthCmd, #c_ForthCmdCount                      wc
        if_c            jmp     #F_WaitForCmd_ret
                        ' we have an invalid command
                        call    #F_CompleteCommand                        
                        jmp     #F_WaitForCmd
                        ' we have a valid command, get the stack pointer address
F_WaitForCmd_Ret        ret

'

'Update stack pointer then exit
F_SaveStackPtrExit      shl     ForthStackPtr, #16
                        wrlong  ForthStackPtr, PAR
                        mov     ForthStackPtr, #0
F_SaveStackPtrExit_ret  ret
'
'
'Indicate system command requested by forth engine is completed, invalid command
F_CompleteCommand       mov     ForthCmd, #0                          
                        wrword  ForthCmd, par
                        mov     ForthStackPtr, #0
F_CompleteCommand_ret   ret

'
'Wait for a command and execute it
F_Do_Cmds               call    #F_WaitForCmd
                        'we have a cmd index 1..XX
                        mov     ForthTmp, #F_JUMP_TABLE
                        add     ForthTmp, ForthCmd
                        movs    :F_Set_Cmd_Jump, ForthTmp
                        nop
:F_Set_Cmd_Jump         mov     :F_Do_Cmd_Jump, (ForthTmp)
                        nop
:F_Do_Cmd_Jump          nop
                        'The called function will return here
F_Do_Cmds_Return        call    #F_SaveStackPtrExit
                        jmp     #F_Do_Cmds
'
'   
'Table of Jump opcodes for each of the supported forth functions
F_JUMP_TABLE            nop                             '(0)
                        call    #F_CMD_RdIna            '(1)
                        call    #F_CMD_RdOuta           '(2)
                        call    #F_CMD_WtOuta           '(3)
                        call    #F_CMD_RdDira           '(4)
                        call    #F_CMD_WtDira           '(5)
                        call    #F_CMD_RdInaPin         '(6)
                        call    #F_CMD_RdOutaPin        '(7)
                        call    #F_CMD_RdDiraPin        '(8)
                        call    #F_CMD_SetOutaPin       '(9)
                        call    #F_CMD_SetDiraPin       '(10)
                        call    #F_CMD_ClrOutaPin       '(11)
                        call    #F_CMD_ClrDiraPin       '(12)
                        call    #F_CMD_InvOutaPin       '(13)
                        call    #F_CMD_InvDiraPin       '(14)
                        call    #F_CMS_WaitPinsEqu      '(15)
                        call    #F_CMS_WaitPinsNotEqu   '(16)


' *********************************[ Forth Commands ]*********************************
'
'
' 1)  Read the I/O pins input register ina@ ( -- dAllPins )
F_CMD_RdIna             mov     ForthTmp, INA
                        call    #F_PushStack32
F_CMD_RdIna_ret         ret
'
' 2)  Read the I/O pins output register
' outa@ ( -- dRegValue )
F_CMD_RdOuta            mov     ForthTmp, OUTA
                        call    #F_PushStack32
F_CMD_RdOuta_ret        ret                        
'                        
' 3)  Write to the I/O pins output register OUTA
' outa! ( dValue -- )
F_CMD_WtOuta            call    #F_PopStack32
                        mov     OUTA, ForthTmp
F_CMD_WtOuta_ret        ret
'
' 4)  Read the I/O pins direction register
' dir@ ( -- dRegValue )
F_CMD_RdDira            mov     ForthTmp, dira
                        call    #F_PushStack32
F_CMD_RdDira_ret        ret                        
'
' 5)  Write to the I/O pins output register OUTA
' dira! ( dValue -- )
F_CMD_WtDira            call    #F_PopStack32
                        mov     DIRA, ForthTmp
F_CMD_WtDira_ret        ret
'
'
' Read the value of port pin. Returns; 0 (FALSE) if low and $FFFF is high (TRUE)
RdPortPin               call    #F_PopStack
                        mov     t1, ForthTmp
                        'We have the bit number (0..31), convert into bit mask
                        mov     ForthTmp, #1
                        shl     ForthTmp, t1
                        'See if bit is high or low
RdPinSource             test    ForthTmp, (ina)   wz
                        mov     ForthTmp, #0
              if_nz     sub     ForthTmp, #1                        
                        call    #F_PushStack
RdPortPin_ret           ret
'
' 6)  Read a sing1e pin from the I/O input register INA.
' ina-pin ( nPin -- TRUE | FALSE )
F_CMD_RdInaPin          movs    RdPinSource, #INA
                        call    #RdPortPin
F_CMD_RdInaPin_ret      ret                        
'                        
' 7)  Read a sing1e pin from the I/O output register OUTA.
' outa-pin ( nPin -- TRUE | FALSE )
F_CMD_RdOutaPin         movs    RdPinSource, #OUTA
                        call    #RdPortPin
F_CMD_RdOutaPin_ret     ret                        
'                        
'
' 8)  Read a sing1e pin from the I/O output direction register DIRA.
' dira-pin ( nPin -- TRUE | FALSE )
F_CMD_RdDiraPin         movs    RdPinSource, #DIRA
                        call    #RdPortPin
F_CMD_RdDiraPin_ret     ret                        
'
'
' Set the state of register pin high.
SetPortPin              call    #F_PopStack
                        'We have the bit number (0..31), convert into bit mask
                        mov     t1, #1
                        shl     t1, ForthTmp
                        'set the bit high
SetPinDest              or      (OUTA), t1
SetPortPin_ret          ret
'

' 9)  Set a sing1e pin high in the I/O output register OUTA.
' outa-setpin ( nPin -- )
F_CMD_SetOutaPin        movd    SetPinDest, #OUTA
                        call    #SetPortPin
F_CMD_SetOutaPin_ret    ret                        
'                        
' 10)  Set a sing1e pin high in the I/O output direction register DIRA.
' dira-setpin ( nPin -- )
F_CMD_SetDiraPin        movd    SetPinDest, #DIRA
                        call    #SetPortPin
F_CMD_SetDiraPin_ret    ret                        
'                        
'

' Set the state of register pin low.
ClrPortPin              call    #F_PopStack
                        'We have the bit number (0..31), convert into bit mask
                        mov     t1, #1
                        shl     t1, ForthTmp
                        'set the bit high
ClrPinDest              andn    (OUTA), t1
ClrPortPin_ret          ret
'

' 11)  Clear a sing1e pin low in the I/O output register OUTA.
' outa-clrpin ( nPin -- )
F_CMD_ClrOutaPin        movd    ClrPinDest, #OUTA
                        call    #ClrPortPin
F_CMD_ClrOutaPin_ret    ret                        
'                        
' 12)  Clear a sing1e pin low in the I/O output direction register DIRA.
' dira-clrpin ( nPin -- )
F_CMD_ClrDiraPin        movd    ClrPinDest, #DIRA
                        call    #ClrPortPin
F_CMD_ClrDiraPin_ret    ret                        
'                        
'
' Invert the state of register pin.
InvPortPin              call    #F_PopStack
                        'We have the bit number (0..31), convert into bit mask
                        mov     t1, #1
                        shl     t1, ForthTmp
                        'set the bit high
InvPinDest              xor     (OUTA), t1
InvPortPin_ret          ret
'

' 13)  Set a sing1e pin low in the I/O output register OUTA.
' outa-invpin ( nPin -- )
F_CMD_InvOutaPin        movd    InvPinDest, #OUTA
                        call    #InvPortPin
F_CMD_InvOutaPin_ret    ret                        
'                        
' 14)  Set a sing1e pin low in the I/O output direction register DIRA.
' dira-invpin ( nPin -- )
F_CMD_InvDiraPin        movd    InvPinDest, #DIRA
                        call    #InvPortPin
F_CMD_InvDiraPin_ret    ret                        
'                        
' 15) wait for the input pins spcified in the double dValue to equal
'     the INA register anded with the value of dMask.
' waitpeq ( dValue dMask -- )
F_CMS_WaitPinsEqu       call    #F_PopStack32
                        mov     t1, ForthTmp
                        call    #F_PopStack32
WaitPinsEq              waitpeq ForthTmp, t1
F_CMS_WaitPinsEqu_ret   ret
'
' 16) wait for the input pins spcified in the double dValue to not equal
'     the INA register anded with the value of dMask.
' waitpne ( dValue dMask -- )
F_CMS_WaitPinsNotEqu    call    #F_PopStack32
                        mov     t1, ForthTmp
                        call    #F_PopStack32
                        waitpne ForthTmp, t1
F_CMS_WaitPinsNotEqu_ret   ret
'
'
'Variables
'
t1      res   1