\ 4 wire com port for MV4th16

#102 VALUE C2_Blk     \ block where COG driver code is saved
0 VALUE C2_Cog      \ COG index when active, 0 when not running
0 VALUE C2_NxtHead  \ Next Tx Fifo Head value for write to Tx fifo

( serial data format settings )
\ offset_baud_rate    = 0     long sized variable containing the desired serial data rate (Bits per second)
\ offset_data_format  = 4     word sized variable containing the number of data bits, parity and stop bits per RX/TX byte
\                             word sized variable, bits 0..2 -- number of data bits -1, bit 3 -- use parity bit, 
\                             bit 4 -- odd/even parity 1 = odd, bit 5 -- TX extra stop bit.  
( port hardware pins )
\ offset_rx_pin       = 6     word sized variable containing pin number of RX data input pin (0..31)
\ offset_tx_pin       = 8     word sized variable containing pin number of TX data output pin (0..31)
\ offset_cts_pin      = 10    word sized variable containing pin number of CTS flow control input pin (0..31, or 255 to disable flow control)
\ offset_rts_pin      = 12    word sized variable containing pin number of RTS flow control output pin (0..31, or 255 to disable flow control)
( rx fifo settings )
\ offset_rx_head      = 14    address of word sized variable containg the RX fifo Head value (0..rx_buf_size - 1).
\ offset_rx_tail      = 16    address of word sized variable containg the RX fifo Tail value (0..rx_buf_size - 1).
\ offset_rx_buf_size  = 18    number of bytes of storage available in the RX fifo.
\ offset_rx_buf_adr   = 20    address of RX fifo data storage bytes.
( tx fifo settings )
\ offset_tx_head      = 22    address of word sized variable containg the TX fifo Head value (0..tx_buf_size - 1).
\ offset_tx_tail      = 24    address of word sized variable containg the TX fifo Tail value (0..tx_buf_size - 1).
\ offset_tx_buf_size  = 26    number of bytes of storage available in the TX fifo.
\ offset_tx_buf_adr   = 28    address of TX fifo data storage bytes.
( End of com port settings structure )
\ c_ParityOnMsk   = $0008
\ c_ParityOddMsk  = $0010
\ c_ExStopBitMsk  = $0020

#256 CONSTANT C2_RxFifoSize

#128 CONSTANT C2_TxFifoSize

C2_RxFifoSize QBYTES C2_RxBuf

C2_TxFifoSize QBYTES C2_TxBuf

#32 QBYTES C2_Pars

: C2-RxH ( -- adrRxHead )
  C2_Pars #14 + ;


: C2-RxT ( -- adrRxTail )
  C2_Pars #16 + ;


: C2-TxH ( -- adrTxHead )
  C2_Pars #22 + ;


: C2-TxT ( -- adrTxTail )
  C2_Pars #24 + ;


\ Clear the Tx fifo
: COM2-ClrTx ( -- )
  \ we can only alter Tx fifo head so set head to tx tail if in range
  C2-TxT @ DUP C2_TxFifoSize 1- U> IF
    DROP 0 DUP C2-TxT !
  THEN
  C2-TxH ! ;


\ Clear the Rx fifo
: COM2-ClrRx ( -- )  \ we can only alter Rx fifo tail so set tail to Rx head if in range
  C2-RxH @ DUP C2_RxFifoSize 1- U> IF
    DROP 0 DUP C2-RxH !
  THEN
  C2-RxT ! ;


\ calc the serial data format word used by the serial port driver
: C2-CalcFrmt ( uDataBits, fExtStopBit, CParity -- uFormat )
  ( now set parity bits 3..4, Parity is O -- odd, E -- even, else no parity )
  #32 OR  \ set parity char to lower case
  DUP [CHAR] e = IF
    DROP $08
  ELSE
    [CHAR] o = IF
      $18
    ELSE
     0
    THEN  
  THEN  
  ( uDataBits, fExtStopBit, uFormat )
  ( now add in extra Tx stop bit if required, bit 5 )
  SWAP IF
    $20 OR
  THEN
  ( uDataBits, uFormat )
  ( last add in data byte count bits 0..2 )
  SWAP 1- #7 AND OR ;


: C2-Init ( dBaudRate, uDataBits, fExtStopBit, CParity, uRxPin, uTxPin, uCtsPin, uRtsPin -- )
  C2_Pars >R
  C2_RxFifoSize R@ #18 + ! \ set Rx buffer size
  C2_RxBuf R@ #20 + !      \ set Rx fifo buffer address
  C2_TxFifoSize R@ #26 + ! \ set Tx buffer size
  C2_TxBuf R@ #28 + !      \ set Tx fifo buffer address
  ( save hardware pin numbers )
  R@ #12 + !         \ set rts pin number (0..31)
  R@ #10 + !         \ set cts pin number (0..31)
  R@ #08 + !         \ set tx data pin number (0..31)
  R@ #06 + !         \ set rx data pin number (0..31)
  ( calc the com port data format word )
  ( dBaudRate, uDataBits, fExtStopBit, CParity )
  C2-CalcFrmt
  ( dBaudRate, uFormat )
  R@ #04 + !         \ set serial data format  
  ( last save the buad rate in Propeller cip byte order )
  R@ Q!
  RDROP ;
  
: COM2-Stop ( -- )
  C2_Cog DUP 0<> IF
    COG-STOP
    0 TO C2_Cog
  ELSE
    DROP
  THEN ;
  
: COM2-Start ( dBaudRate, uDataBits, fExtStopBit, CParity, uRxPin, uTxPin, uCtsPin, uRtsPin -- )
  COM2-Stop
  COM2-ClrRx
  COM2-ClrTx
  C2-Init
  C2_Pars C2_Blk COG-LOAD TO C2_Cog ;
  

\ return flag indicating if there is data to read from Rx Fifo  
: COM2-IN? ( -- fCanRead )
  C2-RxT @ C2-RxH @ <> ;
  

\ wait for data byte in Rx fifo and return it
: COM2-IN ( -- cData )
  BEGIN
    COM2-IN?
  UNTIL
  C2-RxT @ DUP C2_RxBuf + C@
  ( uRxTail cData )
  SWAP 1+ DUP C2_RxFifoSize >= IF
    DROP 0
  THEN
  C2-RxT ! ;

\ return flag indicating if there is room to write data to the Tx Fifo  
: COM2-OUT? ( -- fCanWrite )
  C2-TxT @ C2-TxH @
  ( uTxTail uTxHead )
  1+ DUP C2_TxFifoSize >= IF
    DROP 0
  THEN
  ( uTxTail uTxHead+1 )
  DUP TO C2_NxtHead <> ;


\ wait for space in Tx fifo then add cData to Tx fifo
: COM2-OUT ( cData -- )
  BEGIN
    COM2-OUT?
  UNTIL
  C2-TxH @
  ( cData uTxHead )
  C2_TxBuf + C!
  C2_NxtHead C2-TxH ! ;
  
: COM2>CON ( -- )
  ['] COM2-IN? IS KEY? ['] COM2-IN IS KEY
  ['] COM2-OUT? IS EMIT? ['] COM2-OUT IS EMIT
  ['] COM2-ClrRx IS EMPTY-KEYS ;
  
\ 57600. 8 FALSE CHAR n 24 25 255 255 COM2-Start 

\ 9600. 8 FALSE CHAR n 15 16 14 13 COM2-Start 

\ : test BEGIN KEY? 0= WHILE C2-IN? IF C2-IN . THEN REPEAT ;
  