''***************************************
''*  Full-Duplex Serial Driver v1.1     *
''*  Author: Chip Gracey                *
''*  Copyright (c) 2006 Parallax, Inc.  *
''*  See end of file for terms of use.  *
''***************************************


con
  c_ParityOnMsk                 = $0008
  c_ParityOddMsk                = $0010
  c_ExStopBitMsk                = $0020

  ' The setting for serial port, must be long aligned
  ' address passed via PAR register when cog starts.
  '
  'serial data format settings
  offset_baud_rate              = 0                     ' long sized variable containing the desired serial data rate (Bits per second)
  offset_data_format            = 4                     ' word sized variable containing the number of data bits, parity and stop bits per RX/TX byte
                                                        ' word sized variable, bits 0..2 -- number of data bits -1, bit 3 -- use parity bit, 
                                                        ' bit 4 -- odd/even parity 1 = odd, bit 5 -- TX extra stop bit.  
  'port hardware pins
  offset_rx_pin                 = 6                     ' word sized variable containing pin number of RX data input pin (0..31)
  offset_tx_pin                 = 8                     ' word sized variable containing pin number of TX data output pin (0..31)
  offset_cts_pin                = 10                    ' word sized variable containing pin number of CTS flow control input pin (0..31, or 255 to disable flow control)
  offset_rts_pin                = 12                    ' word sized variable containing pin number of RTS flow control output pin (0..31, or 255 to disable flow control)
  'rx fifo settings
  offset_rx_head                = 14                     ' address of word sized variable containg the RX fifo Head value (0..rx_buf_size - 1).
  offset_rx_tail                = 16                     ' address of word sized variable containg the RX fifo Tail value (0..rx_buf_size - 1).
  offset_rx_buf_size            = 18                     ' number of bytes of storage available in the RX fifo.
  offset_rx_buf_adr             = 20                     ' address of RX fifo data storage bytes.
  'tx fifo settings
  offset_tx_head                = 22                    ' address of word sized variable containg the TX fifo Head value (0..tx_buf_size - 1).
  offset_tx_tail                = 24                    ' address of word sized variable containg the TX fifo Tail value (0..tx_buf_size - 1).
  offset_tx_buf_size            = 26                    ' number of bytes of storage available in the TX fifo.
  offset_tx_buf_adr             = 28                    ' address of TX fifo data storage bytes.
  'End of com port settings structure
  '
  'Min free bytes in rx fifo for RTS to be active (low)
  c_RxFullLevel                 = 16

PUB start

  
DAT

'***********************************
'* Assembly language serial driver *
'***********************************
                        org
'Start of cog assembly code for serial port
                        jmp     #entry
'
' Calc Bit mask (t2 = bit number on entry, on exit t2 = mask)
ctemp                   long    0
clc_msk                 mov     ctemp, t2
                        mov     t2, #0
                        ' not using pin, mask should be 0 (no bit in mask)
                        cmp     ctemp, #255        wz
              if_z      jmp     #clc_msk_ret
                        ' have pin number
                        mov     t2, #1
                        shl     t2, ctemp
clc_msk_ret             ret
'                        
'
'Unsigned 32 bit division with 32 bit result and 32 bit remainder.
'On entry T1 - dividend, T2 = divisor. On exit T3 = T1 / T2 
DoDivU32                tjnz    T2, #DoDivU32_1         'If T2 = 0 we have a divide by zero error
                        'Handle division by zero, return max value and remainder of 0
                        mov     T3, #0
                        sub     T3, #1
                        jmp     #DoDivU32_ret
DoDivU32_1              'Perform a normal division                                                              
                        mov     T3, #0
                        mov     T4, #0
                        mov     T5, #32                                   
                        'Do the shift and sub a max of 32 times.
DoDivU32_2              shl     T1, #1          wc
                        rcl     T4, #1
                        cmpsub  T4, T2         wc      'C flag is set (1) if a subtraction is possible.
                        rcl     T3, #1
                        djnz    T5, #DoDivU32_2
DoDivU32_ret            ret
'
'
                        ' First load baud rate (first 4 bytes of structure)
entry                   mov     t1, par                 'get structure address
                        mov     TotRxDataBits, #1       'set total RX bits in one data byte to one for stop bit
                        ' First comes baud rate
                        rdlong  t2, par                 'read baud rate from 
                        rdlong  t1, #0                  'read propeller chip clock frequency from address 0 in global memoru
                        call    #DoDivU32               'calc bit ticks (Prop Freq / Baud rate)
                        mov     BitTicks, t3
                        ' save half bit tick time for start bit offset delay
                        mov     BitTicksDiv2, BitTicks
                        shr     BitTicksDiv2, #1
                        ' Load pin settings and uart format options.
                        '
                        ' Read number of data bits per word (5,7,8)
                        mov     t1, par                 'get structure address
                        add     t1, #offset_data_format 'skip past baud rate
                        rdword  DataBits, t1
                        ' decode parity settings
                        mov     ParityMode, #0
                        test    DataBits, c_ParityOnMsk wz
                        test    DataBits, c_ParityOddMsk wc
           if_nz        add     TotRxDataBits, #1       ' add one to bit count for parity bit            
           if_nz_and_c  mov     ParityMode, #1          ' odd parity mode
           if_nz_and_nc mov     ParityMode, #2          ' even parity mode
                        ' decode extra stop bit
                        mov     ExtraStopBitMode, #0
                        test    DataBits, c_ExStopBitMsk wz
           if_nz        mov     ExtraStopBitMode, #1                        
                        ' decode number of data bits per word (1..8)
                        and     DataBits, #7            ' read data bit count -1
                        add     DataBits, #1            ' add 1 to get number of data bits per byte
                        add     TotRxDataBits, DataBits ' add number of data bits to bit count            
                        'Set data byte mask based on data bits
                        mov     DataMask, #$FF
                        mov     t2, #8
                        sub     t2, DataBits    wc, wz
        if_nz_and_nc    shr     DataMask, t2
                        '                        
                        ' Read hardware pin numbers for RX, TX, CTS, and RTS pins, then calc pin masks.
                        movd    :next_pin_save, #rxmask ' first pin to calculate mask for is rx pin    
                        mov     t3, #4                  ' 4 pin settings to read
:next_pin               add     t1, #2                  ' t1 = address of next pin number
                        rdword  t2, t1
                        call    #clc_msk
:next_pin_save          mov     (rxmask), t2
                        add     :next_pin_save, D0     
                        djnz    t3, #:next_pin
                        '
                        ' Read rx fifo structure
                        'calc rx fifo head global memory address
                        mov     RxHeadAdr, #offset_rx_head
                        add     RxHeadAdr, par
                        'calc rx fifo tail global memory address
                        mov     RxTailAdr, #offset_rx_tail
                        add     RxTailAdr, par
                        'calc rx fifo byte size address, and read fifo buffer byte size from global memory word 
                        mov     RxBuffSize, #offset_rx_buf_size
                        add     RxBuffSize, par
                        rdword  RxBuffSize, RxBuffSize
                        'calc rx fifo memory address address, and read fifo buffer address from global memory word 
                        mov     RxBuffadr, #offset_rx_buf_adr
                        add     RxBuffAdr, par
                        rdword  RxBuffAdr, RxBuffAdr
                        '
                        ' Read tx fifo structure
                        'calc tx fifo head global memory address
                        mov     TxHeadAdr, #offset_tx_head
                        add     TxHeadAdr, par
                        'calc tx fifo tail global memory address
                        mov     TxTailAdr, #offset_tx_tail
                        add     TxTailAdr, par
                        'calc tx fifo byte size address, and read fifo buffer byte size from global memory word 
                        mov     TxBuffSize, #offset_tx_buf_size
                        add     TxBuffSize, par
                        rdword  TxBuffSize, TxBuffSize
                        'calc tx fifo memory address address, and read fifo buffer address from global memory word 
                        mov     TxBuffadr, #offset_tx_buf_adr
                        add     TxBuffAdr, par
                        rdword  TxBuffAdr, TxBuffAdr
                        '
                        'copy current Rx fifo Head to shadow RXHead
                        rdword  SdwRxHead, RxHeadAdr    'SdwRxHead = rx head
                        'copy current Tx fifo Tail to shadow TxTail
                        rdword  SdwTxTail, TxTailAdr    'SdwTxTail = tx tail
                        '
                        mov     txcode, #transmit       'initialize ping-pong multitasking
                        '
                        'configure tx data pin for output then set high
                        or      outa, txmask            'set tx pin high
                        or      dira, txmask            'set tx pin as an output
                        'set RTS pin as an output, set pin low to allow data transfer
                        andn    outa, rtsmask           'set RTS pin low
                        or      dira, rtsmask           'set RTS pin as output
'
'
' Receive
'
receive                 jmpret  rxcode,txcode         'run a chunk of transmit code, then return
                        'Set RTS to low if fifo is not full
                        rdword  t2, RxTailAdr           'T2 = rx tail
                        sub     t2, SdwRxHead   wc, wz  'T2 = tail - head
                        'If Head >= tail
        if_c_or_z       add     t2, #RxBuffSize        ' head < tail, Count=Tail - Head
                        'See if enough space to allow reception
                        cmp     t2, #c_RxFullLevel wc
                        muxc    outa, rtsmask           'Set RTS, if we have >= c_RxFullLevel bytes of free space set low else set high
                        '
                        'Check for start bit, rx pin low
                        test    rxmask, ina     wz
        if_nz           jmp     #receive              'not ready to receive  
                        'we have a start bit, get ready to receive byte
                        mov     rxbits, TotRxDataBits    'number of bits in a data word (should be 5,7,8 plus parity bit, and stop bit) 
                        mov     rxcnt, bitTicksDiv2    'set count to one full bit time divided by 2 to put us in the middle of bit  
                        add     rxcnt,cnt             'calc system start time            
                        'receive next bit
:bit                    add     rxcnt,bitTicks        'ready next bit period
:wait                   jmpret  rxcode,txcode         'run a chuck of transmit code, then return
                        'wait for bit time to expire
                        mov     t1, rxcnt             'check if bit receive period done
                        sub     t1, cnt
                        cmps    t1, #0          wc
        if_nc           jmp     #:wait
                        'bit time is up read next bit from rx pin
                        test    rxmask,ina      wc    'receive bit on rx pin
                        rcr     rxdata,#1
                        djnz    rxbits,#:bit
                        'we have all data bits plus a stop bit.
                        'The data is in the MSB byte need to shift it to LSB
                        'if parity is on, we need to remove parity bit from data
                        rol     rxdata, TotRxDataBits     
                        and     rxdata, datamask
                       ' we have a byte write it to the fifo
                       'check for, fifo full, head + 1 = tail
                        mov     t3, SdwRxHead           'T3 = rx head
                        add     t3, #1                  'T3 = head + 1, still need to wrap
                        cmpsub  t3, RxBuffSize          'T3 = head +1 mod RxFifoSize, in case head wrapped
                        rdword  t2, RxTailAdr           'T2 = rx tail
                        cmp     t2, t3          wz      'if head + 1 = tail, fifo full
        if_z            jmp     #receive                'if full can not receive, dump char
                        'write the data byte to the fifo
                        'save received byte to rxHead and then inc rxHead
                        mov     t1, SdwRxHead         'T2 = current rxHead value
                        add     t1, RxBuffAdr         'Add rxBuffer global memory address
                        wrbyte  rxdata, t1            'Save the new data to the fifo
                        mov     SdwRxHead, t3         'Point head to next location
                        wrword  SdwRxHead, RxHeadAdr  'Update the rxHead in global memory
                        jmp     #receive              'byte done, receive next byte
'
'
' Transmit
'
transmit                jmpret  txcode,rxcode           'run a chunk of receive code, then return
                        'if CTS is low we can send, else we are done
                        test    ctsmask, ina    wc      'Carry = CTS pin value
        if_nc           rdword  t2,TxHeadAdr            'T2 = txHead
        if_nc           cmp     t2, SdwTxTail   wz      'see if Head = Tail, indicates nothing to send
        if_c_or_z       jmp     #transmit               'if no data to send do recieve code
                        'return to receive code and run a chunk
:txdorx1                jmpret  txcode,rxcode         'run a chunk of receive code, then return
                        'There is at least one byte to read from tx fifo
                        mov     t3, SdwTxTail           'Read the tx tail from our cache
                        add     t3, TxBuffAdr           'calc address of next fifo byte to read
                        rdbyte  txdata, t3              'txdata = next byte to send
                        'return to receive code and run a chunk
:txdorx2                jmpret  txcode,rxcode         'run a chunk of receive code, then return
                        'Data in txdata, now inc tx tail
                        add     SdwTxTail, #1                  ' point to next fifo location
                        cmpsub  SdwTxTail, TxBuffSize          ' SdwTxTail = Tail +1 mod TxFifoSize, in case head wrapped
                        wrword  SdwTxTail, TxTailAdr           ' save tx tail to global memory
                        'Data byte to transmit...
                        and     txdata, datamask
                        'Add a stop bit(s) and possibly parity to our data
                        mov     txbits, databits        ' number of data bits to send
                        add     txbits, #3              '3 extra bits; start, stop, back to idle
                        'See if we need 1 or 2 stop bits
                        mov     temp, #1
                        tjz     ExtraStopBitMode, #:txdorx5
                        'Add a second stop bit
                        add     temp, #2
                        add     txbits, #1              'extra bit for second stop bit
                        
                        'See if we need parity bit
:txdorx5                tjz     paritymode, #:txdorx4
                        ' we need to add parity
                        or      txdata, txdata  wc      'Calc data parity in case we need it
                        rcl     temp, #1                'Add in the parity
                        test    paritymode, #1  wz      'see if we want odd parity
        if_nz           xor     temp, #1                'flip parity bit
                        add     txbits, #1              'extra bit for parity bit
                        ' Add in our stop bits
:txdorx4                shl     temp, databits
                        or      txdata, temp
                        'Add the start bit and idle bit to the data
                        shl     txdata,#2
                        or      txdata,#1
                        mov     txcnt,cnt
                        'return to receive code and run a chunk
:txdorx3                jmpret  txcode,rxcode         'run a chunk of receive code, then return

:bit                    shr     txdata,#1       wc
                        muxc    outa,txmask        
                        add     txcnt,bitTicks        'ready next cnt

:wait                   jmpret  txcode,rxcode         'run a chunk of receive code, then return

                        mov     t1,txcnt              'check if bit transmit period done
                        sub     t1,cnt
                        cmps    t1,#0           wc
        if_nc           jmp     #:wait

                        djnz    txbits,#:bit          'another bit to transmit?

                        jmp     #transmit             'byte done, transmit next byte
'
'
' Constants
D0                      long    $0200

' Com port data format and baud rate settings
BitTicks                res     1
BitTicksDiv2            res     1
DataBits                res     1
TotRxDataBits           res     1
ParityMode              res     1
ExtraStopBitMode        res     1
DataMask                res     1
'
'I/O masks for pins (must be in order (rx, tx, cts, rts).
rxmask                  res     1
txmask                  res     1
ctsmask                 res     1
rtsmask                 res     1
'
' RX fifo params must have same order as settings
RxHeadAdr               res     1
RxTailAdr               res     1
RxBuffSize              res     1
RxBuffAdr               res     1
'
' TX fifo params must have same order as settings
TxHeadAdr               res     1
TxTailAdr               res     1
TxBuffSize              res     1
TxBuffAdr               res     1
'
' Shadow copies of fifo pointers (Rx Head, Tx Tail) we modify in this COG
SdwRxHead               res     1         'Only this cog updates rxHead so we keep a shadow copy for speed
SdwTxTail               res     1         'Only this cog updates txTail so we keep a shadow copy for speed
'
' Uninitialized data
t1                      res     1
t2                      res     1
t3                      res     1
t4                      res     1
t5                      res     1
temp                    res     1
'
rxdata                  res     1
rxbits                  res     1
rxcnt                   res     1
rxcode                  res     1
'
txdata                  res     1
txbits                  res     1
txcnt                   res     1
txcode                  res     1
'