\ MV4th16 Hello World Example
\
\ No addition forth libraries are required.
\ 
\ Requires an led and a 300..600 ohm resistor.
\ Connect one end of resistor to a pin on the propeller chip.
\ Connect the other end of the resistor to the positive (anode) of led.
\ Connect the led negative side (cathode) to ground.
\
\ set the pin number as a constant
0 CONSTANT led_pin

\ pin mask constant to set led pin on propeller chip I/O
#1. led_pin DLSHIFT 2CONSTANT led_set_mask

\ pin mask constant to clear led pin on propeller chip I/O
led_set_mask $FFFFFFFF. DXOR 2CONSTANT led_clr_mask

\ set bits high for the specified propeller chip COG register
: cog-or-reg ( dPinMask uCogReg -- )
  \ duplicate and save cog register to return stack
  DUP >R
  ( dPinMask uCogReg  R: uCogReg )
  \ read the current register value
  COG@
  ( dPinMask dRegValue  R: uCogReg )
  \ logical OR requested bits
  DOR 
  ( dNewRegValue  R: uCogReg )
  \ restore COG register index from return stack
  R> 
  ( dNewRegValue uCogReg )
  \ update COG register
  COG! ; 

\ clear bits low for the specified propeller chip COG register
: cog-and-reg ( dPinMask uCogReg -- )
  \ duplicate and save cog register to return stack
  DUP >R
  ( dPinMask uCogReg  R: uCogReg )
  \ read the current register value
  COG@
  ( dPinMask dRegValue  R: uCogReg )
  \ logical OR requested bits
  DAND 
  ( dNewRegValue  R: uCogReg )
  \ restore COG register index from return stack
  R> 
  ( dNewRegValue uCogReg )
  \ update COG register
  COG! ; 

\ enable the led pin as an output
: led-enable ( -- )
  led_set_mask $1F6 cog-or-reg ;

\ set the led pin as an input
: led-disable ( -- )
  led_clr_mask $1F6 cog-and-reg ;

\ set led pin to 3.3v
: led-on ( -- )
  led_set_mask $1F4 cog-or-reg ;

\ set the led pin to ground
: led-off ( -- )
  led_clr_mask $1F4 cog-and-reg ;

\ flash the led the specified number of times
: led-flash ( uCount -- )
  \ activate pin connected to led as an output
  led-enable
  \ loop for the number of flashes requested
  0 ?DO
    \ trun on led
    led-on
    \ delay for 600 milli-seconds
    #600 MS
    \ trun off the led
    led-off
    \ delay for 600 milli-seconds
    #600 MS
  LOOP
  \ all done set led pin to input  
  led-disable ;
    
