''Forth word Find implemented in propeller assembly
''For MV4th16 project
''Version 1.0
''Sept 9, 2012
''www.shaels.net
''
CONST
  c_MaxSearchLen = 64
  c_ForthCmdCount = 2
  c_ImmediateFlag = 1
  
PUB Start

DAT

Entry                   jmp     #F_Do_Cmds

'FORTH system command support routines
'
' Remove top 16 bit value from forth stack.T1 = 16 bit value on return
                        'make sure we are in a command and stack pointer is valid
F_PopStack              tjz     ForthStackPtr, #F_PopStack_Ret
                        'We are in a command OK to pop value from stack
                        rdword  T1, ForthStackPtr
                        add     ForthStackPtr, #2
F_PopStack_Ret          ret             
'
' Add 16 bit value to top of forth stack.T1 = 16 bit value to write.
                        'make sure we are in a command and stack pointer is valid
F_PushStack             tjz     ForthStackPtr, #F_PushStack_Ret
                        'We are in a command OK to push a value to the stack
                        sub     ForthStackPtr, #2
                        wrword  T1, ForthStackPtr
F_PushStack_Ret         ret             
'
'
'Wait for a system command to be requested by forth engine
F_WaitForCmd            rdlong  ForthCmd, par
                        test    ForthCmd, DataWordMask  wz
        if_z            jmp     #F_WaitForCmd
                        'Seperate stack address from command number
                        mov     ForthStackPtr, ForthCmd
                        shr     ForthStackPtr, #16
                        and     ForthCmd, DataWordMask
                        'see if we have a valid command
                        cmp     ForthCmd, #c_ForthCmdCount                      wc
        if_c            jmp     #F_WaitForCmd_ret
                        ' we have an invalid command
                        call    #F_CompleteCommand                        
                        jmp     #F_WaitForCmd
                        ' we have a valid command, get the stack pointer address
F_WaitForCmd_Ret        ret

'
'Update stack pointer then exit
F_SaveStackptr          mov     T1, par
                        add     T1, #2
                        wrword  ForthStackPtr, T1
'Indicate system command requested by forth engine is completed
F_CompleteCommand       mov     ForthCmd, #0                          
                        wrword  ForthCmd, par
                        mov     ForthStackPtr, #0
F_SaveStackptr_ret                              
F_CompleteCommand_ret   ret
'
'
'Load the search string into local buffer
'Search string is prefixed by length byte
'On Entry
'  SearchStrAdr = address of string to search for in dictionary.
'
LoadStr                 mov     T1, SearchStrAdr
                        rdbyte  SearchByteCnt, T1
                        add     T1, #1
                        'Copy the search string to the buffer.
                        'if search string is zero length we are done.
                        tjz     SearchByteCnt, #LoadStr_ret
                        'There is at least 1 character to copy.
                        'First save the length byte
                        mov     SearchBuf, SearchByteCnt
                        movd    LoadStrCpy, #SearchBuf + 1    
                        mov     T2, SearchByteCnt
                        'Limit copy to c_MaxSearchLen bytes
                        max     T2, #c_MaxSearchLen
                        'Copy the bytes to the buffer
LoadStrCpy              rdbyte  (SearchBuf), T1
                        add     T1, #1
                        add     LoadStrCpy, D0
                        djnz    T2, #LoadStrCpy
LoadStr_ret             ret
'
'
'Compare a length prefixed string to the string in search buffer for an exact match
'On Entry:
'  T1 = address of string to compare with search string
'On Exit:
'  Z = 0 if no match, Z = 1 string matches

CmpStr                  rdbyte  T2, T1                  'read string length
                        cmp     T2, SearchByteCnt    wz 'see if length same as search string
              if_nz     jmp     #CmpStr_ret
                        'String is same length, see if it matches
                        'Limit compare to c_MaxSearchLen bytes
                        max     T2, #c_MaxSearchLen
                        'Start compare at first character after length                                                
                        movd    CmpStrCompare, #SearchBuf + 1    
                        'Loop and compare string bytes
CmpStrNext              add     T1, #1          'point to next char
                        rdbyte  T3, T1
CmpStrCompare           cmp     (SearchBuf), T3     wz
                        add     CmpStrCompare, D0
              if_z      djnz    T2, #CmpStrNext
CmpStr_ret              ret
'
'
'Read the next forth word NFA (name field address)
'On Entry:
'  LatestAdr = address of current forth word name field
'On Exit:
'  LatestAdr = address of next forth word name field or zero if at last word
GetNextNFA              tjz     LatestAdr, #GetNextNFA_ret
                        'Convert NFA to LFA
                        sub     LatestAdr, #4
                        'read word at link field address to get prior NFA
                        rdword  LatestAdr, LatestAdr
GetNextNFA_ret          ret
'
'
'Perform forth FastFind function
' Foth stack has search string address, and value of forth LATEST variable
'  FastFind ( c-addr uLatest -- c-addr 0 ) or ( c-addr uLatest -- xt 1 | -1 )
F_CMD_FIND              call    #F_PopStack     'LatestAdr = address of last forth word name field
                        mov     LatestAdr, T1
                        call    #F_PopStack     'SearchStrAdr = search string memory address
                        mov     SearchStrAdr, T1
                        'Copy search string to buffer so we only read once
                        call    #LoadStr
                        'If search string is blank no match is possible
                        tjz     SearchByteCnt, #F_CMD_FIND_Failed
                        'see if we can match any dictionary name
                        tjz     LatestAdr, #F_CMD_FIND_Failed
F_CMD_FIND_Next         mov     T1, LatestAdr
                        call    #CmpStr
              if_z      jmp     #F_CMD_FIND_Ok
                        'no match try next forth word
                        call    #GetNextNFA
                        tjnz    LatestAdr, #F_CMD_FIND_Next
                        'failed to match, return c-addr and 0 on stack
F_CMD_FIND_Failed       mov     T1, SearchStrAdr
                        call    #F_PushStack
                        mov     T1, #0
                        call    #F_PushStack
                        jmp     #F_CMD_FIND_ret
                        'We have a match, LatestAdr = NFA of matching forth word
                        'Convert the NFA to the CFA (code field address) of forth word
                        'Return a 1 if word is immediate, return a -1 if normal    
F_CMD_FIND_Ok           mov     T1, LatestAdr   'Calculate forth word FFA (flag field address)
                        sub     T1, #2
                        'T1 = matching word FFA
                        rdword  T1, T1          'T1 = matching forth word flags
                        test    T1, #c_ImmediateFlag    wc
                        'Immediate flag saved in carry flag
                        'C = 0 word is normal, C = 1 word is immediate
                        'Convert matcing forth word NFA to CFA (code field address)
                        rdbyte  T1, LatestAdr   'Read forth word name length
                        add     T1, #1          'Add one to length for string length byte
                        'if the string length is odd word align to next word boundry
                        test    T1, #1          wz
              if_nz     add     T1, #1                        
                        'Add string size to NFA to get CFA
                        add     T1, LatestAdr
                        call    #F_PushStack
                        mov     T1, DataWordMask        'set to -1 (16 bit)
              if_c      mov     T1, #1                                                
                        call    #F_PushStack
F_CMD_FIND_ret          ret
'
'
'Wait for a command and execute it
F_Do_Cmds               call    #F_WaitForCmd
                        'we have a cmd index 1..XX
                        mov     T1, #F_JUMP_TABLE
                        add     T1, ForthCmd
                        movs    :F_Set_Cmd_Jump, T1
                        nop
:F_Set_Cmd_Jump         mov     :F_Do_Cmd_Jump, (T1)
                        nop
:F_Do_Cmd_Jump          nop
                        'The called function will return here
F_Do_Cmds_Return        call    #F_SaveStackptr
                        jmp     #F_Do_Cmds
'
'   
'Table of Jump opcodes for each of the supported forth functions
F_JUMP_TABLE            nop                             '(0)
                        call    #F_CMD_FIND             '(1)

                        
'                        
'
'Forth command variables ForthCmdAdr (16 bits) followed directly by StackPtrAdr (16 bits)
ForthCmd                long    0
ForthStackPtr           long    0
'constants
DataWordMask            long    $FFFF
D0                      long    512
'Variables
SearchStrAdr            long    0
SearchByteCnt           long    0
'SearchBuf should be last allocation
SearchBuf               res     c_MaxSearchLen + 1
T1                      res     1
T2                      res     1
T3                      res     1
LatestAdr               res     1