\ Fast Find code, test for EEPROM and loads COG code if required
\ SD card I/O contains fast find code so no additional COG required

#3 VALUE FF_Blk \ indicates what block to load COG from
  
0 VALUE FF_Cog  \ if we need COG for fast find, save COG index

QVARIABLE FF_Data  \ command buffer for COG-EX if we need it

\ replacement FIND function
: (ff-find) ( c-adr -- xt 1 | -1 ) ( c-adr -- c-adr 0 )
  LATEST @
  FF_Cog IF
    \ fast find COG is running
    FF_Data 1 COG-EX
  ELSE
    \ fast find COG is not running use SD Card I/O COG 
    #25 SYSTEM
  THEN ;

: FF-STOP ( -- )
  FF_Cog ?DUP IF
    COG-STOP
    0 TO FF_Cog
  THEN
  \ unhook the fast find code, revert to dictionary FIND
  ['] (find) IS FIND ;

: FF-RUN ( -- )
  FF-STOP
  blk-count #128 <= IF
    \ EEPROM block device load COG for fast find 
    0. FF_Data Q!  \ clear the command buffer
    FF_Data FF_Blk COG-LOAD  \ load and run COG
    DUP TO FF_Cog  \ save the COG index (0..7)
    \ if no COG report error and die, better then a lockup
    0= ABORT" Unable to load FF COG"
  THEN
  \ hook the fast find code
  ['] (ff-find) IS FIND ;
  
  