'' Test Si4703 FM rx IC

const
  SDA_PIN                       = 0
  CLK_PIN                       = 1
  RST_PIN                       = 2
  INT_PIN                       = 3
  
  c_ForthCmdCount               = 5
  
  c_Si4703_ReadCmd              = $21
  c_Si4703_WriteCmd             = $20

pub Start
  
dat

Entry                   jmp     #F_Do_Cmds

'*********************************[ FORTH system command support routines]*********************************
'
' Variables used by forth SYSTEMEX interface
ForthTmp                long    0
ForthCmd                long    0
ForthStackPtr           long    0
ForthScratch            long    0   
' Constants
ForthWordMask           long    $FFFF


' Remove top 16 bit value from forth stack.ForthTmp = 16 bit value on return
                        'make sure we are in a command and stack pointer is valid
F_PopStack              tjz     ForthStackPtr, #F_PopStack_Ret
                        'We are in a command OK to pop value from stack
                        rdword  ForthTmp, ForthStackPtr
                        add     ForthStackPtr, #2
F_PopStack_Ret          ret             
'
' Add 16 bit value to top of forth stack.ForthTmp = 16 bit value to write.
                        'make sure we are in a command and stack pointer is valid
F_PushStack             tjz     ForthStackPtr, #F_PushStack_Ret
                        'We are in a command OK to push a value to the stack
                        sub     ForthStackPtr, #2
                        wrword  ForthTmp, ForthStackPtr
F_PushStack_Ret         ret             
'
'
'Wait for a system command to be requested by forth engine
F_WaitForCmd            rdlong  ForthCmd, par
                        test    ForthCmd, ForthWordMask      wz
        if_z            jmp     #F_WaitForCmd
                        'Seperate stack address from command number
                        mov     ForthStackPtr, ForthCmd
                        shr     ForthStackPtr, #16
                        and     ForthCmd, ForthWordMask
                        'see if we have a valid command
                        cmp     ForthCmd, #c_ForthCmdCount   wc, wz
        if_c_or_z       jmp     #F_WaitForCmd_ret
                        ' we have an invalid command
                        call    #F_CompleteCommand                        
                        jmp     #F_WaitForCmd
                        ' we have a valid command, get the stack pointer address
F_WaitForCmd_Ret        ret

'

'Update stack pointer then exit
F_SaveStackPtrExit      shl     ForthStackPtr, #16
                        wrlong  ForthStackPtr, PAR
                        mov     ForthStackPtr, #0
F_SaveStackPtrExit_ret  ret
'
'
'Indicate system command requested by forth engine is completed, invalid command
F_CompleteCommand       mov     ForthCmd, #0                          
                        wrword  ForthCmd, par
                        mov     ForthStackPtr, #0
F_CompleteCommand_ret   ret

'
'Wait for a command and execute it
F_Do_Cmds               call    #F_WaitForCmd
                        'we have a cmd index 1..XX
                        mov     ForthTmp, #F_JUMP_TABLE
                        add     ForthTmp, ForthCmd
                        movs    :F_Set_Cmd_Jump, ForthTmp
                        nop
:F_Set_Cmd_Jump         mov     :F_Do_Cmd_Jump, (ForthTmp)
                        nop
:F_Do_Cmd_Jump          nop
                        'The called function will return here
F_Do_Cmds_Return        call    #F_SaveStackPtrExit
                        jmp     #F_Do_Cmds
'
'   
'Table of Jump opcodes for each of the supported forth functions
F_JUMP_TABLE            nop                              '(0)
                        call    #Cmd_Reset               '(1)
                        call    #Cmd_RdRegisters         '(2)
                        call    #Cmd_WtRegisters         '(3)
                        call    #Cmd_SetDeviceIdx        '(4)
                        call    #Cmd_GetDeviceIdx        '(5)


' *********************************[ Forth Commands ]*********************************
'
' Initialize the I2C interface, on exit SCLK and SDA are high
i2c_Init                andn    outa, SdaPinMsk         ' ready i2c data pin for open drain output
                        andn    outa, ClkPinMsk         ' ready i2c clk pin for open drain output
                        ' next two opcodes are not needed unless we call Entry again
                        andn    dira, SdaPinMsk         ' set i2c data pin high
                        andn    dira, ClkPinMsk         ' set i2c clk pin high
                        ' The i2c data and clock pins are now high
i2c_Init_ret            ret 

' Delay for one half an i2c bit time, for si470x on my LCD this is 1.25us (400khz max clock rate / 2)
Delay_Temp    long      0
'
' For PC8574 on my LCD max i2c clock rate is 100Khz.
' This makes the min period for a clock high or low 5us (10us / 2).
'
' Delay for one half of i2c clock bit time or 2.5us (10us / 4)
HalfBitDelay            mov     Delay_Temp, I2C_Bit_Delay
                        shr     Delay_Temp, #1          ' half i2c bit delay clock time
                        ' we have five opcodes + call into this + return = (5 * 4 + 4 + 4) = 24 cycles
                        sub     Delay_Temp, #24
                        add     Delay_Temp, cnt
                        waitcnt Delay_Temp, #0
HalfBitDelay_ret        ret                                                                              
'
'
' Set the i2c clk line high, delay half a min clock period
ClkHigh                 andn    dira, ClkPinMsk         ' set i2c clk pin high
                        call    #HalfBitDelay
ClkHigh_ret             ret
'
' 
' Set the i2c clk line low, delay half a min clock period
ClkLow                  or      dira, ClkPinMsk         ' set i2c clk pin low
                        call    #HalfBitDelay
ClkLow_ret              ret
'
' 
' Set the i2c data line high, delay half a min clock period
SdaHigh                 andn    dira, SdaPinMsk         ' set i2c data pin high
                        call    #HalfBitDelay
SdaHigh_ret             ret
'
' 
' Set the i2c clk line low, delay half a min clock period
SdaLow                  or      dira, SdaPinMsk         ' set i2c data pin low
                        call    #HalfBitDelay
SdaLow_ret              ret
'
' 
' Send an i2c start bit, SCLK will be set Low on exit, ausume SCLK and SDAT high on entry 
i2c_start               call    #SdaLow
                        call    #ClkLow
i2c_start_ret           ret
'
'
' Send an i2c stop bit, assume clk is low and data is undetermined, both data and clk will be set high.
i2c_stop                call    #SdaLow                 ' set i2c data pin low
                        call    #ClkHigh                ' set i2c clock pin high
                        call    #SdaHigh                ' set i2c data pin High
i2c_stop_ret            ret
'
'
' write a data bit based on carry flag and clock it out, assume clk is low on entry, on exit clk will be low
i2c_WriteBit            muxnc   dira, SdaPinMsk         ' set i2c data pin high or low based on carry bit
                        call    #HalfBitDelay
                        call    #ClkHigh
                        call    #HalfBitDelay            ' need to hold data for a full clock state period
                        call    #ClkLow
i2c_WriteBit_ret        ret
'
'
' Send 8 bits (LSB) of i2c_data MSb first, assume clock is low on entry, clock is low on exit
' ACK bit is returned in carry flag, Data must be saved into i2c_data variable before calling this function. 
i2c_data      long      0
'
i2c_WrByte              mov     i2c_bitCnt, #8          ' we have 8 bits to send
                        shl     i2c_data, #24           ' shift data into MSB (and clear LS bits)
                        ' clock out next MSb data bit first
:i2c_WrByteNxt          shl     i2c_data, #1    wc
                        call    #i2c_WriteBit
                        djnz    i2c_bitCnt, #:i2c_WrByteNxt
                        'need to read ack bit, clock is low
                        call    #i2c_ReadBit
i2c_WrByte_ret          ret    
'
'
' read next data bit and save to carry flag, assume clk is low on entry, on exit clk will be low
i2c_ReadBit             call    #SdaHigh                ' set data pin to input mode
                        call    #ClkHigh                ' clock out next data bit             
                        test    SdaPinMsk, ina     wc   ' save data to carry flag
                        call    #HalfBitDelay           ' need to hold data for a full clock state period
                        call    #ClkLow                 ' set clock low
i2c_ReadBit_ret         ret
'
'
' Read 8 bits into (LSB) of i2c_data variable MSb first, assume clock is low on entry, clock is low on exit
i2cLastByte             long    0
'
'
i2c_RdLastByte          mov     i2cLastByte, #1
                        jmp     #i2c_DoRdByte
'                        
i2c_RdByte              mov     i2cLastByte, #0
i2c_DoRdByte            mov     i2c_bitCnt, #8          ' we have 8 bits to read
                        mov     i2c_data, #0
                        ' clock out next MSb data bit first
:i2c_RdByteNxt          call    #i2c_ReadBit
                        rcl     i2c_data, #1
                        djnz    i2c_bitCnt, #:i2c_RdByteNxt
                        'need to write ack bit, clock is low
                        cmp     i2cLastByte, #1         wz
              if_nz     call    #SdaLow                 ' set data pin low
              if_z      call    #SdaHigh                ' set data pin high
                        call    #ClkHigh                ' set clock high
                        call    #HalfBitDelay           '
                        call    #ClkLow                                  
i2c_RdByte_ret
i2c_RdLastByte_ret      ret    
'
'
' Read from 1 to 16 Si4703 16bit registers and save them to a global memory array of words.

' Before calling this function:
'       1) The variable SiRegCount is set to the number of 16 bit registers to read from the Si4703 - 1.
'       2) The destination buffer address is copied to the variable SiRegBuffAdr (buffer must hold 16 words). 

' Valid values for SiRegCount are (0..15).
' Each Si4703 register is read and saved to the global memory starting from register address 0Ah.
' The value from register 0Ah is saved to buffer address = (SiRegBuffAdr + (0Ah * size of word)).

' For the number set in SiRegCount each succesive register is read and saved to the next global
' memory word. After register 0Fh is read from the Si4703 and saved to the equivilent global memory
' address the next register read is 00h and will be save to the first global memory block word. 

' The buffer address is assumed to point to the address of register 0.
'
SiRegCount              long    0               ' number of registers to read
SiRegBuffAdr            long    0               ' global memory address to copy registers to
SiCopyRes               long    0               ' set to indicate if operation succeded on exit
'
'                       ' first register is register index 0A
RdSiRegs                mov     regIndex, #$0A
                        and     SiRegCount, #$0F        ' (set range from 0..F for 1..16 registers)
                        add     SiRegCount, #1          ' inc register count to use 1..16 range
                        
                        ' send i2c start condition
                        call    #i2c_start
                        ' send read register request                              
                        mov     i2c_data, #c_Si4703_ReadCmd
                        call    #i2c_WrByte
       if_c             jmp     #:RdSiError           
                        ' if write cmd was acked read high byte of 16 bit register value
                        
:RdNxtReg               call    #i2c_RdByte
                        mov     regData, i2c_data
                        shl     regData, #8
                        ' now read low 8 bits
                        cmp     SiRegCount, #1  wz
              if_nz     call    #i2c_RdByte
              if_z      call    #i2c_RdLastByte
                        or      regData, i2c_data

                        ' we have the register value, save to global memory address
                        mov     temp1, regIndex
                        shl     temp1, #1
                        add     temp1, SiRegBuffAdr
                        wrword  regData, temp1 

                        ' data saved pont to next register
                        add     regIndex, #1
                        cmpsub  regIndex, #$10     

                        ' see if we have more registers to read
                        djnz    SiRegCount, #:RdNxtReg
                        ' all done, indicate read register(s) was completed
                        mov     SiCopyRes, ForthTrueValue
                        jmp     #:RdSiDone
                        ' indicate we could not read registers
:RdSiError              mov     SiCopyRes, ForthFalseValue
                        ' send a stop condition
:RdSiDone               call    #i2c_stop
RdSiRegs_ret            ret
'
'
' Write from 1 to 14 Si4703 16bit registers from data passed in a global memory array of 16 words.
' The first 2 resister are read-only, wirting starts from register 02h.

' Before calling this function:
'       1) The variable SiRegCount is set to the number of 16 bit registers to write to the Si4703 - 1.
'       2) The source buffer address is copied to the variable SiRegBuffAdr (buffer must hold 16 words). 

' Valid values for SiRegCount are (0..15).
' Each Si4703 register is written with the global memory starting from register address 02h.
' The value from register 02h is written with the value in buffer address = (SiRegBuffAdr + (02h * size of word)).

' For the number set in SiRegCount each succesive register is wriiten with the next global
' memory word. After register 0Fh is written with the equivilent global memory
' address value the next register write is 00h. 

' The buffer address is assumed to point to the address of register 0.
'
WtSiRegs                mov     regIndex, #$02
                        and     SiRegCount, #$0F        ' (set range from 0..F for 1..16 registers)
                        add     SiRegCount, #1          ' inc register count to use 1..16 range
                        
                        ' send i2c start condition
                        call    #i2c_start
                        ' send write register request                              
                        mov     i2c_data, #c_Si4703_WriteCmd
                        call    #i2c_WrByte
       if_c             jmp     #:WtSiError           
                        ' we have the register value, read global memory address
:WtNxtReg               mov     temp1, regIndex
                        shl     temp1, #1
                        add     temp1, SiRegBuffAdr
                        rdword  regData, temp1 
                        ' write register, high byte first
                        mov     i2c_data, regData
                        shr     i2c_data, #8
                        call    #i2c_WrByte
       if_c             jmp     #:WtSiError           
                        ' write register low byte next
                        mov     i2c_data, regData
                        and     i2c_data, #255
                        call    #i2c_WrByte
       if_c             jmp     #:WtSiError           
                        ' register was written, see if there are more
                        add     regIndex, #1
                        cmpsub  regIndex, #$10     
                        ' see if we have more registers to read
                        djnz    SiRegCount, #:WtNxtReg
                        ' all done indicate copy was ok
                        mov     SiCopyRes, ForthTrueValue
                        jmp     #:WtSiDone
                        ' indicate we could not write registers
:WtSiError              mov     SiCopyRes, ForthFalseValue
                        ' send a stop condition
:WtSiDone               call    #i2c_stop
WtSiRegs_ret            ret
'
'
' select the radio device we are connected to
' valid device numbers for this project are 0..3
                        'see if selected device index is in range (0..3)
SelDevice               cmp     DeviceIdx, #4           wc
              if_nc     jmp     #SelDevice_ret
                        'calculate pin offset from device 0
                        mov     temp1, DeviceIdx
                        shl     temp1, #2               ' mult index by 4 to selct device pin shift
                        ' temp1 = device index * 4 (4 pins per device)
                        'set serial data pin mask   
                        mov     SdaPinMsk, BaseSdaPinMsk
                        shl     SdaPinMsk, temp1
                        'set seial clock pin mask
                        mov     ClkPinMsk, BaseClkPinMsk
                        shl     ClkPinMsk, temp1
                        'calculate device reset pin mask
                        mov     ResetPinMsk, BaseResetPinMsk
                        shl     ResetPinMsk, temp1
SelDevice_ret           ret

' COMMAND FUNCTIONS
'
' reset the Si4703 IC, si_reset ( -- )
'
Cmd_Reset               andn    outa, ResetPinMsk
                        or      dira, ResetPinMsk
                        ' reset is low, need to pull SDIO low to force 2 wire mode (I2C)
                        andn    outa, SdaPinMsk
                        or      dira, SdaPinMsk
                        ' reset and sdio are low, hold reset for at least 100usec
                        mov     temp1, #$1FF
:restWait               call    #HalfBitDelay
                        djnz    temp1, #:restWait
                        ' set reset high
                        or      outa, ResetPinMsk
                        mov     temp1, #50
:restWait2              call    #HalfBitDelay
                        djnz    temp1, #:restWait2
                        ' set sdio and clock to allow i2c comunications
                        call    #i2c_Init
                        call    #HalfBitDelay
Cmd_Reset_ret           ret
'
'
' read Si4703 registers  si_read( adrDestArray RegCnt -- fRead)
'
Cmd_RdRegisters         call    #F_PopStack
                        mov     SiRegCount, ForthTmp
                        call    #F_PopStack
                        mov     SiRegBuffAdr, ForthTmp
                        call    #RdSiRegs
                        ' return result
                        mov     ForthTmp, SiCopyRes
                        call    #F_PushStack
Cmd_RdRegisters_ret     ret
'
'
' write Si4703 registers  si_write( adrSourceArray RegCnt -- fWritten )
'
Cmd_WtRegisters         call    #F_PopStack
                        mov     SiRegCount, ForthTmp
                        call    #F_PopStack
                        mov     SiRegBuffAdr, ForthTmp
                        call    #WtSiRegs
                        ' return result
                        mov     ForthTmp, SiCopyRes
                        call    #F_PushStack
Cmd_WtRegisters_ret     ret
'
'
' select the device we are connected to  si_device! ( uDeviceIdx -- )
                        'see if selected device index is in range (0..3)
Cmd_SetDeviceIdx        call    #F_PopStack
                        cmp     ForthTmp, #4           wc
              if_nc     jmp     #Cmd_SetDeviceIdx_ret
                        ' set variable to track device index
                        mov     DeviceIdx, ForthTmp
                        ' set pin masks to select correct device
                        call    #SelDevice
Cmd_SetDeviceIdx_ret    ret
'
'
' read index of selected device  si_device@ ( -- uDeviceIdx )
Cmd_GetDeviceIdx        mov     ForthTmp, DeviceIdx
                        call    #F_PushStack
Cmd_GetDeviceIdx_ret    ret
'
'
'Base device pin masks
BaseSdaPinMsk               long    1 << SDA_PIN
BaseClkPinMsk               long    1 << CLK_PIN
BaseResetPinMsk             long    1 << RST_PIN
DeviceIdx                   long    0
'Currently selected device pin masks, at start up device 0 is selecetd
SdaPinMsk               long    1 << SDA_PIN
ClkPinMsk               long    1 << CLK_PIN
ResetPinMsk             long    1 << RST_PIN

ForthTrueValue          long    $FFFF
ForthFalseValue         long    0
'
'I2C Variables
I2C_Bit_Delay           long    (80_000_000 / 400_000)  {2.5us}
i2c_bitCnt              long    0
'
'General use variables
regIndex                res     1
regData                 res     1
temp1                   res     1

