\ LCD	PC8574
\ ---	------
\ R/S	P0
\ R/W	P1
\ E	P2
\ D0	
\ D1
\ D2
\ D3
\ D4	P4
\ D5	P5
\ D6	P6
\ D7	P7
\ A  (connected to GND? pin of two pin header other side is VCC)
\ K


#22 CONSTANT i2c_clk_pin 
#21 CONSTANT i2c_data_pin

\ set clock line high  
: i2c-clk1 ( -- )
  i2c_clk_pin dira-clrpin ;
  
\ set clock line low  
: i2c-clk0 ( -- )
  i2c_clk_pin dira-setpin ;
  
\ set data line high  
: i2c-data1 ( -- )
  i2c_data_pin dira-clrpin ;
  
\ set data line low  
: i2c-data0 ( -- )
  i2c_data_pin dira-setpin ;
    
\ send an I2C start bit  
: i2c-sendStart ( -- )
  \ set the clock line low
  i2c-clk0
  \ set data line high
  i2c-data1
  \ bring the clock line high 
  i2c-clk1
  \ set the data line low
  i2c-data0 ;
  
\ send an I2C stop bit  
: i2c-sendStop ( -- )
  \ set the clock line low
  i2c-clk0
  \ set data line low
  i2c-data0
  \ bring the clock line high 
  i2c-clk1
  \ set the data line high
  i2c-data1 ;
  
\ send 8 data bits to 8574 via I2C, MSb first  
: i2c-sendByte ( cByte -- fAck )
  $80 
  #8 0 DO
    \ set the clock line low
    i2c-clk0
    2DUP 
    ( cByte cMask cByte cMask )
    AND IF
      \ set data high
      i2c-data1
    ELSE
      \ set data low
      i2c-data0
    THEN
    \ bring the clock line high 
    i2c-clk1
    1 RSHIFT
    ( cByte cMask' )
  LOOP
  2DROP
  \ last read the ack bit from the 8574
  \ set the clock line low
  i2c-clk0
  \ set data pin as input (set data pin high)
  i2c-data1
  \ bring the clock line high 
  i2c-clk1
  \ read data pin to get device ack bit
  i2c_data_pin ina-pin 0= ;
  
\ : reg-wrt ( cData -- fOk )
\   i2c-sendStart
\   $4E i2c-sendByte IF
\     i2c-sendByte
\   ELSE
\     DROP
\     FALSE
\   THEN  
\   i2c-sendStop ;

\ set data and clock pins high, and ready for open drain drive
: i2c-init ( -- )
  \ ready pins to use DIRA to control output levels
  i2c_clk_pin outa-clrpin
  i2c_data_pin outa-clrpin
  \ set data and clock pins high
  i2c-data1
  i2c-clk1 ;

: lcd-regwrtnib ( cNibble -- fOK )
  i2c-sendStart
  $4E i2c-sendByte IF
    \ $0C = RS=0 RW=0 E=1
    #4 LSHIFT $0C OR DUP
    ( cNibble'  cNibble' )
    i2c-sendByte
    ( cNibble'  fOK )
    SWAP $FB AND
    \ RS=0 RW=0 E=0
    i2c-sendByte AND
  ELSE
    DROP FALSE
  THEN 
  i2c-sendStop ;    
  
: lcd-regwrt ( cbyte -- fOK )
  DUP #4 RSHIFT lcd-regwrtnib
  SWAP $0F AND lcd-regwrtnib
  AND ; 
  
: lcd-wrtnib ( cNibble -- fOK )
   \ $0C = RS=0 RW=0 E=1
   #4 LSHIFT $0D OR DUP
   ( cNibble'  cNibble' )
   i2c-sendByte
   ( cNibble'  fOK )
   SWAP $FB AND
   \ RS=0 RW=0 E=0
   i2c-sendByte AND ;
  
: lcd-wrt ( cbyte -- fOK )
  i2c-sendStart
  $4E i2c-sendByte IF
    DUP #4 RSHIFT lcd-wrtnib
    SWAP $0F AND lcd-wrtnib
    AND DROP
  ELSE
    DROP FALSE
  THEN 
  i2c-sendStop ;    
  
\ Initalize the LCD
\ Setup LCD controller with correct settings
: lcd-init ( -- )
  i2c-init
  #15 MS
  $3 lcd-regwrtnib DROP
  #5 MS
  $3 lcd-regwrtnib DROP
  #1 MS  
  $32 lcd-regwrt DROP
  $28 lcd-regwrt DROP
  $01 lcd-regwrt DROP
  $06 lcd-regwrt DROP
  \ LCD is reset, now configure LCD
  $01 lcd-regwrt DROP
  $0F lcd-regwrt DROP ;

: lcd-page ( -- )
  $01 lcd-regwrt DROP 2 MS ;
    
: lcd-type ( c-adr uCount -- )
  i2c-sendStart
  $4E i2c-sendByte IF
    0 ?DO 
      COUNT 
      DUP #4 RSHIFT lcd-wrtnib
      SWAP $0F AND lcd-wrtnib
      AND DROP
    LOOP 
    DROP  
  ELSE
    2DROP FALSE
  THEN 
  i2c-sendStop ;    

: lcd-space ( -- )
  BL lcd-wrt ;
  
: lcd-words ( -- )
  LATEST
  BEGIN
    @ DUP
  WHILE
    DUP COUNT lcd-type lcd-space
    NFA>LFA
  REPEAT ;    

: words ( -- )
  LATEST
  BEGIN
    @ DUP
  WHILE
    DUP COUNT TYPE SPACE
    NFA>LFA
  REPEAT ;    
  
\ Base line addresses for 20x4 LCD
VARIABLE lcd_ofs #4 CHARS ALLOT
0 lcd_ofs !           ( line 0 addresses 0..$13)
$40 lcd_ofs 1+ C!     ( line 1 addresses $40..$53)
$14 lcd_ofs 2+ C!     ( line 2 addresses $14..$27)
$54 lcd_ofs #3 + C!   ( line 3 addresses $54..$74)

 
\ set addrtess to specific x and y (0 based, Y - 0..3, X - 0..19)  
: lcd-xy ( cChar uX uY -- )
  \ set address
  #3 MIN lcd_ofs + C@
  SWAP #19 MIN + 
  $80 OR
  lcd-regwrt DROP ;
  
\ display char at specific x and y address (0 based, Y - 0..3, X - 0..19)  
: lcd-out ( cChar uX uY -- )
  lcd-xy
  lcd-wrt ; 
  
: lcd-cursor ( fShow -- )
  IF
    $0F
  ELSE
    $0C
  THEN
  lcd-regwrt DROP ;     

: lcd>con 
  ['] lcd-page IS PAGE
  ['] lcd-wrt IS EMIT
  ['] TRUE IS EMIT? ;

