'' Example 20x4 LCD alphanumeric display with PC8574 interface

const
  SDA_PIN       = 1
  CLK_PIN       = 0
  c_ForthCmdCount = 2

pub Start
  
dat

Entry                   jmp     #F_Do_Cmds

'*********************************[ FORTH system command support routines]*********************************
'
' Variables used by forth SYSTEMEX interface
ForthTmp                long    0
ForthCmd                long    0
ForthStackPtr           long    0
ForthScratch            long    0   
' Constants
ForthWordMask           long    $FFFF


' Remove top 16 bit value from forth stack.ForthTmp = 16 bit value on return
                        'make sure we are in a command and stack pointer is valid
F_PopStack              tjz     ForthStackPtr, #F_PopStack_Ret
                        'We are in a command OK to pop value from stack
                        rdword  ForthTmp, ForthStackPtr
                        add     ForthStackPtr, #2
F_PopStack_Ret          ret             
'
' Add 16 bit value to top of forth stack.ForthTmp = 16 bit value to write.
                        'make sure we are in a command and stack pointer is valid
F_PushStack             tjz     ForthStackPtr, #F_PushStack_Ret
                        'We are in a command OK to push a value to the stack
                        sub     ForthStackPtr, #2
                        wrword  ForthTmp, ForthStackPtr
F_PushStack_Ret         ret             
'
'
'Wait for a system command to be requested by forth engine
F_WaitForCmd            rdlong  ForthCmd, par
                        test    ForthCmd, ForthWordMask  wz
        if_z            jmp     #F_WaitForCmd
                        'Seperate stack address from command number
                        mov     ForthStackPtr, ForthCmd
                        shr     ForthStackPtr, #16
                        and     ForthCmd, ForthWordMask
                        'see if we have a valid command
                        cmp     ForthCmd, #c_ForthCmdCount                      wc
        if_c            jmp     #F_WaitForCmd_ret
                        ' we have an invalid command
                        call    #F_CompleteCommand                        
                        jmp     #F_WaitForCmd
                        ' we have a valid command, get the stack pointer address
F_WaitForCmd_Ret        ret

'

'Update stack pointer then exit
F_SaveStackPtrExit      shl     ForthStackPtr, #16
                        wrlong  ForthStackPtr, PAR
                        mov     ForthStackPtr, #0
F_SaveStackPtrExit_ret  ret
'
'
'Indicate system command requested by forth engine is completed, invalid command
F_CompleteCommand       mov     ForthCmd, #0                          
                        wrword  ForthCmd, par
                        mov     ForthStackPtr, #0
F_CompleteCommand_ret   ret

'
'Wait for a command and execute it
F_Do_Cmds               call    #F_WaitForCmd
                        'we have a cmd index 1..XX
                        mov     ForthTmp, #F_JUMP_TABLE
                        add     ForthTmp, ForthCmd
                        movs    :F_Set_Cmd_Jump, ForthTmp
                        nop
:F_Set_Cmd_Jump         mov     :F_Do_Cmd_Jump, (ForthTmp)
                        nop
:F_Do_Cmd_Jump          nop
                        'The called function will return here
F_Do_Cmds_Return        call    #F_SaveStackPtrExit
                        jmp     #F_Do_Cmds
'
'   
'Table of Jump opcodes for each of the supported forth functions
F_JUMP_TABLE            nop                              '(0)
                        call    #i2c_Init                '(1)
                        call    #WrPC8574Reg             '(2)
'                        call    #F_CMD_WtOuta           '(3)
'                        call    #F_CMD_RdDira           '(4)
'                        call    #F_CMD_WtDira           '(5)
'                        call    #F_CMD_RdInaPin         '(6)
'                        call    #F_CMD_RdOutaPin        '(7)
'                        call    #F_CMD_RdDiraPin        '(8)
'                        call    #F_CMD_SetOutaPin       '(9)
'                        call    #F_CMD_SetDiraPin       '(10)


' *********************************[ Forth Commands ]*********************************

' Initialize the I2C interface, on exit SCLK and SDA are high
i2c_Init                andn    outa, SdaPinMsk         ' ready i2c data pin for open drain output
                        andn    outa, ClkPinMsk         ' ready i2c clk pin for open drain output
                        ' next two opcodes are not needed unless we call Entry again
                        andn    dira, SdaPinMsk         ' set i2c data pin high
                        andn    dira, ClkPinMsk         ' set i2c clk pin high
                        ' The i2c data and clock pins are now high
i2c_Init_ret            ret 

' Delay for one half an i2c bit time, for si470x on my LCD this is 1.25us (400khz max clock rate / 2)
Delay_Temp    long      0
'
' For PC8574 on my LCD max i2c clock rate is 100Khz.
' This makes the min period for a clock high or low 5us (10us / 2).
'
' Delay for one half of i2c clock bit time or 2.5us (10us / 4)
HalfBitDelay            mov     Delay_Temp, I2C_Bit_Delay
                        shr     Delay_Temp, #1          ' half i2c bit delay clock time
                        ' we have five opcodes + call into this + return = (5 * 4 + 4 + 4) = 24 cycles
                        sub     Delay_Temp, #24
                        add     Delay_Temp, cnt
                        waitcnt Delay_Temp, #0
HalfBitDelay_ret        ret                                                                              
'
'
' Set the i2c clk line high, delay half a min clock period
ClkHigh                 andn    dira, ClkPinMsk         ' set i2c clk pin high
                        call    #HalfBitDelay
ClkHigh_ret             ret
'
' 
' Set the i2c clk line low, delay half a min clock period
ClkLow                  or      dira, ClkPinMsk         ' set i2c clk pin low
                        call    #HalfBitDelay
ClkLow_ret              ret
'
' 
' Set the i2c data line high, delay half a min clock period
SdaHigh                 andn    dira, SdaPinMsk         ' set i2c data pin high
                        call    #HalfBitDelay
SdaHigh_ret             ret
'
' 
' Set the i2c clk line low, delay half a min clock period
SdaLow                  or      dira, SdaPinMsk         ' set i2c data pin low
                        call    #HalfBitDelay
SdaLow_ret              ret
'
' 
' Send an i2c start bit, SCLK will be set Low on exit, ausume SCLK and SDAT high on entry 
i2c_start               call    #SdaLow
                        call    #ClkLow
i2c_start_ret           ret
'
'
' write a data bit based on carry flag and clock it out, assume clk is low on entry, on exit clk will be low
i2c_WriteBit            muxnc   dira, SdaPinMsk         ' set i2c data pin high or low based on carry bit
                        call    #HalfBitDelay
                        call    #ClkHigh
                        call    #HalfBitDelay            ' need to hold data for a full clock state period
                        call    #ClkLow
i2c_WriteBit_ret        ret
'
'
' Send an i2c stop bit, assume clk is low and data is undetermined, both data and clk will be set high.
i2c_stop                call    #SdaLow                 ' set i2c data pin low
                        call    #ClkHigh                ' set i2c clock pin high
                        call    #SdaHigh                ' set i2c data pin High
i2c_stop_ret            ret
'
'
' Send 8 bits (LSB) of i2c_data MSb first, assume clock is low on entry, clock is low on exit 
i2c_data      long      0
i2c_WrByte              mov     i2c_bitCnt, #8          ' we have 8 bits to send
                        shl     i2c_data, #24           ' shift data into MSB (and clear LS bits)
                        ' clock out next MSb data bit first
:i2c_WrByteNxt          shl     i2c_data, #1    wc
                        call    #i2c_WriteBit
                        djnz    i2c_bitCnt, #:i2c_WrByteNxt
                        'need to read ack bit, clock is low
                        call    #SdaHigh                ' set data pin to input mode
                        test    SdaPinMsk, ina  wc      ' read the ack bit and save to carry bit
                        call    #HalfBitDelay
                        call    #ClkLow
i2c_WrByte_ret          ret    
'
'
' Write a byte to PC8574 data register.
' To make a pin an input write the pin as a high.
' WrPC8574Reg ( c-data -- fOK)
WrPC8574Reg             call    #i2c_start
                        {send PC8574 address}
                        mov     i2c_data, #$4E
                        call    #i2c_WrByte
                        call    #F_PopStack    
                        mov     i2c_data, ForthTmp
WrPC8574Reg_ret         ret       

'
  
'Hardware constants
SdaPinMsk               long      1 << SDA_PIN
ClkPinMsk               long      1 << CLK_PIN
'Variables
I2C_Bit_Delay           long      (80_000_000 / 3200)  {2.5us}
i2c_bitCnt              long      0
         