( PS2 keyboard driver )

#43 VALUE kbd_blk

0 VALUE kbd_cog

\ keyboard driver requires 19 longs for params
#19 #2 LSHIFT QBYTES kbd_params

\ Define some word to make it simple to read params

: kbd-tail@ ( -- uTail )
  kbd_params @ ;

: kbd-tail! ( uNewTail -- )
  kbd_params ! ;

: kbd-head@ ( -- uHead )
  kbd_params #4 + @ ;

: kbd-found@ ( -- uFound )
  kbd_params #8 + @ ;

\ read from keyboard fifo

\   locks = lock setup
\           bit 6 disallows shift-alphas (case set soley by CapsLock)
\           bits 5..3 disallow toggle of NumLock/CapsLock/ScrollLock state
\           bits 2..0 specify initial state of NumLock/CapsLock/ScrollLock
\           (eg. %0_001_100 = disallow ScrollLock, NumLock initially 'on')
\
\   auto  = auto-repeat setup
\           bits 6..5 specify delay (0=.25s, 1=.5s, 2=.75s, 3=1s)
\           bits 4..0 specify repeat rate (0=30cps..31=2cps)
\           (eg %01_00000 = .5s delay, 30cps repeat)


: kbd-clr ( -- )
  kbd-head@ kbd-tail! ;

: EKEY? ( -- f )
  kbd-head@ kbd-tail@ <> ;

: kbd-buffer ( -- uDataAdr )
  \ calc start of data buffer
  kbd_params #11 #2 LSHIFT + ;

\ return the current key in buffer, does not inc tail
: kbd-peek ( -- u )
  \ get address of keyboard buffer
  kbd-buffer
  \ add tail offset, 16 bits per key data
  kbd-tail@ 2* + @ ;

: EKEY ( -- u )
  BEGIN
    EKEY?
  UNTIL
  kbd-peek
  \ increment the tail, check for fifo wrap
  kbd-tail@ 1+ $F AND kbd-tail! ;

: EKEY>CHAR ( uEKey -- cChar TRUE | uEKey FASLE )
  DUP #8 RSHIFT
  \ See if we have a normal key
  DUP IF
    ( uKey uKeyState )
    \ not a normal key see if it is a control character
    \ 1 = shift, 2 = ctrl, 4 = alt, 8 = win
    #2 = IF
      ( uKey )
      \ we have a control key
      \ if key is lowercase convert to upper case
      DUP DUP $FF AND [CHAR] z 1+ [CHAR] a WITHIN IF
        ( uKey uKey )
        $DF AND
      ELSE
        $FF AND
      THEN
      ( uKey uKey' )
      DUP #96 #64 WITHIN IF
        ( uKey uKey' )
        \ we have a valid control key
        NIP $40 - $1F AND TRUE
      ELSE
        \ not a valid control key
        DROP FALSE
      THEN      
    ELSE
      \ Special key not a character
      FALSE
    THEN
  ELSE
    \ key state = 0, no modifiers pressed
    ( uKey uKeyState )
    \ a normal character key just return
    \ see if it is a function key
    DROP DUP #128 < IF
      ( uKey )
      TRUE
    ELSE
      \ see if it is back space
      DUP $C8 = IF
        ( uKey )
        \ we have a back space key
        DROP #8 TRUE
      ELSE
        \ not a valid key
        FALSE
      THEN
    THEN
  THEN ;

: kbd-key? ( -- f )
  FALSE
  BEGIN
    EKEY?
  WHILE
    kbd-peek EKEY>CHAR IF
      2DROP TRUE EXIT
    ELSE
      \ not a valid character, remove it
      EKEY 2DROP
    THEN
  REPEAT ;

: kbd-key ( -- c )
  BEGIN
    kbd-key?
  UNTIL
  EKEY EKEY>CHAR DROP ;

: kbd>con ( -- )
  ['] kbd-key? IS KEY?
  ['] kbd-key IS KEY
  ['] kbd-clr IS EMPTY-KEYS ;

MARKER KEYBOARD

: kbd-init ( uClkPin uDataPin -- )
  \ set all paramater values to zero
  kbd_params #19 #2 LSHIFT ERASE
  \ set the first 4 longs of data buffer with start up params
  \ Save data pin setting
  S>D kbd-buffer Q!
  \ Save clock pin setting
  S>D kbd-buffer #4 + Q!
  \ Save key lock setting
  %0000100 S>D kbd-buffer #8 + Q!
  \ Save repeat rate setting
  %0100010 S>D kbd-buffer #12 + Q!
  \ Start up params are set start COG
  kbd_params kbd_blk COG-LOAD TO kbd_cog
  KEYBOARD ;

\ #27 #26 kbd-init  