MARKER unload-crc7

\ Calculate CRC7 for SD/MMC cards

$09 CONSTANT crc7_poly


: crc7-encode ( uCrcReg uByte uBits -- uCrcReg' )
  \ msb first of uBbyte for uBits elements
  0 ?DO
    ( uCrcReg uByte )
    \ Shift seven bit register left and put in MSB value of input to LSB.
    SWAP 1 LSHIFT
    ( uByte uCrcReg )
    OVER $80 AND IF
      1 OR
    THEN
    \ Test D7 of shift register as MSB of byte, test if 1 and therefore XOR.
    DUP $80 AND IF
      crc7_poly XOR
    THEN
    SWAP
    ( uCrcReg uByte )
    \ Shift data left one bit to ready next bit
    1 LSHIFT
  LOOP
  ( uCrcReg uByte )
  DROP $7F AND ;


: CRC7 ( uDataAdr uByteCnt -- uCrc7 )
  \ Set result to 0 and loop for uByteCnt bytes
  0 SWAP
  0 ?DO
    ( uDataAdr uCrc7 )
    OVER C@
    ( uDataAdr uCrc7 uByte )
    #8 crc7-encode
    ( uDataAdr uCrc7 )
    \ point to next data byte
    SWAP 1+ SWAP
  LOOP
  ( uDataAdr uCrc7 )
  NIP
  ( uCrc7 )
  \ mix in last 7 bits of 0s to augment, and then shift final CRC
  \ remainder left and OR in fixed 1 LSB.
  0 #7 crc7-encode
  ( uCrc7 )
  1 LSHIFT 1 OR ;  

\ Buffer to hold 5 bytes for SD card command crc7 calc
VARIABLE crc7_cmdbuf #4 ALLOT

\ calc crc7 for SD card command. dData is 32 bit value uCmd is a byte
: CRC7-CMD ( dData uCmd -- uCmdCrc7 dData uCmd )
  \ Save command value to buffer as a byte
  DUP crc7_cmdbuf C!
  \ save command value on return stack
  >R
  ( dData ) ( R: uCmd )
  \ Save data to buffer as 4 bytes MSB first
  2DUP 2>R
  ( dData ) ( R: uCmd dData )
  #5 1 DO
    \ get MSB byte
    2DUP $FF000000. DAND NIP 8 RSHIFT
    ( dData uByte )
    crc7_cmdbuf I + C!
    ( dData )
    \ ready next byte
    #8 DLSHIFT
  LOOP
  2DROP
  \ data is in buffer Calc CRC7
  crc7_cmdbuf 5 CRC7
  ( uCmdCrc7 ) ( R: uCmd dData )
  2R> R>
  ( uCmdCrc7 dData uCmd )
  ;
  
  