\ program to run DS1620 digital thermometer
\ requires COG Pins forth code to control Propeller pins

#0 CONSTANT dt_pin_dq
#1 CONSTANT dt_pin_clk
#2 CONSTANT dt_pin_rst

: dt-init ( -- )
  dt_pin_clk outa-setpin        \ set clock pin high
  dt_pin_rst outa-clrpin        \ set reset pin low
  dt_pin_dq dira-clrpin         \ set data pin as input
  dt_pin_clk dira-setpin        \ set clock pin as output
  dt_pin_rst dira-setpin ;      \ set reset pin as output

  \ write command 8 bits (LSB first)
: (dt-cmd) ( uCmd -- )
  dt_pin_rst outa-clrpin        \ set reset pin low
  dt_pin_clk outa-setpin        \ set clock pin high
  dt_pin_rst outa-setpin        \ set reset pin high
  dt_pin_dq dira-setpin        \ set data pin as output
  #8 0 DO
    dt_pin_clk outa-clrpin      \ set clock pin low
    DUP 1 AND IF
      dt_pin_dq outa-setpin    \ set data pin high
    ELSE
      dt_pin_dq outa-clrpin    \ set data pin low
    THEN
    \ data bit set
    dt_pin_clk outa-setpin      \ set clock pin high
    1 RSHIFT
  LOOP
  DROP
  \ Clock, Reset pins are high, set data to input
  dt_pin_dq dira-clrpin ;        \ set data pin as input

\ read up to 16 bits from DS1620  
: (dt-read) ( uBits uCmd -- uValue )
  (dt-cmd)
  \ command written now do a read of requested bits
  \ Clock, Reset pins are high, data set to input
  ( uBits )
  DUP
  ( uBits uBits )
  0                             \ init data read value
  ( uBits uBits uValue )
  SWAP
  ( uBits uValue uBits )
  0 ?DO
    dt_pin_clk outa-clrpin      \ set clock pin low
    1 RSHIFT
    dt_pin_dq ina-pin IF
      $8000
    ELSE
      0
    THEN
    dt_pin_clk outa-setpin      \ set clock pin high
    ( uValue uBit )  
    OR
  LOOP
  dt_pin_rst outa-clrpin        \ set reset pin low
  ( uBits uValue )
  SWAP   
  ( uValue uBits )
  #16 SWAP - RSHIFT ; 


\ write up to 16 bits to DS1620  
: (dt-write) ( uValue uBits uCmd -- )
  (dt-cmd)
  \ command written now do a read of requested bits
  \ Clock, Reset pins are high, set data to output
  dt_pin_dq dira-setpin          \ set data pin as output
  ( uValue uBits )
  0 ?DO
    ( uValue )
    dt_pin_clk outa-clrpin      \ set clock pin low
    DUP 1 AND IF
      dt_pin_dq outa-setpin    \ set data pin high
    ELSE
      dt_pin_dq outa-clrpin    \ set data pin low
    THEN
    \ data bit set
    dt_pin_clk outa-setpin      \ set clock pin high
    1 RSHIFT
    ( uValue' )
  LOOP
 ( uValue' )
  DROP
  dt_pin_rst outa-clrpin         \ set reset pin low
  dt_pin_dq dira-clrpin ;        \ set data pin as input

\ read the 9 bit temp register
: dt-temp@ ( -- uTempReg )
  #9 $AA (dt-read) ;
  
\ read the 9 bit count register
: dt-count@ ( -- uCountReg )
  #9 $A0 (dt-read) ;
  
\ read the 9 bit slope register
: dt-slope@ ( -- uSlopeReg )
  #9 $A9 (dt-read) ;
  
\ read the 8 bit config register
: dt-cfg@ ( -- uConfigReg )
  #8 $AC (dt-read) ;

\ write the 8 bit config register
: dt-cfg! ( uValue -- )
  #8 $0C (dt-write) ;
  
\ write the 9 bit high temp alert value
: dt-th! ( uTH_Value -- )
  #9 $01 (dt-write) ;

\ read the 9 bit high temp alert value
: dt-th@ ( -- uTH_Value )
  #9 $A1 (dt-read) ;

\ write the 9 bit low temp alert value
: dt-tl! ( uTL_Value -- )
  #9 $02 (dt-write) ;

\ read the 9 bit low temp alert value
: dt-tl@ ( -- uTL_Value )
  #9 $A2 (dt-read) ;

\ Start temp conversion
: dt-convert ( -- )
  $EE (dt-cmd) ;

\ Stop temp conversion
: dt-stop ( -- )
  $22 (dt-cmd) ;

: temp
  dt-convert
  BEGIN
    dt-cfg@ #128 AND
  UNTIL
  dt-temp@ 
  DUP #256 AND IF
    \ convert 9 bit negative to a 16 bit negative number
    $FF00 OR 
  THEN 2/ ;
  
: temp+
  temp #100 *
  ( u100Temp )
  dt-slope@ DUP dt-count@
  ( u100Temp uSlope uSlope uCount )
   - SWAP
  ( u100Temp uRemain uSlope )
  #100 SWAP
  ( u100Temp uRemain u100, uSlope )
  */ 
  ( u100Temp uDecTemp )
  #25 - +
  \ now round to single decimal point
  DUP 0< IF
    -#5
  ELSE
    #5
  THEN
  + #10 /   
  . ;
  

: dt-test ( -- )
  BEGIN
    KEY? NOT
  WHILE
    temp+
  REPEAT ;      
  
      


