\ functions to load execute and unload forth words
\ such as edit, load-blocks.

\ execute the function by name
\ the function is loaded from block storage
\ executed then unloaded (no COGS allowed)
: exec-blocks ( cadrName uFirstBlock uLastBlock )
  \ save dictionary pointers to return stack
  LATEST @ HERE 2>R
  \ load the code
  (  cadrName uFirstBlock uLastBlock  R: uLatest@ uHere )
  2DUP = IF DROP LOAD ELSE THRU THEN
  \ code is loaded now find the word
  ( cadrName  R: uLatest@ uHere )
  DUP FIND DUP IF
    ( cadrName uFindAddress fWasFound  R: uHere uLatest@ )
    \ make sure address is in loaded code, not some prior instance
    OVER R@
    ( cadrName uFindAddress fWasFound uFindAddress uPriorLastAddress  R: uHere uLatest@ )
    > IF
      OVER
      ( cadrName uFindAddress fWasFound uFindAddress  R: uHere uLatest@ )
      \ remove stack values so command params at TOS
      2>R
      ( cadrName uFindAddress  R: uHere uLatest@ uFindAddress fWasFound )
      SWAP >R
      EXECUTE
      R>
      ( cadrName  R: uLatest@ uHere uFindAddress fWasFound )
      2R>
      ( cadrName fWasFound uFindAddress R: uHere uLatest@ )
      SWAP
      ( cadrName uFindAddress fWasFound  R: uHereu Latest@ )
      ROT DROP
      ( uFindAddress fWasFound  R: uHereu Latest@ )
    ELSE
      ( cadrName uFindAddress fWasFound  R: uHere uLatest@ )
      \ requested word was not in loaded code
      2DROP FALSE
      ( cadrName fWasFound  R: uHere uLatest@ )
    THEN  
  ELSE  
    ( cadrName uFindAddress fWasFound  R: uHere uLatest@ )
    NIP
    ( cadrName fWasFound  R: uHere uLatest@ )
  THEN
  ( cadrName | uFindAddress fWasFound  R: uHere uLatest@ )
  \ unload the code from memory, restore dictionary pointers
  2R> DP !  LATEST !
  ( uFindAddress fWasFound )
  \ Find returns the name string address on failure, 
  \ name becomes abort message
  0= SWAP COUNT
  ( fWasNotFound cadrName uCount )
  \ report error if any, 
  ?ABORT ;

\ Examples of using the exec-blocks function.
\ Set block numbers to actual location of forth code
    
\ load block editior from blocks 11 thru 19 
\ and excute, unload when done.
: edit ( uBlockIndex -- )
  C" edit" #11 #19 exec-blocks ;

\ load load-blocks ForthTerm transfer software
\ from block 21 and execute, unload when done.
: load-blocks ( uFromBlock uLastBlock -- )
  C" load-blocks" #21 #21 exec-blocks ; 

\ load send-blocks ForthTerm transfer software
\ from block 22 and excute, unload when done.
: send-blocks ( uFromBlock uLastBlock -- )
  C" send-blocks" #22 #22 exec-blocks ;   

: noload ( uFromBlock uLastBlock -- )
  C" noload" #220 #220 exec-blocks ;   
