'' ANSI Terminal Emulation for One/Two Pin TV driver
'
' The screen buffer must be long aligned and requires
' one byte per 7 bit ASCII character. The high bit indicates
' if the character is drawn in reverse video. The number of
' columns must evely divide by 4 (ie; 64, 68, 72, 76, 80).


Con
  C_Cols = 80
  C_Rows = 25
  C_ScreenBytes = (C_Cols * C_Rows)
  C_LastRow = (C_Rows - 1)
  C_LastCol = (C_Cols - 1)
  C_RowLongs    = (C_Cols / 4)
  '
  ' Terminal options
  to_LineWrap   = 1
  to_Invert     = 2
  to_ShowCursor = 4
  '
  ' Terminal reset options state
  to_Default = to_LineWrap | to_ShowCursor 

  'Tab stop constants
  c_TabClear    =  0            'Indicates no tab stop, MUST BE ZERO
  c_TabSet      =  1
  
var
  ' Params required by ANSI COG
  long  vTabStopsAdr
  long  vScreenAdr
  long  vCursorAdr
    
Pub Start

Dat

Entry                   mov     T1, PAR
                        rdlong  ScreenAdr, T1
                        add     T1, #4
                        rdlong  CursorAdr, T1
                        add     T1, #4
                        rdlong  TabStopsAdr, T1
                        ' Calculate commonly used values
                        mov     ScreenLastAdr, ScreenAdr
                        add     ScreenLastAdr, ScreenBytes
                        call    #Reset

'
'
' Reset all terminal parameters, clear screen and home cursor
Reset                   mov     ScreenWrPtr, ScreenAdr
                        mov     ScreenRow, #0        
                        mov     ScreenCol, #0
                        call    #ClrParams        
                        call    #ClrScreen
                        ' Scroll region set to full screen
                        mov     RegStart, #0
                        mov     RegLast, #C_LastRow
                        ' Set default terminal options
                        mov     TermOptions, #to_Default
Reset_ret               ret



' Calculate the start address of the row in T1, T1 = screen address
tmp1                    long 0
tmp2                    long 0
'

RowStartAdr             mov     tmp2, #C_Cols            'Multiply by number of columns
                        mov     tmp1, T1
                        mov     T1, #0                                   
                        'Do the add and shift a max of 32 times, quit when (T1 << n) = 0
RowStartAdr_1           shr     tmp1, #1                  wc, wz
        if_c            add     T1, tmp2
                        shl     tmp2, #1
        if_nz           jmp     #RowStartAdr_1
                        add     T1, ScreenAdr
RowStartAdr_ret         ret
'
'
' Clear row, row number is in ScreenRow
' Retuens next write address in T1
ClrRow        mov       T1, ScreenRow
              call      #RowStartAdr
' Clear next row, row screen address is in T1
' Retuens next screen write address in T1
ClrNextRow              mov     T2, #C_RowLongs
:ClrRowRpt              wrlong  ClearLongFill, T1
                        add     T1, #4
                        djnz    T2, #:ClrRowRpt
ClrNextRow_ret          
ClrRow_ret              ret
'
'
' Fill entire screen with clear character
ClrScreen               mov     T1, ScreenAdr
ClrScreenRpt            call    #ClrNextRow
                        cmp     T1, ScreenLastAdr       wc
        if_c            jmp     #ClrScreenRpt
ClrScreenRpt_ret
ClrScreen_ret           ret
'
'
' Clear the current row from the current column to the end of the line.
ClrEol                  mov     T1, ScreenWrPtr
                        mov     T2, #C_Cols
                        sub     T2, ScreenCol
:ClrEolRpt              wrbyte  ClearLongFill, T1
                        add     T1, #1
                        djnz    T2, #:ClrEolRpt
ClrEol_ret              ret
'
'                      
' Clear the current row from the current column to the begining of the line.
ClrBol                  mov     T1, ScreenWrPtr
                        sub     T2, ScreenCol
                        add     T2, #1
:ClrBolRpt              wrbyte  ClearLongFill, T1
                        sub     T1, #1
                        djnz    T2, #:ClrBolRpt
ClrBol_ret              ret
'
'
' Clear the current row from the current column to the end of the screen
ClrEos                  call    #ClrEol
                        ' Clear the remaing rows if any
                        cmp     ScreenRow, #C_LastRow   wz
              if_z      jmp     #ClrEos_ret
                        call    #ClrScreenRpt
ClrEos_ret              ret
'
' Clear the current row from the current column to the start of the screen
ClrBos                  call    #ClrBol
                        ' Clear the remaing rows if any
                        tjz     ScreenRow, #ClrBos_ret
                        ' There is at least one more row to clear
                        mov     T1, ScreenAdr
                        mov     T3, ScreenRow
                        sub     T3, #1
ClrBosRpt               call    #ClrNextRow
                        djnz    T3, #ClrBosRpt
ClrBos_ret              ret
'
'
' Clear the esc sequence parameters
ClrParams               movd    ParamCnt, #0    
                        mov     Param1, #0
                        mov     Param2, #0
ClrParams_ret           ret
'
'
' Move the cursor to the prior row, if Param1 > 1 repeat Param1 times.
PriorRow                tjnz    Param1, #:PriorRow1                        
                        mov     Param1, #1
:PriorRow1              tjz     ScreenRow, #PriorRow_ret
                        sub     ScreenWrPtr, #C_Cols
                        sub     ScreenRow, #1 
                        djnz    Param1, #:PriorRow1
PriorRow_ret            ret
'
'
' Move the cursor to the next row, if Param1 > 1 repeat Param1 times.
NextRow                 tjnz    Param1, #:NextRow1                        
                        mov     Param1, #1
:NextRow1               cmp     ScreenRow, #C_LastRow   wz
              if_z      jmp     #NextRow_ret
                        add     ScreenWrPtr, #C_Cols
                        add     ScreenRow, #1 
                        djnz    Param1, #:NextRow1
NextRow_ret             ret
'
'
' Move the cursor to the prior column, if Param1 > 1 repeat Param1 times.
PriorCol                tjnz    Param1, #:PriorCol1                        
                        mov     Param1, #1
:PriorCol1              cmp     ScreenCol, Param1       wc
              if_c      mov     Param1, ScreenCol
                        sub     ScreenWrPtr, Param1
                        sub     ScreenCol, Param1 
PriorCol_ret            ret
'
'
' Move the cursor to the next column, if Param1 > 1 repeat Param1 times.
NextCol                 tjnz    Param1, #:NextCol1                        
                        mov     Param1, #1
:NextCol1               mov     T1, #C_LastCol
                        sub     T1, ScreenCol
                        cmp     T1, Param1              wc
              if_c      mov     Param1, T1
                        add     ScreenWrPtr, Param1
                        add     ScreenCol, Param1 
NextCol_ret             ret
'
'
' Move the cursor to the column in Param1, the cursor row is maintained
GotoCol                 tjz     Param1, #:GotoCol1                        
                        sub     Param1, #1
:GotoCol1               cmp     Param1, #C_LastCol      wc
              if_nc     mov     Param1, #C_LastCol
                        mov     T1, ScreenRow
                        call    #RowStartAdr
                        mov     ScreenWrPtr, T1
                        add     ScreenWrPtr, Param1
                        mov     ScreenCol, Param1
GotoCol_ret             ret                 
'
'
' Move the cursor to the row in Param1, the cursor column is maintained
GotoRow                 tjz     Param1, #:GotoRow1                        
                        sub     Param1, #1
:GotoRow1               cmp     Param1, #C_LastRow      wc
              if_nc     mov     Param1, #C_LastRow
                        mov     T1, Param1
                        call    #RowStartAdr
                        mov     ScreenWrPtr, T1
                        add     ScreenWrPtr, ScreenCol
                        mov     ScreenRow, Param1
GotoRow_ret             ret                 


GotoRowCol              call    #GotoRow
                        mov     Param1, Param2
                        call    #GotoCol
GotoRowCol_ret          ret


' Set scroll region to values in Param1 and Param2
SetScrollReg            tjnz    ParamCnt, #SetScrollRegParams
                        mov     RegStart, #0
                        mov     RegLast, #C_LastRow
                        jmp     #SetScrollReg_ret
                        ' we have at least one paramatere, make sure we have two
SetScrollRegParams      cmp     ParamCnt, #2            wz
              if_nz     jmp     #SetScrollReg_ret
                        ' make sure Param1 <= Param2
                        cmp     Param2, Param1          wc
              if_c      jmp     #SetScrollReg_ret
                        ' make sure Param1 <> 0
                        tjz     Param1, #SetScrollReg_ret
                        sub     Param1, #1                                                                              
                        sub     Param2, #1
                        ' Make sure Param2 < screen rows
                        cmp     Param2, #C_Rows         wc                                                                              
              if_nc     jmp     #SetScrollReg_ret
                        ' assign new values
                        mov     RegStart, Param1
                        mov     RegLast, Param2
SetScrollReg_ret        ret
'
'
' Delete row(s) starting with row in T1.  Startong row must be in scroll region.
' Param1 = number of rows to delete.
DeleteRows              tjnz    Param1, #:DelRows1
                        mov     Param1, #1
                        ' First see if start row is in scroll region
:DelRows1               mov     T2, RegLast
                        sub     T2, T1     wc
              if_nc     cmp     T1, RegStart    wc
              if_c      jmp     #DeleteRows_ret
                        ' start row is in scroll region, calc how many lines we can delete
                        add     T2, #1
                        ' range check rows to delete value in Param1
                        max     Param1, T2 
                        ' Param1 = rows to delete, T1 = first row to delete
                        mov     Param3, T1
                        add     Param3, Param1
                        mov     Param2, T1
                        ' Param1 = rows to delete, Param2 = first row to delete
                        ' Param3 = first row to copy to first deleted row  
                        call    #RowStartAdr            ' calc copy dest address
                        mov     T2, T1                  ' dest address, save in T2
                        mov     T1, RegLast             ' T1 = last region row index
                        add     T1, #1                  ' add one more row to get address at start of next row
                        call    #RowStartAdr            ' calc memory address of last region row + 1
                        mov     T3, T1                  ' save in T3
                        mov     T1, Param3
                        call    #RowStartAdr            ' calc copy source address, save in T1
                        ' T1 = copy source memory address, T2 = copy dest memory address, T3 = end of region memory address
                        ' copy until T1 => last screen scroll region memory addresss
:DelRows2               cmp     T1, T3       wc         ' see if we are done copy
              if_nc     jmp     #:DelRows3
                        rdlong  T4, T1
                        add     T1, #4
                        wrlong  T4, T2
                        add     T2, #4
                        jmp     #:DelRows2
                        ' We have copied all the rows, now clear the lower region rows
                        ' T2 = first address to clear                         
:DelRows3               cmp     T2, T3       wc         ' see if we are done erasing row(s)
                        wrlong  ClearLongFill, T2
                        add     T2, #4
                        jmp     #:DelRows3
DeleteRows_ret          ret              
'
'
' Insert row(s) starting with row in T1.  Starting row must be in scroll region.
' Param1 = number of rows to insert.
InsertRows              tjnz    Param1, #:InsRows1
                        mov     Param1, #1
                        ' First see if start row is in scroll region
:InsRows1               mov     T2, RegLast
                        sub     T2, T1     wc
              if_nc     cmp     T1, RegStart    wc
              if_c      jmp     #InsertRows_ret
                        ' start row is in scroll region, calc how many lines we can insert
                        add     T2, #1
                        ' range check rows to insert value in Param1
                        max     Param1, T2 
                        mov     Param2, T1
                        ' Param1 = rows to insert, Param2 = first screen row to insert
                        ' Reverse move screen data down number of rows to insert.
                        ' calc copy destination last address 
                        mov     T1, RegLast             
                        add     T1, #1
                        call    #RowStartAdr
                        sub     T1, #4
                        mov     T2, T1                  ' T2 = memory address of last scroll region row's long value
                        ' calc copy source last address 
                        mov     T1, RegLast
                        add     T1, #1             
                        sub     T1, Param1
                        call    #RowStartAdr
                        sub     T1, #4
                        mov     T3, T1                  ' T3 = last source address
                        ' T2 = copy dest address, T3 = copy source address
                        ' We stop the copy when we reach start address of inserted row
                        mov     T1, Param2
                        call    #RowStartAdr
:InsRows2               cmp     T3, T1          wc
              if_c      jmp     #:InsRows3
                        ' copy a long of screen memory
                        rdlong  T4, T3
                        sub     T3, #4                        
                        wrlong  T4, T2
                        sub     T2, #4                        
                        jmp     #:InsRows2
                        ' screen data copied down, now blank lines, Dest copy address is last address to blank
:InsRows3               cmp     T2, T1          wc
              if_c      jmp     #InsertRows_ret
                        ' copy a long of screen memory
                        wrlong  ClearLongFill, T2
                        sub     T2, #4                        
                        jmp     #:InsRows3
InsertRows_ret          ret              
'
'
' Delete Param1 characters from curent position (row, col)
' Characters are shifted left and characters at end of row are cleared
DeleteChars             tjnz    Param1, #:DelChars1
                        mov     Param1, #1
                        ' Calc max chars we can delete, save in Param1
:DelChars1              mov     T1, #C_Cols
                        sub     T1, ScreenCol     
                        max     Param1, T1
                        ' First shift chars left starting at current position
                        ' Calc source copy memory address
                        mov     T3, ScreenWrPtr
                        add     T3, Param1
                        ' T2 = copy dest memory address
                        mov     T2, ScreenWrPtr
                       ' Calc start address of next row
                        mov     T1, ScreenRow
                        add     T1, #1  
                        call    #RowStartAdr
                        ' Copy characters left on row
:DelChars2              cmp     T3, T1          wc
              if_nc     jmp     #:DelChars3
                        rdbyte  T4, T3
                        add     T3, #1
                        wrbyte  T4, T2
                        add     T2, #1
                        jmp     #:DelChars2
                        ' Blank end of line                           
:DelChars3              cmp     T2, T1          wc
              if_nc     jmp     #DeleteChars_ret
                        wrbyte  ClearLongFill, T2
                        add     T2, #1
                        jmp     #:DelChars3
DeleteChars_ret         ret
'
'
' Insert Param1 characters at the curent position (row, col)
' Characters are shifted right and inserted characters are cleared.
' The operation is limited to the current row.
InsertChars             tjnz    Param1, #:InsChars1
                        mov     Param1, #1
                        ' Calc max chars we can insert, use to ramge check Param1
:InsChars1              mov     T1, #C_Cols
                        sub     T1, ScreenCol     
                        max     Param1, T1
                        ' First shift chars right starting at current position
                        ' Need to do a reverse copy.
                       ' Calc last character address of row
                        mov     T1, ScreenRow
                        add     T1, #1  
                        call    #RowStartAdr
                        sub     T3, #1          ' last row character address
                        ' Calc end of copy source memory address
                        mov     T3, T1
                        sub     T3, Param1
                        ' T2 = copy dest end memory address
                        mov     T2, T1
                        ' copy characters right
:InsChars2              cmp     T3, ScreenWrPtr         wc, wz
        if_nc_and_nz    jmp     #:InsChars3
                        rdbyte  T4, T3
                        sub     T3, #1
                        wrbyte  T4, T2
                        sub     T2, #1
                        jmp     #:InsChars2
                        ' now clear inserted characters
:InsChars3              mov     T2, ScreenWrPtr
:InsChars4              wrbyte  ClearLongFill, T2
                        add     T2, #1
                        djnz    Param1, #:InsChars4
InsertChars_ret         ret
'
'                        
'
'Set the current column as a tab stop.
DoSetTab                mov     t2, #c_TabSet           'we want to set a tab stop.
                        jmp     #UpdateTabStop
'Clear the current column as a tab stop.
DoClearTab              mov     t2, #c_TabClear         'we want to clear a tab stop.
'                       'update the current column tab stop
UpdateTabStop           mov     t1, ScreenCol                  
                        add     t1, TabStopsAdr
                        wrbyte  t2, t1                             
                        'go wait for next command
DoClearTab_ret          ret
'
'
'Clear all tab stops
DoClearTabs             mov     t1, TabStopsAdr
                        mov     t2, #C_Cols
                        mov     t3, #c_TabClear
DoClearTabsRpt          wrbyte  t3, t1
                        add     t1, #1
                        djnz    t2, #DoClearTabsRpt                  
                        'go wait for next command
DoClearTabs_ret         ret
'
'
'Find the next tab stop. Tabs never line wrap.
'On Exit: The global param[0] will contain the new column. 
'if no tab is found the retuen value is the current column.
'Tabs are saved as an array of bytes in global memory.
DoNextTab               mov     t1, ScreenCol           'Read current column
                        'T1 = current column
                        mov     t2, t1                  'T2 = current column
                        mov     t3, TabStopsAdr         'point to first tab stop
                        add     t3, t2                  'T3 = memory address of next tab to check
                        'T1 = return value, T2 = next tab stop to check, T3 = address of tab to check
DoNextTabRpt            add     t2, #1                  'T2 = next column index to check
                        add     t3, #1                  'T3 = next tab global memory address
                        cmp     t2, #C_Cols     wz      'See if we are done
              if_z      jmp     #DoNextTab_ret          'No tab found just return current value
                        'see if this is a tab stop (<> 0)
                        rdbyte  t4, t3          wz
                        'if tab stop set (<> 0), copy column to T1 and quit
              if_nz     mov     t1, t2
                        'Not a valid tab stop, try next column if any.
              if_z      jmp     #DoNextTabRpt           'No tab try next column.
                        'All done T1 = return value
DoNextTab_ret           ret                                                                 
'
'
'Constants
ScreenBytes             long    C_ScreenBytes
ClearLongFill           long    $20202020

'These Variables are initialized at start up. Must stay in order
TabStopsAdr             res     1
CursorAdr               res     1
ScreenAdr               res     1
'Variables
ScreenLastAdr           res     1
ScreenWrPtr             res     1
ScreenRow               res     1
ScreenCol               res     1
RegStart                res     1
RegLast                 res     1
TermOptions             res     1

' Temp variables
T1                      res     1
T2                      res     1
T3                      res     1
T4                      res     1
ParamCnt                res     1
Param1                  res     1
Param2                  res     1
Param3                  res     1