.( timer demo, uses a COG to keep time )

\ This value must reflect the block number 
\ used to store COG binary image.
#22 VALUE timer_blk

\ used to save COG index timer is running on
0 VALUE timer_cog

#16 QBYTES timer_data ( save space for four longs )

\ timer_data returns address of packed BCD time (HHMMSSDD 4 bytes)
\ where HH = hours (0..23), MM = minutes, SS = seconds, DD = hundreths of a second

\ returns address of packed BCD date (YYYYMMDD 4 bytes)
\ where YYYY = year (0..9999), MM = month (1..12), DD = day of month (1..31)
: timer_date
  timer_data #4 + ; 
  
\ returns address of timer muxtex lock index 0..7
: timer_lock
  timer_data #8 + ; 

\ returns address of unsigned double, number of system ticks per 10ms interval
\ can be adjusted to trim clock.
: timer_ticks 
  timer_data #12 + ; 

\ allocate a hardware mutex and save index
LOCK-NEW S>D timer_lock Q! 

\ calculate the ideal number of system clock ticks per 10ms interval 
CLKFREQ #100. D/ timer_ticks Q!

\ lock the timer data so cog will not be able to update time/date
: timer-lock ( -- ) 
  BEGIN 
    timer_lock Q@ D>S LOCK-SET
  UNTIL ; 
  
\ unlock the timer data so cog will be able to update time data
: timer-unlock ( -- )
  timer_lock Q@ D>S LOCK-CLR DROP ;

\ set the time in packed BCD format (HHMMSSDD)
: set-time ( dBcdTime -- )
  timer-lock timer_data Q! timer-unlock ;

\ set the date in packed BCD format (YYYYMMDD)
: set-date ( dBcdDate -- )
  timer-lock timer_date Q! timer-unlock ;

\ set the date and time at the same time
: set-datetime ( dBcdDate dBcdTime -- )
  timer-lock
  timer_data Q!
  timer_date Q!
  timer-unlock ;

\ set the data and time to zero, handy when clock is used as a long term timer  
: clr-time ( -- )
  0. 0. set-datetime ;

\ convert a packed 4 byte BCD time into a string
: bcdtimestr ( dBcdTime -- cadr uCount )
  \ save current numeric base and set to base 16 
  BASE @ >R HEX
  \ format value as BCD time with seperators
  <# # # $2E HOLD # # $3A HOLD # # $3A HOLD # # #> 
  \ restore save numeric base value
  R> BASE ! ;

\ convert a packed 4 byte BCD date into a string
: bcddatestr ( dBcdTime -- cadr uCount )
  \ save current numeric base and set to base 16 
  BASE @ >R HEX
  \ format value as BCD date with seperators
  <# # # $2F HOLD # # $2F HOLD # # # # #> 
  \ restore save numeric base value
  R> BASE ! ;
    
\ display the current time on the current console device
: time. ( -- ) 
  timer_data Q@ bcdtimestr TYPE ;

\ display the current date on the current console device
: date. ( -- ) 
  timer_date Q@ bcddatestr TYPE ;

\ start the timer cog passing the address of the time data structure
: timer-start ( -- ) 
  timer_data timer_blk COG-LOAD 
  TO timer_cog ;

