\ implements Forth DATE and TIME functions
\ requires Timer.forth (cog timer)

MARKER unload-time&date
  
\ Zeller's Congruence 
: zeller ( uMonth -- days since March 1 ) 
  #9 + #12 MOD 1-
  #26 #10 */ #3 + ; 

\ calculate the day of the week 0..6 using
\ Zeller's Congruence 
: WEEKDAY ( uDay uMonth uYear -- uDayOfWeek )   
\ Monday..Sunday 0..6
  OVER #3 < IF 1- THEN DUP 4 /  
  OVER #100 / -  
  OVER #400 / +  +  SWAP 
  zeller + +  1+ #7 MOD ;
  
\ read low byte as bcd and covert to byte
\ the uValue is returned shifted right 8 bits
: bcd>byte ( uValue -- uValue' uBinValue )
  DUP #8 RSHIFT SWAP $FF AND
  ( uValue' uValueLSB )
  DUP #4 RSHIFT 10 *
  ( uValue' uValueLSB uValueMSN*10 )
  SWAP $0F AND
  ( uValue' uValueMSN*10 uValueLSN )
  + 
  ( uValue' uBinValue )
  ; 
  
\ convert a 4 digit bcd to a word  
: bcd>word ( uBcdValue -- uWord )
  bcd>byte
  ( uBcdValue' uLowValue ) 
  SWAP bcd>byte NIP
  ( uLowValue uHighValue )
  #100 * + ; 

\ convert a value into a two digit packed BCD value
\ value must be in range of 0..99
: byte>bcd ( uByteValue -- uBcdValue )
  #10 /MOD 
  #4 LSHIFT OR ;
  
\ convert a value into a four digit packed BCD value
\ value must be in range of 0..9999)  
: word>bcd ( uWordValue -- uBcdValue)  
  #100 /MOD
  byte>bcd #8 LSHIFT SWAP 
  byte>bcd OR ;
    
  
\ read and decode current time value
: TIME ( -- uSec uMin uHour )
    \ read time and remove hundreths of a second
    timer_data Q@
    >R #8 RSHIFT bcd>byte NIP
    R> bcd>byte SWAP bcd>byte NIP ;

\ read and decode current date
: DATE ( -- uDay uMonth uYear )
    timer_date Q@
    >R bcd>byte SWAP bcd>byte NIP
    R> bcd>word ;

\ read the time and date from the Timer COG 
: TIME&DATE ( -- uSec uMin uHour uDay uMonth uYear ) 
  TIME DATE ;
  
\ set the date
: DATE! ( uDay uMonth uYear -- )
  word>bcd
  ( uDay uMonth uBcdYear )
  >R byte>bcd #8 LSHIFT
  ( uDay uBcdMonth  R: uBcdYear )
  SWAP byte>bcd OR R>
  ( dBcdDate )
  set-date ;
   
\ set the time
: TIME! ( uSec uMinute uHour -- )
  byte>bcd #8 LSHIFT SWAP byte>bcd OR
  ( uSec uBcdHourMin )
  SWAP byte>bcd #8 LSHIFT SWAP
  ( dBcdTime )
  set-time ;
   
\ read the time and date from the Timer COG 
: TIME&DATE! ( uSec uMin uHour uDay uMonth uYear ) 
  DATE! TIME! ;
  
