MARKER NET-CON
\ test program....

0     VALUE    nc_rxcnt    \ Number of bytes in rx buffer
0     VALUE    nc_rxidx    \ Index of next byte to read from rx buffer

0     VALUE    nc_txidx    \ Index of next Tx buffer byte to write to

-1    VALUE    nc_def_skt  \ Socket used to auto open console
-1    VALUE    nc_socket   \ current console socket
0     VALUE    nc_port     \ Console port number
FALSE VALUE    net_constat \ Indicates if socket is connected
0     VALUE    nc_session   \ Incremented each time a new connection is made
  
#512 CONSTANT nc_rxsize     \ Size of rx buffer in bytes
#512 CONSTANT nc_txsize     \ Size of tx buffer in bytes

nc_rxsize QBYTES nc_rxbuf  \ Rx buffer 
nc_txsize QBYTES nc_txbuf  \ Tx buffer 

TRUE VALUE net_autocon

\ place to save last tx byte system tick count
2VARIABLE nc_nag

\ calc number of system ticks for tx buffer Nagle time out
CLKFREQ #100. D/ 2CONSTANT nc_nagmax ( 10ms Nagle time )

\ clear the tx buffer Nagle timer
: nc-nagclr ( -- )
  0. nc_nag 2! ;
  
nc-nagclr
  
\ update the tx buffer Nagle timer, called each time byte is written
: nc-nagtx ( -- )
  COG-CNT nc_nag 2! ;

\ see if the tx buffer Nagle timer has timed out  
: nc-nag? ( -- f )
  nc_nag 2@ 2DUP
  ( dTimerCnt dTimerCnt  ) 
  D0= IF
    ( dTimerCnt )
    \ timer is = 0 so just return FALSE (0)
    DROP
  ELSE
    \ see if timer has timed out
    ( dTimerCnt )
    COG-CNT 2SWAP D-
    ( dTimeDelta )
    nc_nagmax UD>
    ( fTimedOut ) 
    DUP IF
      \ if we timed out clear tx buffer timer
      nc-nagclr
    THEN  
  THEN ;

: nc-init ( -- )
  FALSE TO net_constat
  nc-nagclr
  0 TO nc_rxidx
  0 TO nc_rxcnt 
  0 TO nc_txidx ;
  
: nc-rxbuf? ( -- f ) 
  nc_rxidx nc_rxcnt < ;

  
: nc-rxbuf ( -- c ) 
  nc_rxidx DUP nc_rxbuf + C@ SWAP
  1+ TO nc_rxidx ;

: nc-rxfill? ( -- )
  nc_socket -1 = IF
    FALSE
  ELSE
    nc_socket skt-status DUP #23 = SWAP #28 = OR
  THEN ;

\ Indicate no data in rx buffer
: nc-rxclr ( -- ) 
  0 TO nc_rxidx
  0 TO nc_rxcnt ;

\ read characters from TCP stream to fill rx buffer
: nc-rxfill ( -- )
  nc-rxclr
  nc-rxfill? IF
    nc_socket nc_rxbuf nc_rxsize tcp-read TO nc_rxcnt
  THEN
  nc_rxcnt 0= IF
    nc_socket skt-status #28 = IF
      nc_socket tcp-disconnect 
    THEN
  THEN ;

\ wait for socket to enter specified state
: nc-waitfor ( uSocket uStatus -- f )
  >R BEGIN 
    DUP skt-status DUP R@ = SWAP 0= OR DUP 0= IF
      #2 MS
    THEN 
  UNTIL 
  skt-status R> = ;
  
\ wait for a socket to enter listen state  
: nc-wtlisten ( uSocket -- f ) #20 nc-waitfor ;  

\ write the tx data from the buffer to the TCP stream
: nc-flush ( -- )
  nc_txidx 0> net_constat AND IF
    \ there is at least one byte to write
    \ TCP write blocks until data writen or not connected
    nc_socket nc_txbuf nc_txidx tcp-write DROP
    0 TO nc_txidx
    nc-nagclr
  THEN ;
  
: nc-closeskt ( uSocket -- )
  ( see if socket can close )
  DUP skt-status #23 = IF
    ( uSocket )
    ( tear down tcp connection )
    DUP tcp-disconnect
    ( uSocket )
    ( wait for socket to close )
    BEGIN 
      DUP skt-status 0= DUP NOT IF
        #2 MS
      THEN 
    UNTIL
  THEN
  ( uSocket )
  skt-close ;
       
: nc-close ( -- )
  nc_socket -1 <> IF
    net_constat IF
      ( we are connected buffers should be flushed )
      nc-flush
    THEN
    ( close socket )
    nc_socket nc-closeskt
    -1 TO nc_socket
  THEN
  \ clear all buffers
  0 TO nc_txidx
  nc-nagclr
  nc-rxclr
  ( indicate we are no longer connected )
  FALSE TO net_constat ;
  
: nc-openskt ( uSocket uPort -- f )
  OVER DUP nc-closeskt
  ( uSocket uPort uSocket )
  -1 > IF
    ( uSocket uPort )
    OVER SWAP 1 SWAP
    ( uSocket uSocket uProt uPort ) 
    0 0 skt-open 
    ( uSocket ) 
    ( increment session count )
    nc_session 1+ TO nc_session
    ( place socket into listen mode )
    DUP tcp-listen nc-wtlisten
   ELSE 
    ( uSocket uPort )
    2DROP FALSE
   THEN ;
   
: nc-open ( uSocket uPort -- f )
  net_constat IF nc-close THEN
  ( save the port )
  TO nc_port
  ( save the socket )
  net_autocon IF
    \ set default socket to allow reopen
    DUP
  ELSE
    \ set default socket to prohibit reopen
    -1 SWAP
  THEN
  TO nc_socket
  ( save copy of the socket in case we auto open )
  TO nc_def_skt
  ( open the socket port )
  nc_socket nc_port nc-openskt ;
  
: nc-reopen ( -- )
  nc_def_skt -1 <> IF
    nc_def_skt DUP TO nc_socket 
    nc_port nc-openskt DROP
  THEN ;
  
\ tell W5x00 to send a TCP KEEP-ALIVE packet
\ This will test to see if peer is still there
: nc-keep ( -- )
  nc_socket DUP -1 <> IF
    ( uSocketIdx )
    DUP skt-status #23 = IF
      ( uSocketIdx )
      $22 SWAP skt-cr!
    ELSE  
      ( uSocketIdx )
      DROP    
    THEN
  ELSE
    ( uSocketIdx )
    DROP    
  THEN ;  
  

: nc-tcpbg ( -- )
  nc_socket -1 <> IF
    nc_socket skt-status DUP 
    #23 = DUP TO net_constat IF
      ( we are connected )
      nc-nag? IF 
        nc-flush 
      THEN
      DROP
    ELSE
      0= IF
        ( socket is closed, re-open and set to listen )
        nc-close
        nc-reopen
      THEN
    THEN
  ELSE
    nc-reopen
  THEN ;

: nc-rx? ( -- f )
  nc-rxbuf? ?DUP NOT IF
    \ nothing in buffer try to fill
    nc-tcpbg
    nc-rxfill 
    nc-rxbuf?
  THEN ;

: nc-rx ( -- c ) 
  BEGIN 
    nc-rx? 
  UNTIL 
  nc-rxbuf ;

( transmit character code )

\ see if we have room to write another byte to TX buffer
: nc-txbuf? ( -- f ) 
  nc_txidx nc_txsize < ;
  
\ write byte c to TX buffer
: nc-txbuf ( C -- ) 
  nc_txidx SWAP OVER nc_txbuf + C!
  1+ TO nc_txidx ;

: nc-tx? ( -- f )
  nc-txbuf? ?DUP NOT IF
    \ buffer is full try to send
    nc-tcpbg
    nc-flush 
    nc-txbuf?
  THEN ;

: nc-tx ( c -- ) 
  BEGIN 
    nc-tx? 
  UNTIL 
  nc-nagtx nc-txbuf ;

\ initialize net-console

nc-init

: NET>CON ( -- )
  ['] nc-rx? IS KEY? 
  ['] nc-rx IS KEY
  ['] nc-tx? IS EMIT? 
  ['] nc-tx IS EMIT 
  ['] nc-rxclr IS EMPTY-KEYS ;
