\ W5100 TCP Driver

\ First block that holds binary (2 blocks)
#6 VALUE net-blk

\ Number of sockets we can use for TCP
#4 CONSTANT net_sockets 

\ Our local network settings
2VARIABLE net_mac #2 ALLOT
2VARIABLE net_ip 
2VARIABLE net_mask
2VARIABLE net_blankmask
2VARIABLE net_gw

\ used by NetConfig and WebServer
#8080 VALUE net_defport

0 VALUE net_cog
0. net_blankmask 2!

\ constants
\  1. 16 DLSHIFT 2CONSTANT c_readSPI
\  2. 16 DLSHIFT 2CONSTANT c_writeSPI

\  3. 16 DLSHIFT 2CONSTANT c_setMAC
\  4. 16 DLSHIFT 2CONSTANT c_setGW
\  5. 16 DLSHIFT 2CONSTANT c_setMASK
\  6. 16 DLSHIFT 2CONSTANT c_setIP

\  7. 16 DLSHIFT 2CONSTANT c_PingBlock
\  8. 16 DLSHIFT 2CONSTANT c_rstHW
\  9. 16 DLSHIFT 2CONSTANT c_rstSW

\ 10. 16 DLSHIFT 2CONSTANT c_Sopen
\ 11. 16 DLSHIFT 2CONSTANT c_Sdiscon
\ 12. 16 DLSHIFT 2CONSTANT c_Sclose
\ 13. 16 DLSHIFT 2CONSTANT c_Sread
\ 14. 16 DLSHIFT 2CONSTANT c_Sread_tcp
\ 15. 16 DLSHIFT 2CONSTANT c_Swrite
\ 16. 16 DLSHIFT 2CONSTANT c_Swrite_tcp
\ 17. 16 DLSHIFT 2CONSTANT c_Sstatus



\ socket address constants
$0400 CONSTANT c_BASE_MR
$0401 CONSTANT c_BASE_CR
\ $0403 CONSTANT c_BASE_SR
\ $0420 CONSTANT c_BASE_TX_FSR
\ $0424 CONSTANT c_BASE_TX_WR
\ $0426 CONSTANT c_BASE_RX_RSR
\ $0428 CONSTANT c_BASE_RX_RD                                    
$040C CONSTANT c_BASE_DIPR
$0410 CONSTANT c_BASE_DPORT


\ $6000 CONSTANT c_RX-BASE
\ $4000 CONSTANT c_TX-BASE

\ $07FF CONSTANT c_RX_MASK
\ $07FF CONSTANT c_TX_MASK

$13 CONSTANT c_SOCK_INIT
$17 CONSTANT c_SOCK_ESTAB

$02 CONSTANT c_CMD_LISTEN
$04 CONSTANT c_CMD_CONNECT
\ $40 CONSTANT c_CMD_RECV
\ $20 CONSTANT c_CMD_SEND
\ $22 CONSTANT c_CMD_SEND_KEEP

\ a place to hold command and up to 7 parameters
QVARIABLE net_command
net_command #4 ERASE
\ a place to hold data for single values
#24 QBYTES net_data
net_data #24 ERASE
QVARIABLE net_temp
net_temp #4 ERASE

\ copy n bytes from stack and save at address a-adr
\ bytes are saved in reverse order
: net-wrnet ( cN...c1 c0 n a-adr -- )
  OVER + 1- SWAP 0 ?DO
    ( cN...c1 c0 a-adr' -- )
    SWAP OVER C! 1- 
  LOOP DROP ;

\ copy 4 bytes from stack and save to a-adr
: net-wrnet4 ( c3 c2 c1 c0 a-adr -- )
  #4 SWAP net-wrnet ;
  
\ convert a socket register offset to actual W5100 reg
: skt-reg ( uSocket uRegOfs -- uSktRegister )
  SWAP $100 * + ;

\ Set MAC address to 6 bytes on data stack
: net-mac! ( c5 c4 c3 c2 c1 c0 -- )
  #6 net_mac net-wrnet ;

\ Set IP address to 4 bytes on data stack
: net-ip! ( c3 c2 c1 c0 -- )
  net_ip net-wrnet4 ;

\ Set net address mask to 4 bytes on data stack
: net-mask! ( c3 c2 c1 c0 -- )
  net_mask net-wrnet4 ;

\ Set gateway address to 4 bytes on data stack
: net-gw! ( c3 c2 c1 c0 -- )
  net_gw net-wrnet4 ;

\ default network settings
$00 $1D $61 $E9 $5B $88 net-mac!
#10 0 1 #13 net-ip!
#255 #255 #255 0 net-mask!
#10 0 1 1 net-gw!

: net-byteswap ( u -- u' ) DUP #8 LSHIFT SWAP #8 RSHIFT OR ;

\ execute command and then wait for command to complete
: w51-do-cmd ( dCmd fParams -- )
  IF
    \ send the command
    NIP net_data SWAP net_command Q! 
  ELSE
    net_command Q! \ send the command
  THEN
  \ wait for command to complete
  BEGIN net_command Q@ #0. D= UNTIL ; 

: w51-push ( Un...U1 U0 nCount -- )
  DUP net_data SWAP 1- #2 CELLS * + SWAP
  ( Un...U1 U0 uLastParamAdr nCount -- )
  0 ?DO 
    ( Un...U1 U0 uLastParamAdr -- )
    SWAP OVER 0 SWAP Q! #2 CELLS -
  LOOP
  DROP ;

( commands to W5100 COG 1..18 )

\ single 16 bit param W5100 cog command
: w51-cmd1 ( nParam dCmd -- ) 
  SWAP DROP FALSE w51-do-cmd ;
  
: w51-spi-read ( uRegister uDataPtr uNumbytes -- )
  #3 w51-push 0 1 TRUE w51-do-cmd ;

: w51-spi-write ( uRegister uDataPtr uNumbytes -- )
  #3 w51-push 0 #2 TRUE w51-do-cmd ;

\ write 6 byte MAC address to W5100
: w51-mac! ( a-adr -- ) 0 #3 w51-cmd1 ;
  
\ write 4 byte gateway address to W5100
: w51-gw! ( a-adr -- ) 0 #4 w51-cmd1 ;

\ write 4 byte subnet mask to W5100
: w51-mask! ( a-adr -- ) 0 #5 w51-cmd1 ;

\ write 4 byte ip address to W5100
: w51-ip! ( a-adr -- ) 0 #6 w51-cmd1 ;

\ Set if W5100 responds to an echo (ping) request.
: w51-canping ( f -- ) 
  0= S>D net_data Q!
  0 #7 TRUE w51-do-cmd ;

\ Hardware reset of W5100.
: w51-resethw 0 #8 FALSE w51-do-cmd ;

\ Software reset of W5100.
: w51-resetsw 0 #9 FALSE w51-do-cmd ;

: skt-open ( uSocket uMode uSrcPort uDestPort uDestIPAdr )
  \ push all params
  #5 w51-push
  0 #10 TRUE w51-do-cmd ;

\ Close the specified socket on the W5100. 
: skt-close ( uSocket -- ) 
  net_temp ! net_temp 0 #12 w51-cmd1 ;

\ Read a number of bytes from the socket rx buffer
: skt-read ( uSocket c-Adr, uBytes -- uBytes' )
  \ no socket registers are updated.
  #3 w51-push 0 #13 TRUE w51-do-cmd net_data @ ;

\ write a number of bytes to the socket tx buffer
: skt-write ( uSocket c-Adr, uBytes -- uBytes' )
  \ no socket registers are updated.
  #3 w51-push 0 #15 TRUE w51-do-cmd net_data @ ;

: skt-status ( uSocket -- uStatus )
  1 w51-push 0 #17 TRUE w51-do-cmd 
  net_data @ ;

\ Read number of bytes available in socket RX buffer
: skt-read# ( uSocket -- uBytes )
  1 w51-push 0 #18 TRUE w51-do-cmd 
  net_data @ ;

( end of W5100 COG commands )


\ Read destination IP address
: skt-dipr@ ( uSocket -- uRX_RSR )
  c_BASE_DIPR skt-reg net_temp #4 w51-spi-read
  net_temp 2@ ;

\ Set the command register of the specified socket
: skt-cr! ( cCommand uSocket -- )
  c_BASE_CR skt-reg SWAP net_temp !   
  net_temp 1 w51-spi-write ;
  
\ Set the mode register of the specified socket
: skt-mr! ( cMode uSocket -- )
  c_BASE_MR skt-reg SWAP net_temp ! 
  net_temp 1 w51-spi-write ;

\ Read the mode register of the specified socket
: skt-mr@ ( uSocket -- cMode )
  c_BASE_MR skt-reg net_temp 1 w51-spi-read
  net_temp C@ ;

\ Read the protocol of the specified socket
: skt-proto@ ( uSocket -- cProto ) 
  skt-mr@ $0F AND ;

( tcp support functions )

\ Disconnects the specified socket on the W5100. 
: tcp-disconnect ( utcp-connectSocket -- ) 
  net_temp ! net_temp 0 #11 w51-cmd1 ;

\ Check if a socket has established a TCP connection
: tcp-connected ( uSocket -- f ) 
  skt-status  c_SOCK_ESTAB = ;

\ Read a number of bytes from the TCP stream
\  the rx socket pointer register is updated. 
\ The command RECV is sent to the W5100
: tcp-read ( uSocket c-Adr, uBytes -- uBytes' )
  #3 w51-push 0 #14 TRUE w51-do-cmd net_data @ ;
  
\ write as many bytes to the tcp stream as will fit the TX buffer
\ The tx socket pointer register is updated. 
\ The command SEND is sent to the W5100
: tcp-write ( uSocket c-Adr, uBytes -- uBytes' )
  \ no socket registers are updated.
  #3 w51-push 0 #16 TRUE w51-do-cmd net_data @ ;

\ repeat tcp-write until all bytes written or disconnected
: tcp-write ( uSocket uDataAdr uNumBytes -- uBytesRemaining )
  BEGIN
    ROT DUP tcp-connected
    ( uDataAdr uNumBytes uSocket fCon )
    2SWAP DUP 0>
    ( uSocket fCon DataAdr uNumBytes fData )
    >R ROT R> AND 
    ( uSocket DataAdr uNumBytes fWrite )
  WHILE
    2>R
    ( uSocket )
    DUP 2R@ tcp-write 2R>
    ( uSocket uBytesTX uDataAdr uNumBytes )
    ROT DUP ROT SWAP
    ( uSocket uDataAdr uBytesTX uNumBytes uBytesTX )
    - ROT ROT + SWAP
    ( uSocket uDataAdr' uBytesRemain )
  REPEAT
  NIP NIP ;

\ Tell the W5100 to listen on the particular socket
: tcp-listen ( uSocket -- )
  DUP skt-status c_SOCK_INIT = IF
    c_CMD_LISTEN SWAP skt-cr!
  ELSE
    DROP
  THEN ;

\ Tell the W5100 to connect to destination ip and port
: tcp-connect ( uSocket -- )
  DUP skt-status c_SOCK_INIT = IF
    ( set the correct subnet mask )
    net_mask w51-mask!
    c_CMD_CONNECT SWAP skt-cr!
    ( set the blank subnet mask )
    net_blankmask w51-mask!
  ELSE
    DROP
  THEN ;

( end of tcp support functions )


\ set network and MAC addresses from defaults
: w51-defaults 
  net_mac  w51-mac! 
  net_blankmask w51-mask!
  net_ip   w51-ip! 
  net_gw   w51-gw! ;

\ stop the network and stop COG
: net-stop ( -- )
  net_cog IF
    \ disconnect all tcp connections
    net_sockets 0 DO 
      I skt-status DUP #23 = SWAP #28 = OR IF 
        I tcp-disconnect
      THEN 
    LOOP
    \ all tcp sockets have been told to close
    #1000 MS
    \ close all sockets
    net_sockets 0 DO I skt-close LOOP
    \ stop the cog
    net_cog COG-STOP 0 TO net_cog 
  THEN ;

\ reset W5100 and load default ip, mask and gateway
: net-start ( -- )
  net-stop
  \ clear the network cog command variable
  0. net_command Q! 
  \ start the W5100 control COG
  net_command net-blk COG-LOAD TO net_cog 
  \ reset the w5100 and set IP, MASK, GW and MAC
  w51-resethw w51-defaults ;

\ : tt $1F2 COG@ HEX UD. DECIMAL ;