\ Code to control list of user names and passwords for basic web authentication.
\ User names and passwords are stored in a single block.
\ Each 64 byte block contains one user name:password string
\ followed by a user id (1..99).

MARKER TEST

#83 VALUE pwd_blk       \ block index that contains user nams and passwords
#15 CONSTANT pwd_max
#99 CONSTANT pwd_admin_level
#7 CONSTANT pwd_min_admin
CHAR : CONSTANT COLON
CHAR ; CONSTANT SEMICOL


\ admin account user name
: admin_str ( -- cadrAdminStr uLenAdmin )
  S" admin" ;

\ clear user info at specified index (1..pwd_max)
: pwd-delete ( uUsrIdx -- fWasDeleted )
  \ not allowed to clear admin settings (allways index 0)
  ?DUP IF
    DUP pwd_max <= IF
      \ multiply by 64 to get line start
      #6 LSHIFT
      \ add to start of block buffer
      pwd_blk BLOCK +
      \ blank user info at selected slot
      #64 BLANK
      UPDATE FLUSH
      TRUE
    ELSE
      DROP
      FALSE
    THEN
  ELSE
    FALSE    
  THEN ;

: pwd-level ( uUserIndex -- uUserLevel )
  DUP pwd_max <= IF
    \ multiply by 64 to get line start
    #6 LSHIFT
    \ add to start of block buffer
    pwd_blk BLOCK +
    \ scan for a space char
    ( cadrUserInfo )
    \ skip user name and password ( pwd could be blank )
    #64 COLON SCAN 1 /STRING COLON SCAN
    ( cadrUserInfo' uLenRemaing )
    \ find start of number
    COLON SKIP  
    ( cadrUserLevel' uLenRemaing )
    \ find end of number
    2DUP BL SCAN
    ( cadrUserLevel' uLenRemaing cadrUserLevel'' uLenRemaing' )
    NIP -
    ( cadrUserLevel' uLenLevel )
    0. 2SWAP
    ( dNumber cadrUserLevel' uLenLevel )
    >NUMBER
    ( dNumber cadrUserLevel'' uLenLeft )
    2DROP D>S
  ELSE
    DROP
    0
  THEN ;  


\ set user info at specified index (1..pwd_max)
: pwd-set ( uLevel cadrUserStr uLenUserStr uUsrIdx -- )
  DUP pwd_max <= IF
    \ multiply by 64 to get line start
    #6 LSHIFT
    \ add to start of block buffer
    pwd_blk BLOCK + DUP
    \ blank user info at selected slot
    #64 BLANK
    ( uLevel cadrUserStr uLenUserStr cadrUsrIdx )
    \ calc user level string address in block buffer
    2DUP + >R  \ save address to write user level
    \ copy user string to buffer
    ( uLevel cadrUserStr uLenUserStr cadrUsrIdx )
    SWAP MOVE
    ( uLevel )
    S>D <# #S COLON HOLD #>
    ( cadrLevel uLenLevel )
    R>
    ( cadrLevel uLenLevel cadrBlkLevel )
    SWAP MOVE
    UPDATE FLUSH
  ELSE
    2DROP 2DROP
  THEN ;  
 
\ find the first blank user entry, skip first entry always admin  
: pwd-fndblank ( -- FALSE | uBlankIndex TRUE)
  pwd_blk BLOCK
  0
  BEGIN
    ( cadrUserData uUserIndex)
    SWAP #64 +
    SWAP 1+
    DUP pwd_max <=
    ( cadrUserData' uUserIndex' fContinue)
  WHILE
    ( cadrUserData' uUserIndex' )
    \ skip over all blank chars report remaining
    OVER #64 BL SKIP NIP
    ( cadrUserData' uUserIndex' uRemaing)
    \ if line was all blank chars uRemaining will be zero
    0= IF
      ( cadrUserData' uUserIndex' )
      NIP TRUE
      EXIT
    THEN
  REPEAT
  ( cadrUserData' uUserIndex' )
  \ did not find blank line report FALSE
  2DROP FALSE ;
    
\ This word will append a colon char to the user name if it not already terminated
\ with a colon char.   
\ The buffer that holds the user name must have space for one extra char.
: pwd-colon+ ( cadrStr uLenStr -- cadrStr uLenStr' )
  2DUP 
  ( cadrStr uLenStr cadrStr uLenStr )
  + 1- DUP C@
  ( cadrStr uLenStr cadrStrEnd cLastChar )
  COLON <> IF
    ( cadrStr uLenStr cadrStrEnd )
    \ append colon char
    1+ COLON SWAP C!
    \ increase length of string on return stack
    ( cadrStr uLenStr )
    1+ 
    ( cadrUser uLenUser' )
  ELSE
    ( cadrStr uLenStr cadrStrEnd )
    \ last char is a full colon
    DROP
    ( cadrStr uLenStr )
  THEN ; 
    
\ allocate a variable to build user:password string   
VARIABLE pwd_temp #64 ALLOT
0 VALUE pwd_error

\ Find the specified user entry matching on user name.
: pwd-fndusr ( cadrUser uLenUser -- FALSE | uUserIndex TRUE)
  \ copy user name to temp buffer
  pwd_temp >COUNTED pwd_temp COUNT
  \ make sure user name ends in a full colon ':' char
  pwd-colon+  \ append a colon to search string if required
  2>R  \ save user string address and length to return stack
  \ setup block line address and index to one line before start 
  pwd_blk BLOCK #64 -
  -1
  \ loop until we find user name or check last line in block
  BEGIN
    ( cadrUserData uUserIndex)
    SWAP #64 +
    SWAP 1+
    DUP pwd_max <=
    ( cadrUserData' uUserIndex' fContinue )
  WHILE
    ( cadrUserData' uUserIndex'  R: cadrUser uLenUser )
    \ see if user name matches
    OVER 2R@ S=
    0= IF
      ( cadrUserData' uUserIndex'  R: cadrUser uLenUser )
      NIP TRUE
      2RDROP
      ( uUserIndex' TRUE )
      EXIT
    THEN
  REPEAT
  ( cadrUserData' uUserIndex'  R: cadrUser uLenUser )
  \ did not find blank line report FALSE
  2RDROP 2DROP FALSE ;

\ Add or update the specified user block line entry.
: pwd-addusr ( uUserLevel cadrPwd uLenPwd cadrUser uLenUser uLineIndex --  uLineIndex )
  >R  \ save the line index
  \ clear temp buffer
  pwd_temp #64 BLANK
  \ copy the user name to temp buffer
  ( uUserLevel cadrPwd uLenPwd cadrUser uLenUser  R: uLineIndex )
  2DUP pwd_temp SWAP MOVE NIP
  ( uUserLevel cadrPwd uLenPwd uLenUser  R: uLineIndex )
  \ append a colon to user name if string does not end with one
  pwd_temp SWAP pwd-colon+ 
  ( uUserLevel cadrPwd uLenPwd cadrTempStr uLenUser'  R: uLineIndex )
  \ save total string length and calc address to copy the password string
  2OVER NIP 
  ( uUserLevel cadrPwd uLenPwd cadrTempStr uLenUser' uLenPwd  R: uLineIndex )
  OVER + >R 
  ( uUserLevel cadrPwd uLenPwd cadrTempStr uLenUser'  R: uLineIndex uTotalLen )
  +
  ( uUserLevel cadrPwd uLenPwd cadrTempStr'  R: uLineIndex uTotalLen )
  \ copy password to temp buffer
  SWAP MOVE
  ( uUserLevel  R: uLineIndex uTotalLen )
  \ update the line index in the user database
  pwd_temp R> R@
  ( uLevel cadrTempStr uTotalLen uLineIndex )
  pwd-set R> ; 
  
\ DROP 5 cell from data stack  
: 5DROP ( u0 u1 u2 u3 u4 -- )
  2DROP 2DROP DROP ;
   
\ see if the user name is admin   
: pwd-isadmin ( cadrUser uUserLen -- fIsAdmin )
  admin_str
  ( cadrUser uUserLen cadrAdmin uAdminLen )
  ROT OVER
  ( cadrUser cadrAdmin uAdminLen uUserLen uAdminLen )
  = IF
    ( cadrUser cadrAdmin uAdminLen )
    S= 0=
  ELSE
    ( cadrUser cadrAdmin uAdminLen )
    2DROP DROP FALSE
  THEN ; 

   
\ Add or update the specified user entry.
\ This word overides the defined word pwd-addusr to add error checking
: pwd-addusr ( uUserLevel cadrPwd uLenPwd cadrUser uLenUser -- FALSE | uUserIndex TRUE)
  \ if the user name is blank do not add
  DUP IF
    \ user name is not blank, make sure there are no colon chars in name
    2DUP COLON SCAN NIP 0= IF
      \ user name is ok, now check password
      ( uUserLevel cadrPwd uLenPwd cadrUser uLenUser )
      2OVER COLON SCAN NIP 0= IF
        \ password has no colon chars, ok to use
        \ see if we are manipulating admin
        ( uUserLevel cadrPwd uLenPwd cadrUser uLenUser )
        2DUP pwd-isadmin NOT IF
          0 TO pwd_error
          DUP 2OVER DROP
          ( uUserLevel cadrPwd uLenPwd cadrUser uLenUser uLenUser uLenPwd )
          + #60 <= IF
            \ combined string fits
            ( uUserLevel cadrPwd uLenPwd cadrUser uLenUser )
            \ make sure user level is in range
            2>R 2>R  \ save user name and password to return stack
            DUP 1 pwd_admin_level WITHIN IF
              \ user level is in range
              ( uUserLevel  R: cadrUser uLenUser cadrPwd uLenPwd )
              2R> 2R>  \ restore user name and password from return stack
              ( uUserLevel cadrPwd uLenPwd cadrUser uLenUser )
              2DUP pwd-fndusr IF
                \ user already exists so update
                ( uUserLevel cadrPwd uLenPwd cadrUser uLenUser uLineIndex )
                pwd-addusr
                TRUE
              ELSE
                \ user is not in database find a blank line to add new user info
                ( uUserLevel cadrPwd uLenPwd cadrUser uLenUser )
                pwd-fndblank IF
                  ( uUserLevel cadrPwd uLenPwd cadrUser uLenUser uLineIndex )
                  pwd-addusr
                  TRUE
                ELSE
                  \ no space to add a new user
                  ( uUserLevel cadrPwd uLenPwd cadrUser uLenUser )
                  5DROP
                  #2 TO pwd_error
                  FALSE
                THEN
              THEN
            ELSE
              \ user level out of range
              ( uUserLevel  R: cadrUser uLenUser cadrPwd uLenPwd )
              #6 TO pwd_error
              2RDROP 2RDROP DROP
              FALSE 
            THEN
          ELSE
            ( uUserLevel cadrPwd uLenPwd cadrUser uLenUser )
            \ user name plus password are to long
            #1 TO pwd_error
            5DROP
            FALSE
          THEN  
        ELSE
          \ user name is admin, we only can set password
          ( uUserLevel cadrPwd uLenPwd cadrUser uLenUser )
          2DROP
          ( uUserLevel cadrPwd uLenPwd )
          \ save password to return stack and set level to 
          2>R
          DROP pwd_admin_level
          2R>
          ( uAdminLevel cadrPwd uLenPwd )
          \ make sure the password is long enough
          DUP pwd_min_admin 1- U> IF
            ( uAdminLevel cadrPwd uLenPwd )
            admin_str 0
            pwd-addusr
            TRUE
          ELSE
            ( uAdminLevel cadrPwd uLenPwd )
            \ admin pasword not long enough
            #5 TO pwd_error
            2DROP DROP
            FALSE 
          THEN  
        THEN
      ELSE
        \ password contain colon character
        ( uUserLevel cadrPwd uLenPwd cadrUser uLenUser )
        #7 TO pwd_error
        5DROP
        FALSE
      THEN    
    ELSE
      ( uUserLevel cadrPwd uLenPwd cadrUser uLenUser )
      \ user name contains colon chars
      #4 TO pwd_error
      5DROP
      FALSE
    THEN
  ELSE
    ( uUserLevel cadrPwd uLenPwd cadrUser uLenUser )
    #3 TO pwd_error
    5DROP
    FALSE
  THEN ;  
  
\ Remove a user from the database
: pwd-delusr ( cadrUser uLenUser -- fWasDeleted )
  pwd-fndusr IF
    pwd-delete DUP IF
      0
    ELSE
      #9 
    THEN
    TO pwd_error
  ELSE
    #8 TO pwd_error
    FALSE 
  THEN ;

\ change the password for an existing user
: pwd-setpwd ( cadrPwd uLenPwd cadrUser uLenUser -- fWasUpdated )
  \ get the index of the specified user
  2DUP pwd-fndusr IF
    ( cadrPwd uLenPwd cadrUser uLenUser uDbIndex )
    \ read the existing user level value
    pwd-level
    ( cadrPwd uLenPwd cadrUser uLenUser uUserLevel )
    >L 2>R 2>R
    ( -- )
    L> 2R> 2R>
    ( uUserLevel cadrPwd uLenPwd cadrUser uLenUser )
    pwd-addusr DUP IF NIP THEN
  ELSE
    ( cadrPwd uLenPwd cadrUser uLenUser )
    2DROP 2DROP
    #8 TO pwd_error
    FALSE 
  THEN ;    

\ change the authorization level for an existing user
: pwd-setlevel ( uAuthLevel cadrUser uLenUser -- fWasUpdated )
  \ get the index of the specified user
  pwd-fndusr IF
    ( uAuthLevel uDbIndex )
    \ load database and calc address of DB entry
    #6 LSHIFT pwd_blk BLOCK +
    ( uAuthLevel cadrDbEntry )
    \ skip to start of auth level field
    #64 COLON SCAN 1 /STRING COLON SCAN DROP
    \ append level string to current position
    ( uAuthLevel cadrBuff )
    \ clear old number and colon from record buffer
    DUP #3 BLANK
    \ save field buffer address to return stack
    >R
    ( uAuthLevel  R: cadrBuff )
    \ convert number to a string
    S>D <# #S COLON HOLD #>
    ( cadrLevel uLenLevel  R: cadrBuff )
    R> SWAP
    ( cadrLevel cadrBuff uLenLevel )
    \ copy new level string to record buffer
    MOVE
    UPDATE FLUSH
    TRUE  
  ELSE
    ( cadrPwd uLenPwd cadrUser uLenUser )
    DROP
    #8 TO pwd_error
    FALSE 
  THEN ;    
  
\ clear user password database, then add admin account with default password  
: pwd-clearall ( -- )
  \ clear all passwords
  pwd_blk BUFFER #1024 BLANK
  \ save cleared user info
  UPDATE FLUSH
  \ add admin account ;
  pwd_admin_level admin_str S" password"
  pwd-addusr IF DROP THEN DROP ;
      
\ print last user update error  
: pwd-error. ( -- )
  CASE pwd_error
    #1 OF ." user info is to long, max 60 characters"  ENDOF
    #2 OF ." user database is full" ENDOF
    #3 OF ." user name can not be blank" ENDOF
    #4 OF ." user name can not contain colon ':' characters" ENDOF
    #5 OF ." admin password must be at least " pwd_min_admin . ." characters" ENDOF
    #6 OF ." user level out of range 1.." pwd_admin_level 1- . ENDOF 
    #7 OF ." password can not contain colon ':' characters" ENDOF
    #8 OF ." user not found in database" ENDOF
    #9 OF ." admin account can not be deleted" ENDOF
    \ otherwise
    pwd_error IF ." Unknow error " . ELSE ." Completed " THEN
  ENDCASE ;     
       
  