\ Basic Authentication Module For WEB Demo
\ requires WebServer and Base64Dec

\ forth block index where user database is stored
CHAR : CONSTANT COLON
CHAR ; CONSTANT SEMI
#83 VALUE pwd_blk       \ block index that contains user names and passwords
#15 CONSTANT pwd_max    \ max number of user entries


\ This word will append a colon char to the user name if it not already terminated
\ with a colon char.   
\ The buffer that holds the user name must have space for one extra char.
: str-colon+ ( cadrStr uLenStr -- cadrStr uLenStr' )
  2DUP 
  ( cadrStr uLenStr cadrStr uLenStr )
  + 1- DUP C@
  ( cadrStr uLenStr cadrStrEnd cLastChar )
  COLON <> IF
    ( cadrStr uLenStr cadrStrEnd )
    \ append colon char
    1+ COLON SWAP C!
    \ increase length of string on return stack
    ( cadrStr uLenStr )
    1+ 
    ( cadrUser uLenUser' )
  ELSE
    ( cadrStr uLenStr cadrStrEnd )
    \ last char is a full colon
    DROP
    ( cadrStr uLenStr )
  THEN ; 
         
\ save the USER name to the global buffer             
: ba-savname ( cadrCred uCredLen -- )
  2DUP COLON SCAN NIP -
  #32 MIN  
  user_name >COUNTED ; 

                 
\ scan the user database for a match to the credentials
\ if a match is found return the user's authentication level (1..99)
\ or return 0 if there is no match
: ba-val ( cadrCred uCredLen -- uUserLevel )
  \ clear the user name buffer set length to zero
  0 user_name !                
  ?DUP IF
    str-colon+ 
    ( cadrCred uCredLen' )
    pwd_blk BLOCK #64
    ( cadrCred uCredLen' cadrUserInfo uLenUserInfo )
    2SWAP
    ( cadrUserInfo uLenUserInfo cadrCred uCredLen' )
    \ scan all the user info entries looking for a match
    pwd_max 1+ 0 DO
      ( cadrUserInfo uLenUserInfo cadrCred uCredLen' )
      2OVER 2OVER 
      ( cadrUserInfo uLenUserInfo cadrCred uCredLen' cadrUserInfo uLenUserInfo cadrCred uCredLen' )
      str-starts IF
        ( cadrUserInfo uLenUserInfo cadrCred uCredLen' )
        \ save the user name in to global buffer
        ba-savname
        ( cadrUserInfo uLenUserInfo )
        COLON SCAN 1 /STRING
        COLON SCAN 1 /STRING
        ( cadrUserInfo' uLenUserInfo' )
        \ find end of user authentication level string
        2DUP BL SCAN NIP
        ( cadrUserInfo' uLenUserInfo' uRemaingAfterNum )
        -
        0. 2SWAP
        ( dLevelNum cadrUserInfo' uLenUserLevel' )
        >NUMBER 2DROP D>S 
        ( uUserLevel )
        UNLOOP
        EXIT
      THEN
      ( cadrUserInfo uLenUserInfo cadrCred uCredLen' )
      2SWAP DROP #64 + #64 2SWAP  
      ( cadrUserInfo' uLenUserInfo cadrCred uCredLen' )
    LOOP
    ( cadrUserInfo uLenUserInfo cadrCred uCredLen' )
    2DROP 2DROP 0
  ELSE
    ( uCredLen' )
    DROP 0  
  THEN ; 

\ See if request header is Authentication Header
\ return the base 64 encoded user:password string
: bauth-isauth ( cadr uLen -- FALSE | cAdrCreds uLenCreds TRUE )
  2DUP S" Authorization: " str-starts
  IF
   \ make sure we have basic authorization
    ( cadr uLen )
    \ move to first space
    BL SCAN
    ( cadr' uLen' )
    \ skip to next non space
    BL SKIP
    ( cadr'' uLen'' )
    2DUP S" Basic " str-starts
    IF
      \ remainder of string is user:password in base 64
      \ move to first space after Basic
      BL SCAN
      \ move to start of credentials
      BL SKIP
      \ find end of credentials
      2DUP BL SCAN
      ( cadr'' uLen'' c-adrEnd uLenEnd )
      NIP
      ( cadr'' uLen'' uLenEnd' )
      -
      ( cAdrCreds uLenCreds )
      OVER SWAP
      \ decode string over itself, saves need for second buffer
      ( cAdrCreds cAdrCreds uLenCreds )
      base64dec  
      TRUE
    ELSE
      ( cadr'' uLen'' )
      S" No auth basic" debug-writeln
      2DROP
      FALSE
    THEN
  ELSE
    ( cadr uLen )
    2DROP
    FALSE
  THEN ;
    
\ This function is called for each http request header 
\ after the base server proccesses them.
\ The header is a counted string in the variable WEB_TMP.
: WEB-AUTHHDR ( -- )
  \ see if we have an authorization header, if not just exit.
  WEB_TMP COUNT 
  ( cadr uLen )
  bauth-isauth IF
    \ attempt to validate supplice credentials, save result in to a global value
    \ The value web_auth is declared in the "WebServer.forth" file.
    ( cAdrCreds uLenCreds )
    2DUP ba-val 
    ( cAdrCreds uLenCreds uUserLevel )
    DUP TO web_auth
    \ display the authentication level and user credentials if debug is on 
    ( cAdrCreds uLenCreds uUserLevel )
    debug-writenum 
    ( cAdrCreds uLenCreds )
    debug-writeln
  THEN ;
  
\ WEB-HDR is a deferred word that can be hooked to see every request header
' WEB-AUTHHDR IS WEB-HDR

\ authentication request header    
: bauth-genhdr ( -- )
  ." WWW-Authenticate: Basic realm=Mv4th_Web_Demo" CR ;

\ send http responce that authentication is required  
: rsp-noauth  ( -- ) rsp-htp ." 401 Not Authorized" CR ;

\ Standard 401 error, not authorized
: rsp-auth ( -- )
  \ send responce code for auth required
  rsp-noauth rsp-base
  \ indicate text message is plain text
  hdr-text
  \ add in authentication request header
  bauth-genhdr 
  \ close off headers, start of message follows
  rsp-eoh
  \ show server info 
  hdr-svr 
  \ show server error header 
  rsp-noauth ; 
  