\ Web demo html page generation and transfer functions
\ This page is dynamicaly loaded by the web demo.

\ variables to track html tag substitution
FALSE VALUE was_tag
FALSE VALUE in_value

\ The demo user database assigns an authentication level to each user.
\ The authentication level is a value from one to ninty nine (1..99).
\ The admin account always has a level of 99.
\ If the request is not authenticated the level will be zero.
\ The current authentication level can be read from the value web_auth.
\ The value web_auth is created by the WebServer forth code at startup. 

#32 CONSTANT tag_size
VARIABLE tag_buffer tag_size CELL+ ALLOT 
tag_buffer CONSTANT tag_chars
tag_buffer 1+ CONSTANT tag_buffer

: int-tagbuff ( -- )
  0 tag_chars C!
  tag_buffer tag_size ERASE
  S" FEX-" DUP tag_chars C!
  ( cadr uCount )
  tag_buffer SWAP CMOVE ;

: add_tagbuff ( cChar -- )
  tag_chars C@ DUP tag_size < IF
    ( cChar uTagChars )
    tag_buffer + C!
    tag_chars DUP C@ 1 + SWAP C!
  ELSE
    \ No room in buffer
    2DROP
  THEN ;  
  
: int-parser ( -- )
  int-tagbuff
  FALSE TO was_tag
  FALSE TO in_value ;
  
\ ******* TAG VALUE PARSER FUNCTION *******
\ tag buffer has value to parse
: parse-value ( cadr uCount -- )
  2DUP debug-writeln
  DROP 1- FIND 0=  
  IF
    \ did not find value function
    DROP
    ." ?????"
  ELSE  
    EXECUTE
  THEN ;


: parse&send ( c-adrData uBytesToSend -- )
  0 ?DO
    ( cadrData )
    COUNT
    ( cadrData' cChar )
    in_value IF
      DUP [CHAR] > = IF
       ( cadrData' cChar )
        DROP
       ( cadrData' )
        tag_buffer tag_chars C@
        parse-value
        FALSE TO in_value
        FALSE TO was_tag
      ELSE
        add_tagbuff
       ( cadrData' )
      THEN
    ELSE
      ( cadrData' cChar )
      DUP [CHAR] # = was_tag AND
      IF
        ( cadrData' cChar )
        \ start of a replacement tag
        TRUE TO in_value
        int-tagbuff
        DROP 
      ELSE
        ( cadrData' cChar )
        was_tag IF
          [CHAR] < EMIT
        THEN  
        DUP [CHAR] < = DUP TO was_tag
        ( cadrData' cChar fTagStart)
        NOT IF
          EMIT
        ELSE
          DROP  
        THEN  
      THEN
    THEN  
  LOOP
  ( cadrData' )
  DROP ;
  
DEFER out-http  ( c-adrData uBytesToSend -- )
  
' parse&send IS out-http

: send-http ( uBytes uBlk )
  \ make sure NetConsole TX buffer is clear
  nc-flush
  ( uBytes uBlk )
  BEGIN
    OVER
  WHILE
    ( uBytes uBlk )
    OVER DUP #1024 > IF 
      DROP #1024 
    THEN
    ( uBytes uBlk uBytesToSend )
    OVER
    ( uBytes uBlk uBytesToSend uBlk )
    BLOCK SWAP
    ( uBytes uBlk c-adrBlk uBytesToSend )
    DUP >R
    ( uBytes uBlk c-adrBlk uBytesToSend ) ( R: uBytesToSend )
    nc_socket ROT ROT
    ( uBytes uBlk uSocket c-adrBlk uBytesToSend ) ( R: uBytesToSend )
    \ write data to TCP client 
    out-http DROP
    ( uBytes uBlk ) ( R: uBytesToSend )
    1+ SWAP R> - SWAP
    ( uBytes' uBlk' )
  REPEAT
  2DROP ;  

\ send an html page back to the client.
\ if fParseDocument is false the document is sent with no changes.
\ if fParseDocument is true the document is parsed for field tags of the form; 
\ <#xxxxx> where "xxxxx" is the name of a field function to be called.
\
\ All field tags map to field functions that have the prefix "FEX-" 
\ followed by the field tag name.
\ As an example placing the field tag "<#user-name>" in the html file
\ will cause the forth word "FEX-user-name" to be called. The field function
\ can insert content in to the html document by outputting to the console device.
\ The content output by the field function will replace the field tag in the html.  
: send-html ( uBytes uFirstBlk fParseDocument -- )
  IF
    ['] parse&send IS out-http
    int-parser
    -1 rsp-html
  ELSE
    ['] tcp-write IS out-http
    OVER rsp-html
  THEN
  ( uBytes uFirstBlk )
  send-http ;

\ return a png image file to client
: send-png ( uBytes uFirstBlk -- )
  ['] tcp-write IS out-http
  OVER rsp-png
  ( uBytes uFirstBlk )
  send-http ;

\ Load HTML demo main page
: WEX-demo
  UL_GUEST chkauth IF
    #2048 #90 TRUE send-html
  THEN ;

\ send page to update date and time
: WEX-date ( -- )
  UL_ADMIN chkauth IF
    \ only admin can change the date and time
    #1263 #92 TRUE send-html
  THEN ;
  
\ send page to update user password
: WEX-setpwd ( -- )
  UL_USER chkauth IF
    #2025 #100 TRUE send-html
  THEN ;
  
  \ send site page icon
: WEX-favicon.ico
  #4213 #94 send-png ;

\ change the password of the currently authenticated user  
: WEX-updpwd ( -- )
  UL_USER chkauth IF
    FALSE \ fParamsValid = false
    ( fParamsValid ) 
    S" pwd1=" web-valstr IF
      ( fParamsValid cadrpwd1 uLenpwd1 )
      S" pwd2=" web-valstr IF
        ( fParamsValid cadrpwd1 uLenpwd1 cadrpwd2 uLenpwd2 )
        \ make sure strings are same length
        ROT OVER = IF
          \ same length see if they match
          ( fParamsValid cadrpwd1 cadrpwd2 uLenpwd2 )
          S= 
          ( fParamsValid uMatchRes )
          0= NIP
          ( fParamsValid' )
        ELSE
          \ password and confirm password do not match
          ( fParamsValid cadrpwd1 cadrpwd2 uLenpwd2 )
          DROP 2DROP
          ( fParamsValid )
        THEN
      ELSE
        ( fParamsValid cadrpwd1 uLenpwd1 )
        2DROP
        ( fParamsValid )
      THEN
    THEN      
    ( fParamsValid )
    IF
      \ params are ok, load code to update user database
      #51 #54 THRU
      S" pwd1=" web-valstr IF
        ( cadrPwd uLenPwd )
        user_name COUNT
        ( cadrPwd uLenPwd cadrUser uLenUser )
        \ late bind to set password function
        C" pwd-setpwd" FIND IF 
          ( cadrPwd uLenPwd cadrUser uLenUser xt )
          EXECUTE 
          ( fPasswordSet )
          IF
            0 rsp-text
            ." Password changed for user: " user_name COUNT TYPE CR
          ELSE
            DROP
            \ failed to set password, report system error
            0 rsp-text
            ." Password could not be changed, "
            \ late bind to set password error function
            C" pwd-error." FIND IF 
              EXECUTE 
            ELSE 
              DROP 
              ." unknown Error." 
            THEN 
            CR
          THEN
        ELSE
          ( cadrPwd uLenPwd cadrUser uLenUser )
          2DROP 2DROP
          0 rsp-text
          ." Internal error set password function not found."
        THEN
      ELSE
        \ no password param
        0 rsp-text
        ." New password required" CR
      THEN  
    ELSE
      \ paramaters are not valid
      0 rsp-text
      ." Password and confirmed password do not match." CR
    THEN
  THEN ;  
  
\ functions to supply field tag content in the web page when parsed.
\ the words are called by the parse&send word using late binding
  
: FEX-year ( -- )
  DATE 1 .R 2DROP ;
  
: FEX-month ( -- )
  DATE DROP 1 .R DROP ;
  
: FEX-day ( -- )
  DATE 2DROP 1 .R ;
    
: FEX-hour ( -- )
  TIME 1 .R 2DROP ;
  
: FEX-minute ( -- )
  TIME DROP 1 .R DROP ;
  
: FEX-second ( -- )
  TIME 2DROP 1 .R ;

: FEX-getdate ( -- )
  .GMT ;
  
: FEX-user ( -- )
  user_name COUNT TYPE ;

\ if we are logged in as a user level account (level >= 10)
\  return a CSS display type of "inline" if not return "none"
: FEX-USER ( -- )
  web_auth UL_USER >= IF
    ." inline"
  ELSE
    ." none"
  THEN ;
       
\ if we are logged in as the admin level account (level = 99)
\  return a CSS display type of "inline" if not return "none"
: FEX-ADMIN ( -- )
  web_auth UL_ADMIN = IF
    ." inline"
  ELSE
    ." none"
  THEN ;
       
