'MV4th16 Forth basic I/O EEPROM, I2C, COM1 Stuff
'
CON
  ' Max number of forth blocks in a Forth block file
  c_MaxFileBlocks = 65520

  ' Fast Find max string size
  c_MaxSearchLen = 32
  c_ImmediateFlag = 1

  ' Number of I/O COGS used (including this one)
  c_IO_CogCount = 3             ' This one, serial I/O, SD card SPI

  'I2C EEprom stuf
  'The prototype board has a 64K serial eeprom. The Prop chip only uses the first 32K leaving 32K for your program to store data.
  '
'  c_I2C_SCLBit  = 28                                  'default I2C clock signal pin number
'  c_I2C_SDABit  = 29                                  'default I2C data signal pin number

  'Forth Command constants
  c_ForthCmdCount    = 26
  
  'Forth com port fifo sizes
  c_ComTxFifoSize = Com#c_ComTxFifoSize
  c_ComRxFifoSize = Com#c_ComRxFifoSize
  c_bit_ticks_adr = Com#bit_ticks_adr
  'Fifo buffers
  c_ComRxFifoAdr  = Com#c_ComRxFifoAdr
  c_ComTxFifoAdr  = Com#c_ComTxFifoAdr

  'Start address of SPI data area, runs from c_ComDataStartAdr - sizeof(SPI_Data) to c_ComDataStartAdr
  c_SpiDataStartAdr = Com#c_ComDataStartAdr - 16
  
  'Start address of I/O data area, runs from c_ComDataStartAdr to $7FFF
  c_IODataStartAdr = c_SpiDataStartAdr

  ' Forth I/O command numbers used by this object for console I/O at startup
  cmd_COM_OUT     =  1    ' ( c -- )                  Write char to console, c = 8 bit char to send to console.
  
  cmd_COM_IN      =  2    ' (  -- c )                 Read char from console, c = 8 bit char read from console.
  
  cmd_COM_CAN_IN  =  3    ' (  -- f )                 Test if console has data to read, f = 0 no char ready
'                                                     from console, f= -1 char abailable from console.

  cmd_COM_CAN_OUT =  4    ' (  -- f )                 Test if console can write data, f = 0 no char can be set to
'                                                     the console, f= -1 char can be written to console.

  cmd_COM_CLEAR =  5    ' ( -- )                    Request the console input buffer be cleared

OBJ
  PinCfg : "MV4th16PinCfg"      ' Default propeller chip pin settings (I2C, Async serial, SD card...)
  Com: "ForthBaseSerial"
  SdFat: "fsrw"
  
  
VAR
  'com port changable settings
  word  settings_adr
  'rx fifo settings
  word  rx_head_adr             
  word  rx_tail_adr
  word  rx_buf_size
  word  rx_buf_adr
  'tx fifo settings
  word  tx_head_adr             
  word  tx_tail_adr
  word  tx_buf_size
  word  tx_buf_adr
  'com port hardware settings
  word  rx_pin
  word  tx_pin
  'End of com port settings

  long SystemCmdAdr   'address used to communicate with forth I/O COG
  
  long StrToNumValid
  byte Temp[1024]
  
PUB Start(baudrate, SystemCommandAdr) | cog_idx
  SystemCmdAdr:= SystemCommandAdr
  long[SystemCommandAdr] := 0
  'Start serial port, 2 wire version
  Com.Start(PinCfg#c_Pin_COM_RX, PinCfg#c_Pin_COM_TX, PinCfg#c_ComTxMode, baudrate, @settings_adr)
  ComRxHeadAdr := rx_head_adr
  ComRxTailAdr := rx_tail_adr
  ComRxBufferSize := rx_buf_size
  ComRxBufferAdr := rx_buf_adr
  ComTxHeadAdr := tx_head_adr
  ComTxTailAdr := tx_tail_adr
  ComTxBufferSize := tx_buf_size
  ComTxBufferAdr := tx_buf_adr
  'set our I2C buss pins
  I2C_SDA_Bit:= |< PinCfg#c_Pin_I2C_SDA
  I2C_SCL_Bit:= |< PinCfg#c_Pin_I2C_SCL
  'Start our cog if we can 
  cog_idx:= cognew(@Entry, SystemCommandAdr)
  ' Try to mount the SD card, if no card mounted then system will reboot
  ' See if we can mount a valid SD card (no FAT12 only FAT16 or FAT32)

  long[c_SpiDataStartAdr]:= 0
  
  ActivateSdCard                           

  InitializeBlockFile

  cogstop(cog_idx)
  
  cog_idx:= cognew(@Entry, SystemCommandAdr)

PRI CharOut(Char) | Resp, Stack
  Stack := Char
  word[SystemCmdAdr + 2] := @Stack
  word[SystemCmdAdr] := cmd_COM_OUT
  repeat
     Resp := word[SystemCmdAdr]
  until (Resp == 0)

PRI CanCharIn : Ok |  Resp, Stack
  word[SystemCmdAdr + 2] := @Stack + 2
  word[SystemCmdAdr] := cmd_COM_CAN_IN
  repeat
     Resp := word[SystemCmdAdr]
  until (Resp == 0)
  Ok := Stack
  
PRI CharIn : Char |  Resp, Stack
  word[SystemCmdAdr + 2] := @Stack + 2
  word[SystemCmdAdr] := cmd_COM_IN
  repeat
     Resp := word[SystemCmdAdr]
  until (Resp == 0)
  Char := Stack
  
pub PutDec(value) | i, x
{Send integer value as decimal string.}
  x := value == NEGX                                                            'Check for max negative
  if value < 0
    value := ||(value+x)                                                        'If negative, make positive; adjust for max negative
    CharOut("-")                                                                   'and output sign

  i := 1_000_000_000                                                            'Initialize divisor

  repeat 10                                                                     'Loop for 10 digits
    if value => i                                                               
      CharOut(value / i + "0" + x*(i == 1))                                   'If non-zero digit, output digit; adjust for max negative
      value //= i                                                               'and digit from value
      result~~                                                                  'flag non-zero found
    elseif result or i == 1
      CharOut("0")                                                            'If zero digit (or only digit) output it
    i /= 10                                                                     'Update divisor


PRI Type(StrAdr)
      repeat strsize(StrAdr)
        CharOut(byte[StrAdr])
        StrAdr += 1

PRI CR
    Type(string(13,10))

pub BoolPrompt(PromptAdr, MaxTimeMS, Default): bool | LastChar
{ Display a prompt that requests a boolean responce from the operator.
  Wait up to MaxTimeMS milli-seconds for the operatore to type Y or N.
  If we time-out the default value if returned }

  Type(PromptAdr)
  {see if we have a key}
  repeat until FALSE
    if CanCharIn
      LastChar:= CharIn
      if (LastChar == "Y") or (LastChar == "y")
        CharOut("Y")
        CR
        Return True
      elseif (LastChar == "N") or (LastChar == "n")
        CharOut("N")
        CR
        Return False 
    else
      If (MaxTimeMS > 0)
        waitcnt(cnt + clkfreq / 100)
        MaxTimeMS:= MaxTimeMS - 10
      else
        if Default
          CharOut("Y")
        else
          CharOut("N")
        CR
        Return Default



pub GetString(AdrString, MaxLen, NumbersOnly): CurLen | LastChar
' Read a null terminated string from the console

  CurLen:= 0
  LastChar:= 0
  bytefill(AdrString, 0, MaxLen + 1)
  repeat while (LastChar <> 13)
    if CanCharIn
      LastChar:= CharIn
      case LastChar
        48..57:
          if (CurLen < MaxLen)
            Byte[AdrString]:= LastChar
            AdrString += 1
            CurLen += 1
            CharOut(LastChar)
        33..47, 58..127:
          if (CurLen < MaxLen) and ! NumbersOnly
            Byte[AdrString]:= LastChar
            AdrString += 1
            CurLen += 1
            CharOut(LastChar)
        8:
          if (CurLen > 0)
            AdrString -= 1
            CurLen -= 1
            Byte[AdrString]:= 0
            CharOut(LastChar)
            CharOut(" ")
            CharOut(LastChar)
            
pub StrToNum(AdrString): Value | NextChar
' Convert a string into a number

  Value:= 0
  NextChar:= byte[AdrString]
  StrToNumValid:= (NextChar <> 0)
  repeat while (NextChar <> 0)
    AdrString += 1
    if (NextChar => "0") and (NextChar =< "9")
      ' Multiply Value by 10 then add offset of NextChar from "0"
      Value:= ((Value << 3) + (Value << 1)) + (NextChar - "0")
    else
      StrToNumValid:= FALSE 
      Return   
    NextChar:= byte[AdrString]
            
pub GetNumber(MaxLen): Value

  GetString(@Temp, MaxLen, TRUE)
  Value:= StrToNum(@Temp)
     
pub RebootSystem
  CR
  waitcnt(cnt+ 10_000_000)
  Reboot
    
pub PromptRebootSystem |  LastChar

  sdfat.unmount
  CR
  Type(string("Insert FAT16 or FAT32 formatted SD card. then press R to Reboot: ")) 
  repeat until (LastChar == "R")
    if CanCharIn
      LastChar:= CharIn
      case LastChar
        "R","r":
          RebootSystem

pub ForthBlockFileName
  return string("MV4TH16.BLK")
                                   
pub ActivateSdCard | LastChar, VolumeBlocksCount, CardReady
 'Try to mount the SD card, if no valid SD card mounted then system will reboot.
 ' A valid SD card must be formated as FAT16 or FAT32, no FAT12
  CardReady:= \sdfat.mount_explicit(c_SpiDataStartAdr, PinCfg#c_SPI_DO, PinCfg#c_SPI_CLK, PinCfg#c_SPI_DI, PinCfg#c_SPI_CS)
  if CardReady < 0                                      ' If CardReady  is less then 0 we have an error 
    sdfat.unmount
    Type(string("SD Card could not be accessed. Error Num:  "))
    PutDec(CardReady)
    PromptRebootSystem
  else
    ' SD card mounted, indicate on console
    VolumeBlocksCount:= sdfat.GetTotalVolumeBlocks
    Type(string("SD card FAT"))
    if sdfat.GetFileSystem == 1
      Type(string("16, "))
    else  
      Type(string("32, "))
    if VolumeBlocksCount =< 204800    { if card is less then 100Mbyte report in K bytes }
      PutDec(VolumeBlocksCount / 2)
      Type(string("K bytes"))
    else
      PutDec(VolumeBlocksCount / 2048)
      Type(string("M bytes"))
    CR

pub CheckForFiles: HasFiles
  ' See if the SD card is blank. Block files can only be created on a blank SD card.

  \sdfat.opendir
  HasFiles:= (\sdfat.nextfile(@Temp) => 0)
  sdfat.pclose


pub GetBlockFileSizeMB: FileMegBytes | VolumeBlockCount

  VolumeBlockCount:= sdfat.GetTotalVolumeBlocks / 2048
  ' Do not use all space on card, leave about 12% free
  VolumeBlockCount:= VolumeBlockCount - (VolumeBlockCount >> 3)
  if (VolumeBlockCount > 64)
    VolumeBlockCount:= 64
  FileMegBytes:= -1  
  if (VolumeBlockCount => 1)
    Type(string("New Forth Block File Size (1.."))
    PutDec(VolumeBlockCount)
    Type(string(") MBytes: "))
    FileMegBytes:= GetNumber(2)
    CR
    if StrToNumValid
      if (FileMegBytes < 1)
        FileMegBytes:= 1
      elseif (FileMegBytes > 64)
        FileMegBytes:= 64
    else
      FileMegBytes:= -1
  else
    Type(string("Not enough free space on SD card for Forth block file. "))
  CR
      
pub CreateForthBlockFile(MegBytes): OK | FileBlocks, SDCardBlocks, BlkIdx, StatusStep
  {create a new forth block file on the SD card, returns OK = 0 if all went well}
  {SD card must be blank and formated with FAT 16 or FAT 32}
  {SD card must be mounted before calling create}
  '
  ' Make sure the disk is blank
  if CheckForFiles
    ' Disk is not blank
    Type(string("SD card must be blank (no files/directories) to create Forth block file."))
    CR
    CR
    Return -222
  '
  ' Make sure requested file size is in range
  if (MegBytes < 1)
    MegBytes:= 1
  elseif (MegBytes > 64)
    MegBytes:= 64
  ' calc number of Forth blocks required (1024 bytes)
  FileBlocks:= MegBytes * 1024
  if (FileBlocks > c_MaxFileBlocks)
    FileBlocks:= c_MaxFileBlocks  
  ' calc number of SD card blocks required (512 bytes)
  SDCardBlocks:= FileBlocks * 2

  ' Try to create the file in the root of the SD card FAT
  ' SD card is already mounted see if we can create forth block file
  Ok:= \sdfat.popen(ForthBlockFileName, "w")      
  if Ok => 0
    Type(string("Creating Forth Block File..."))
    CR

    Type(string("Formating: "))
    StatusStep:= SDCardBlocks / 64
  ' Fill temp buffer with ANSI space char (32)
    bytefill(@Temp, 32, 512)
    repeat BlkIdx from 0 to SDCardBlocks - 1
      if (OK => 0)
        Ok:= \sdfat.pwrite(@Temp, 512)
        if (BlkIdx // StatusStep) == 0
          CharOut(".")
    if (OK => 0)
      Type(string(" Ok"))
    else  
      Type(string(" Failed."))
    CR  
    \sdfat.pclose
    
    OK:= \sdfat.popen(ForthBlockFileName, "r")
  else  
    Type(string("Failed to create Forth block file."))
    CR
  CR    
  if (OK => 0)
    OK:= 0
    
pub InitializeBlockFile: Ok | BlockFileOpen, RdRes, DataRegionStart

  ' SD card mounted, see if we have a forth block file, if not prompt to create one
  '
  if \sdfat.popen(ForthBlockFileName, "r")  < 0
    CR
    if BoolPrompt(string("File MV4TH16.BLK not found. Create new block file (Y/N)? "), 10000, FALSE)
      CR
      RdRes:= GetBlockFileSizeMB
      if (RdRes => 1)
        BlockFileOpen:= (CreateForthBlockFile(RdRes) == 0)
      else  
        BlockFileOpen:= FALSE      
    else
      CR  
      BlockFileOpen:= FALSE      
  else  
    BlockFileOpen:= TRUE
  '  
  ' Forth block file open see if it is valid
  '     
  if BlockFileOpen  
    ' Forth block file is open
    ' get file stats...
    ' Calculate the first block index of forth block file
    DataRegionStart:= sdfat.GetDataRegion
    ' initialize COG variables
    ' SPI command buffer address
    SD_SpiCmdBufAdr:= c_SpiDataStartAdr
    'First 512 byte SD block index of forth block file on SD card
    DataRegionStart:= sdfat.GetDataRegion
    SD_FileStartIdx:= DataRegionStart + sdfat.GetFirstFileBlock
    'Number of forth blocks available in forth block file
    SD_FileBlockCount:= sdfat.get_filesize / 1024
    ' Close the file
    \sdfat.pclose
    ' Report info to console device
    Type(string("Using MV4TH16.BLK "))
    PutDec(SD_FileBlockCount)
    Type(string(" blocks."))
    CR
    CR 
  else
    Type(string("File MV4TH16.BLK not found on SD card."))
    CR
    PromptRebootSystem
  Ok:= 0                                              
  
DAT
'
'***********************************
'* Assembly language I2C Code      *
'***********************************
'
                        org   0
                        'Initialize I2C hardware buss
                        'make sure SCL = 0
Entry                   'COG asm entry point
T1                      andn    outa, I2C_SCL_Bit       'SCL bit = 0
T2                      or      dira, I2C_SCL_Bit       'SCL dir = output, eeprom SCL = 0
                        'wait half bit time (> 1 usec will do)
T3                      mov     T1, #100
T4                      add     T1, cnt
T5                      waitcnt T1, #0
                        'make sure SDA = 1 at eeprom
T6                      andn    outa, I2C_SDA_Bit       'SDA output register bit = 0
I2C_Resp                andn    dira, I2C_SDA_Bit       'Set SDA as input = 1 at eeprom
                        'set I2C bitrate and reset I2C buss
I2C_MemAdr              mov     T2, OneMillion          'set I2C buss to 1Mbit
CopyCnt                 call    #SetI2cBitDelay     
                        'reset I2C buss
loop                    call    #F_CMD_I2C_RESET
                        'Process commands
SDBlockResp             jmp     #F_Do_Cmds

'Set the i2c bit delay based on the current system clock tick rate.
'T2 = desired I2C bit rate ie: 1_000_000 = 1 uSec, 400_000 = 2.5 usec.....
'min value for I2C_HALF_BIT_DELAY is 10 system clock ticks.
SetI2cBitDelay          rdlong  T1, #0                                          'read system frequency from long[0] in global memory
                        call    #DoDivU32                                       'Divide system clock frequence by requested bit rate
                        'T3 = system clock ticks per whole bit,
                        shr     T3, #1                  'we need half bit time 
                        cmp     T3, #45         wc
                if_nc   sub     T3, #40                 'compensate for delay function time and bit prep time
                if_c    mov     T3, #5                  'min delay of 5 or waitcnt will take @ 45 sec per bit (lockuo)
                        mov     I2C_HALF_BIT_DELAY, T3
SetI2cBitDelay_ret      ret                  
'
'See if we have a valid I2C command. If we do execute the command.
'The I2CCmdAdr contains the address of the long holding
'the I2C command to perform. The next three longs contain;
'EEprom address, Global memory address, and Byte count.
'Commands:
'        0 - no command
'        1 - EMIT write to serial device
'        2 - KEY read from serial device
'        3 - KEY? see if KEY will not block
'        4 - EMIT? see if EMIT will not block
'        5 - EMPTY-KEYS flush the serial rx buffer
'        6 - Copy EEprom to global memory
'        7 - Copy global memory to EEProm page
'        8 - Retrive block device block count
'        9 - Start new COG (auto select COD index)
'       10 - Initialize the selected COG
'       11 - Stop running COG by index
'       12 - Allocate a lock
'       13 - Release a lock
'       14 - FILL ( c-addr u c -- ) fill memory address c-adr with u characters c
'       15 - CMOVE ( c-addr1 c-addr2 u -- )
'       16 - CMOVE> ( c-addr1 c-addr2 u -- )
'       17 - BAUD ( d -- ) set serial port baud rate
' These commands require the "I2C Interface.forth" file be loaded
'       18 - I2C-RESET ( -- )  
'       19 - I2C-READ ( c-adr u -- )  
'       20 - I2C-WRITE ( c-adr u -- f )  
'       21 - I2C-START ( -- )
'       22 - I2C-STOP ( -- )
'       23 - I2C-SPPED ( dBitRate -- )
' End of commands that require that "I2C Interface.forth" be loaded
'       24 - REBOOT ( -- )
'
'
'Initialize our I2C serial buss.
'
'                        
'Perform I2C one bit delay (Delay > 0.5usec), max eeprom clock = 1mhz
'
'Set SCL to low

I2C_SCL_LOW             andn    outa, I2C_SCL_Bit        'SCL bit = 0
                        mov     loop, I2C_HALF_BIT_DELAY
                        add     loop, cnt
                        waitcnt loop, #0
I2C_SCL_LOW_ret         ret
'                        
'
'Set SCL to low
                        'make sure SCL = 1 at eeprom
I2C_SCL_HIGH            or      outa, I2C_SCL_Bit        'SCL bit = 1
                        mov     loop, I2C_HALF_BIT_DELAY
                        add     loop, cnt
                        waitcnt loop, #0
I2C_SCL_HIGH_ret        ret
'                        
'
'Set SDA to low, also SDA is output
I2C_SDA_LOW             or      dira, I2C_SDA_Bit        'Set SDA as output, data set to 0 at init
                        mov     loop, I2C_HALF_BIT_DELAY
                        add     loop, cnt
                        waitcnt loop, #0
I2C_SDA_LOW_ret         ret
'                        
'
'Set SDA to high, also SDA is input
I2C_SDA_HIGH            andn    dira, I2C_SDA_Bit        'Set SDA as input, pull up resistor forces = 1 at eeprom
                        mov     loop, I2C_HALF_BIT_DELAY
                        add     loop, cnt
                        waitcnt loop, #0
I2C_SDA_HIGH_ret        ret
'                        
'
'Perform a I2C start bit, SDA goes low while SCL is high
F_CMD_I2C_START         call    #I2C_SCL_LOW
                        call    #I2C_SDA_HIGH
                        'SCL = 0, SDA = output = 1 at eeprom
                        call    #I2C_SCL_HIGH
                        call    #I2C_SDA_LOW
                        'Next set SCL = 0, then wait
                        call    #I2C_SCL_LOW
                        'Finally bring SDA = 1
                        call    #I2C_SDA_HIGH
                        'we always exit with SCL = 0, SDA = 1
F_CMD_I2C_START_ret     ret
'                        
'
'Perform a I2C stop bit, SDA goes low while SCL is high
F_CMD_I2C_STOP          call    #I2C_SCL_LOW
                        call    #I2C_SDA_LOW
                        'SCL = 0, SDA = output = 0 at eeprom
                        call    #I2C_SCL_HIGH
                        call    #I2C_SDA_HIGH
                        'Next set SCL = 0, then wait
                        call    #I2C_SCL_LOW
                        'we always exit with SCL = 0, SDA = 1
F_CMD_I2C_STOP_ret      ret
'                        
'
'18) I2C-RESET ( -- )
'reset the I2C interface
'Software reset I2C eeprom
F_CMD_I2C_RESET         call    #F_CMD_I2C_START
                        'Pulse SCL 9 times with SDA = 1
                        mov     t1, #63
I2C_ResetRpt            call    #I2C_SCL_HIGH
                        call    #I2C_SCL_LOW
                        djnz    t1, #I2C_ResetRpt
                        'Perform another start bit
                        call    #F_CMD_I2C_START
                        'Perform stop bit
                        call    #F_CMD_I2C_STOP
                        'we alway exit with SCL = 0, SDA = 1
F_CMD_I2C_RESET_ret     ret
'                        
'
'Read a byte from the I2C buss.
'On entry T2 = 0 to write ack = 0, T2 <> 0 ack = 1
'On exit t1 = data byte. 
I2C_ReadByte            mov     t1, #0
                        call    #I2C_SCL_LOW
                        call    #I2C_SDA_HIGH   'Set SDA = input                                                        
                        mov     t3, #8          'Need to receive 8 bit plus send ack/nak
I2C_ReadByteRpt         shl     t1, #1          'Shift bits ready for next msb of byte
                        'now set SCL high then low to clock data bit
                        call    #I2C_SCL_HIGH
                        'read data bit
                        mov     t4, ina
                        and     t4, I2C_SDA_Bit wz
              if_nz     or      t1, #1                        
                        'Set SCL = 0
                        call    #I2C_SCL_LOW
                        djnz    t3, #I2C_ReadByteRpt
                        'clocked in our eight bits, 
                        'now clock out our ack bit
                        mov     t2, t2          wz      'see if ack is 1 or 0
              if_z      call    #I2C_SDA_LOW                                                           
                        mov     t2, t2          wz      'see if ack is 1 or 0
              if_nz     call    #I2C_SDA_HIGH
                        'Clock SCL to send ack/nak         
                        call    #I2C_SCL_HIGH
                        call    #I2C_SCL_LOW
                        'we alway exit with SCL = 0, SDA = 1
                        call    #I2C_SDA_HIGH
I2C_ReadByte_ret        ret
'
'
'Write a byte to the I2C buss, MSb first.
'After byte is sent eeprom returns ack/nak on 9th clock.
'On entry T1 = data byte to write.
I2C_WriteByte           shl     t1, #24         'move data byte into MSB of T1
                        mov     t2, #8          'number of bits to write to eeprom
I2C_WriteByteRpt        shl     t1, #1    wc    'Move next bit into carry flag
              if_c      call    #I2C_SDA_HIGH                                                           
              if_nc     call    #I2C_SDA_LOW
                        'now set SCL high then low to clock data bit
                        call    #I2C_SCL_HIGH
                        call    #I2C_SCL_LOW
                        djnz    t2, #I2C_WriteByteRpt
                        'clocked out our eight bits, set SDA as input to read act
                        call    #I2C_SDA_HIGH
                        call    #I2C_SCL_HIGH
                        mov     I2C_Resp, ina
                        and     I2C_Resp, I2C_SDA_Bit
                        'we alway exit with SCL = 0, SDA = 1
                        call    #I2C_SCL_LOW
I2C_WriteByte_ret       ret
'
'
'                        
'FORTH system command support routines
'
' Remove top 16 bit value from forth stack.T1 = 16 bit value on return
                        'make sure we are in a command and stack pointer is valid
F_PopStack              tjz     ForthStackPtr, #F_PopStack_Ret
                        'We are in a command OK to pop value from stack
                        rdword  T1, ForthStackPtr
                        add     ForthStackPtr, #2
F_PopStack_Ret          ret             
'
' Add 16 bit value to top of forth stack.T1 = 16 bit value to write.
                        'make sure we are in a command and stack pointer is valid
F_PushStack             tjz     ForthStackPtr, #F_PushStack_Ret
                        'We are in a command OK to push a value to the stack
                        sub     ForthStackPtr, #2
                        wrword  T1, ForthStackPtr
F_PushStack_Ret         ret             
'
'
'Unsigned 32 bit division with 32 bit result and 32 bit remainder.
'On entry T1 - dividend, T2 = divisor. On exit T3 = T1 / T2 
DoDivU32                tjnz    T2, #DoDivU32_1         'If T2 = 0 we have a divide by zero error
                        'Handle division by zero, return max value and remainder of 0
                        mov     T3, #0
                        sub     T3, #1
                        jmp     #DoDivU32_ret
DoDivU32_1              'Perform a normal division                                                              
                        mov     T3, #0
                        mov     T4, #0
                        mov     T5, #32                                   
                        'Do the shift and sub a max of 32 times.
DoDivU32_2              shl     T1, #1          wc
                        rcl     T4, #1
                        cmpsub  T4, T2         wc      'C flag is set (1) if a subtraction is possible.
                        rcl     T3, #1
                        djnz    T5, #DoDivU32_2
DoDivU32_ret            ret
'


'SPI SD card block read and write code
'
' SD card block read/wrote functions
'
' Transfer a block (512 butes) to/from the SD card, T1 = block index to transfer.
' The source of the opcode at SD_RequestIOCmd must be set to "r" or "w".  
' On exit T1 = 0 if operation ok, <> 0 means an error.
SD_RequestIO            mov     T3, SD_SpiCmdBufAdr
                        add     T3, #4
                        wrlong  T1, T3                  ' Set block index for SD command
                        add     T3, #4
                        wrlong  T2, T3      ' Set the SD block buffer address for SD command 
                        ' Now issue the command
SD_RequestIOCmd         mov     T3, #"r"
                        wrlong  T3, SD_SpiCmdBufAdr
                        ' Wait for command to complete
SD_RequestIOWait        rdlong  SDBlockResp, SD_SpiCmdBufAdr
                        cmp     SDBlockResp, T3          wz
              if_z      jmp     #SD_RequestIOWait
                        ' If SDBlockResp = 0 task completed, else return and report error in T1
SD_RequestIO_ret        ret        
'
'
'Copy a region from global memoru to the eeprom
'On Entry: T1 = forth block index to write, T2 = global memory address
'On Exit: SDBlockResp = 0 if OK or <> 0 if failed
'Local variables for I2C read/write to global memory functions.
SD_WriteBlock           movs    SD_RequestIOCmd, #"w"
                        jmp     #SD_TransBlock                  
'Copy a region from the SD card block file to global memoru
'On Entry: T1 = forth block index to read, T2 = global memory address
'On Exit: SDBlockResp = 0 if OK or <> 0 if failed
SD_ReadBlock            movs    SD_RequestIOCmd, #"r"
                        'convert forth block Idx to SD card Idx
SD_TransBlock           shl     T1, #1
                        'add on block file start offset
                        add     T1, SD_FileStartIdx
                        'transfer first half of forth block
                        call    #SD_RequestIO
                        tjnz    SDBlockResp, #SD_ReadBlock_ret
                        'write second half of forth block
                        add     T1, #1          'want to transfer next 512 byte block
                        add     T2, Const_512 
                        call    #SD_RequestIO
SD_WriteBlock_ret
SD_ReadBlock_ret        ret

'
'read or write a 1k block to or from the block device using supplied block number and memory address.
'A block is 1024 bytes and will span two SD card blocks (512 bytes).
'
'Read/Write block ( u adr -- f ), u = mass storage block number, adr = buffer memory address,
'f = 0 I/O failed, f = -1 I/O performed.
IO_OperationCode        long    0                'Set to 0 for read 1 for write
'
F_DO_BLK_IO             call    #F_PopStack     'T1 = global memory address
                        mov     T2, T1          'T2 = global memory address
                        call    #F_PopStack     'T1 = block number
                        'see if block number in range
                        cmp     T1, SD_FileBlockCount     wc
        if_nc           mov     T1, #0
        if_nc           jmp     #F_DO_BLK_IO_DONE
                        'Request is valid, try to read/write eeprom device
                        'T1 = Forth Block Number
                        'T2 = global memory address
                        'we have a valid request and all params are set, read/write eeprom
                        cmp     IO_OperationCode, #0    wz
        if_z            call    #SD_ReadBlock
                        cmp     IO_OperationCode, #0    wz
        if_nz           call    #SD_WriteBlock
                        'I/O operation completed report results
                        cmp     SDBlockResp, #0          wz
                        mov     T1, DataWordMask        'T1 = -1 16 bits
        if_nz           mov     T1, #0
                        'push result to stack and exit
F_DO_BLK_IO_DONE        call    #F_PushStack
F_DO_BLK_IO_ret         ret
'
'Run the cog T3 = #8 COG auto selected, 0..7 run on selected cog
COG_RUN                 call    #F_PopStack     'T1 = PAR (14 bit) value
                        mov     T2, T1          'T2 = PAR value
                        call    #F_PopStack     'T1 = COG (14 bit ) code address
                        shr     T2, #2          'Convert T2 to 14 bit number
                        shr     T1, #2          'Convert T1 to 14 bit number
                        'build destination word
                        shl     T2, #18
                        shl     T1, #4
                        or      T1, T2
                        or      T1, T3
                        'params are ready start new cog
                        coginit T1              wc, wr
              if_c      mov     T1, DataWordMask
              if_nc     and     T1, #7
COG_RUN_ret             ret
'
'
'FORTH system commands
'
'1) EMIT ( c -- ) Write char to console, c = 8 bit char to send to console.
F_CMD_EMIT              call    #F_PopStack             'T1 = character value to send
                        rdword  T2, ComTxHeadAdr        'T2 = tx fifo head
F_CMD_EMIT_WAIT         rdword  T3, ComTxTailAdr        'T3 = tx fifo Tail
                        'see if fifo is full (Head + 1 == Tail)
                        mov     T4, T2                  'T4 = tx fifo head
                        add     T4, #1
                        cmp     T4, ComTxBufferSize     wz
        if_z            mov     T4, #0                  'Head wrapped back to fifo start
                        cmp     T4, T3                  wz
        if_z            jmp     #F_CMD_EMIT_WAIT       'Head + 1 = Tail, no room in fifo, try again
                        'There is room write character
                        mov     T5,  ComTxBufferAdr
                        add     T5, T2                  'Calculate tx fifo write address
                        wrbyte  T1, T5
                        wrword  T4, ComTxHeadAdr        'update tx fifo head with head new value
F_CMD_EMIT_ret          ret
'
'
'2) KEY ( -- c ) Read char from console, c = 8 bit char read from console.
F_CMD_KEY               rdword  T3, ComRxTailAdr        'T3 = tx fifo Tail
F_CMD_KEY_WAIT          rdword  T2, ComRxHeadAdr        'T2 = tx fifo Head
                        cmp     T2, T3                  wz
        if_z            jmp     #F_CMD_KEY_WAIT
                        'read from fifo
                        mov     T4, ComRxBufferAdr
                        add     T4, T3
                        rdbyte  T1, T4
                        call    #F_PushStack
                        'Update tail to show we read fifo
                        add     T3, #1
                        cmp     T3, ComRxBufferSize      wz
        if_z            mov     T3, #0
                        wrword  T3, ComRxTailAdr        'update tx fifo Tail with T3
F_CMD_KEY_ret           ret
'
'
'3) KEY? ( -- f )
'Test if console has data to read, f = 0 no char ready
'from console, f= -1 char available from console.
F_CMD_CAN_KEY           rdword  T3, ComRxTailAdr        'T3 = tx fifo Tail
                        rdword  T2, ComRxHeadAdr        'T2 = tx fifo Head
                        cmp     T2, T3                  wz
        if_z            mov     T1, #0
        if_nz           mov     T1, DataWordMask
                        call    #F_PushStack
F_CMD_CAN_KEY_ret       ret
'
'
'4) EMIT? ( -- f ) f = -1 can send, f= 0 unable to send.
F_CMD_CAN_EMIT          rdword  T2, ComTxHeadAdr        'T2 = tx fifo head
                        rdword  T3, ComTxTailAdr        'T3 = tx fifo Tail
                        add     T2, #1
                        cmp     T2, ComTxBufferSize     wc
        if_nc           mov     T2, #0                  'Head wrapped back to fifo start
                        cmp     T2, T3                  wz
        if_z            mov     T1, #0
        if_nz           mov     T1, DataWordMask
                        call    #F_PushStack
F_CMD_CAN_EMIT_ret      ret
'
'
'5) EMPTYKEYS ( -- ) Request the console input buffer be cleared.
F_CMD_EMPTYKEYS         mov     T1, #0
                        wrword  T1, ComRxHeadAdr        'rx fifo head = 0
                        wrword  T1, ComRxTailAdr        'rx fifo Tail = 0
                        call    #F_CompleteCommand
F_CMD_EMPTYKEYS_ret     ret
'
'
'6) READ_BLK ( block mem-adr -- f )
'Read block from eeprom device to memory buffer.
'f = 0 operation failed, f = -1 (16 bit) operation completed.
F_CMD_READ_BLK          mov     IO_OperationCode, #0
                        call    #F_DO_BLK_IO
F_CMD_READ_BLK_ret      ret
'
'
'7) WRITE_BLK ( block mem-adr -- f )
'Write forth block in memory to the eeprom device.
'f = 0 operation failed, f = -1 (16 bit) operation completed.
F_CMD_WRITE_BLK         mov     IO_OperationCode, DataWordMask
                        call    #F_DO_BLK_IO
F_CMD_WRITE_BLK_ret     ret
'
'
'8) BLK_COUNT ( -- u )
'Returns number blocks avaulable, u = mass storage block count.
F_CMD_BLK_COUNT         mov     T1, SD_FileBlockCount
                        call    #F_PushStack
F_CMD_BLK_COUNT_ret     ret
'
'
'9) COD-INIT ( Code-adr u1 u2 -- )
'Start the COG index number u2, if the COG is already running stop the COG and restart
'the COG with the new code. q-adr1 points to a long aligned address containg
'the cog code to execute. u1 (upper 14 bits) is passes via the COG PAR register.
F_CMD_COG_INIT          call    #F_PopStack     'T1 = PAR (14 bit) value
                        mov     T3, T1          'T2 = PAR value
                        and     T3, #7
                        call    #COG_RUN
F_CMD_COG_INIT_ret      ret
'
'                      
'10) COG-NEW ( Code-adr u1 -- u2 )
'Start a new COG Code-adr points to a long aligned address containg
'the cog code to execute. u1 (upper 14 bits) is passes via the COG PAR register.
'The function returns the COG number started in u2 or u2 = -1 if no cog available.
F_CMD_COG_NEW           mov     T3, #8          'T3 = auto COG select
                        call    #COG_RUN
                        'save result to forth data stack
                        call    #F_PushStack
F_CMD_COG_NEW_ret       ret
'
'
'11) COG-STOP ( u -- )
'Stop the COG index number u if it is running.
F_CMP_COG_STOP          call    #F_PopStack     'T1 = cog index to stop
                        cogstop T1
F_CMP_COG_STOP_ret      ret
'
'
'12) LOCK_NEW ( -- u )
'Allocate a lock u is the lock index or -1 if no locks avaiable
'Execute next system command.
F_CMD_LOCK_NEW          locknew T1              wc
        if_c            mov     T1, DataWordMask
                        call    #F_PushStack
F_CMD_LOCK_NEW_ret      ret
'
'
'13) LOCK_RET ( u -- )
'Release a lock, u is the lock index to release/return
F_CMD_LOCK_RET          call    #F_PopStack     'T1 = lock index to return
                        lockret T1              wc
F_CMD_LOCK_RET_ret      ret
'
'14) C FILL   ( c-addr u char -- ) fill memory with char
'u bytes are filled starting at address c-adr
F_CMD_FILL              call    #F_PopStack     'T1 = char to fill with
                        mov     T2, T1          'T2 = char to fill with
                        call    #F_PopStack     'T1 = fill count
                        mov     T3, T1          'T3 = fill count
                        call    #F_PopStack     'T1 = fill address
                        'Fill memory bytes (count, T3)
                        tjz     T3, #F_CMD_FILL_ret
                        'Count is positive fill memory                        
F_CMD_FILL_NXT          wrbyte  T2, T1
                        add     T1, #1
                        djnz    T3, #F_CMD_FILL_NXT
F_CMD_FILL_ret          ret     
'
'15) CMOVE ( c-addr1 c-addr2 u -- )
'If u is greater than zero, copy u consecutive characters from the data space starting at c-addr1
'to that starting at c-addr2, proceeding character-by-character from lower addresses to higher
'addresses.
F_CMD_CMOVE             call    #F_PopStack     'T1 = char to copy
                        mov     T3, T1          'T3 = chars to copy
                        call    #F_PopStack     'T1 = dest address
                        mov     T2, T1          'T2 = dest address
                        call    #F_PopStack     'T1 = source address
                        'Fill memory bytes (count, T3)
                        tjz     T3, #F_CMD_CMOVE_ret
F_CMD_CMOVE_NEXT        rdbyte  T4, T1
                        add     T1, #1
                        wrbyte  T4, T2     
                        add     T2, #1
                        djnz    T3, #F_CMD_CMOVE_NEXT
F_CMD_CMOVE_ret         ret
'
'16) CMOVE> ( c-addr1 c-addr2 u -- )
'If u is greater than zero, copy u consecutive characters from the data space starting at c-addr1
'to that starting at c-addr2, proceeding character-by-character from higher addresses to lower
'addresses.
F_CMD_CMOVE_REV         call    #F_PopStack     'T1 = char to copy
                        mov     T3, T1          'T3 = chars to copy
                        call    #F_PopStack     'T1 = dest address
                        mov     T2, T1          'T2 = dest address
                        call    #F_PopStack     'T1 = source address
                        'copy from end of region to start
                        add     T1, T3
                        sub     T1, #1
                        add     T2, T3
                        sub     T2, #1
                        'Do copy operation
F_CMD_CMOVE_REV1        rdbyte  T4, T1
                        sub     T1, #1
                        wrbyte  T4, T2     
                        sub     T2, #1
                        djnz    T3, #F_CMD_CMOVE_REV1
F_CMD_CMOVE_REV_ret     ret                        
'
'17) BAUD ( d -- ) set serial port baud rate
F_CMD_BAUD              call    #F_PopStack     'T1 = high word of double
                        mov     T2, T1          'T2 = high word of double 
                        shl     T2, #16         'move word into bits 16..31
                        call    #F_PopStack     'T1 = low word of double
                        or      T2, T1          'T2 = 32bit double value of baud rate
                        rdlong  T1, #0          'read the system frequency from global memory
                        call    #DoDivU32
                        wrlong  T3, ComBitTicksAdr
F_CMD_BAUD_ret          ret
'
'19) I2C-READ ( uDataAdr uByteCount -- )
'read specified number of bytes from I2C interface and save to global memory
F_CMD_I2C_READ          call    #F_PopStack     'CopyCnt = number bytes to copy
                        mov     CopyCnt, T1
                        call    #F_PopStack     'MemAdr = global memory address
                        mov     I2C_MemAdr, T1
                        'loop and copy from I2C to memory
F_CMD_I2C_READ_NEXT     mov     t2, #0          'Send ACK = 0
                        call    #I2C_ReadByte
                        wrbyte  T1, I2C_MemAdr
                        add     I2C_MemAdr, #1
                        djnz    CopyCnt, #F_CMD_I2C_READ_NEXT
F_CMD_I2C_READ_ret      ret
'
'20) I2C-WRITE ( uDataAdr uByteCount -- fAcked )
'read specified number og bytes from I2C interface
F_CMD_I2C_WRITE         mov     I2C_Resp, #0  'Indicate no error
                        call    #F_PopStack     'CopyCnt = number bytes to copy
                        mov     CopyCnt, T1
                        call    #F_PopStack     'MemAdr = global memory address
                        mov     I2C_MemAdr, T1
                        'loop and copy from I2C to memory
F_CMD_I2C_WRITE_NEXT    mov     t2, #0          'Send ACK = 0
                        rdbyte  T1, I2C_MemAdr
                        add     I2C_MemAdr, #1
                        call    #I2C_WriteByte
                        tjnz    I2C_Resp, #F_CMD_I2C_WRITE_ERROR
                        djnz    CopyCnt, #F_CMD_I2C_WRITE_NEXT
                        'All done now report positive result and return
                        mov     T1, DataWordMask
                        jmp     #F_CMD_I2C_WRITE_DONE
                        '( did not get responce report error )
F_CMD_I2C_WRITE_ERROR   mov     T1, #0
F_CMD_I2C_WRITE_DONE    call    #F_PushStack
F_CMD_I2C_WRITE_ret     ret
'

'23) I2C-FAST ( -- )
'Set the I2C interface to run at @ 1 Mbs.

F_CMD_I2C_SPEED         call    #F_PopStack             'get high word of requested bit rate
                        mov     T2, T1                  'save in to T2
                        shl     T2, #16                 'mov word in to high word of 32 bit bit rate
                        call    #F_PopStack             'get low word of requested bit rate
                        or      T2, T1                  'mov word in to high word of 32 bit bit rate
                        max     T2, OneMillion          'limit to our highest bit rate (1 Mbs)
                        call    #SetI2cBitDelay     
F_CMD_I2C_SPEED_ret     ret

'25) REBOOT ( -- )
'Hardware reset of CPU
F_CMD_REBOOT            mov     T1, #$80
                        clkset  T1
F_CMD_REBOOT_WAIT       jmp     #F_CMD_REBOOT_WAIT
'
'
{ ************* FUNCTION FAST FIND CODE ********************** }

'Load the search string into local buffer
'Search string is prefixed by length byte
'On Entry
'  SearchStrAdr = address of string to search for in dictionary.
'
LoadStr                 mov     T1, SearchStrAdr
                        rdbyte  SearchByteCnt, T1
                        add     T1, #1
                        'Copy the search string to the buffer.
                        'if search string is zero length we are done.
                        tjz     SearchByteCnt, #LoadStr_ret
                        'There is at least 1 character to copy.
                        'First save the length byte
                        mov     SearchBuf, SearchByteCnt
                        movd    LoadStrCpy, #SearchBuf + 1    
                        mov     T2, SearchByteCnt
                        'Limit copy to c_MaxSearchLen bytes
                        max     T2, #c_MaxSearchLen
                        'Copy the bytes to the buffer
LoadStrCpy              rdbyte  (SearchBuf), T1
                        add     T1, #1
                        add     LoadStrCpy, D0
                        djnz    T2, #LoadStrCpy
LoadStr_ret             ret
'
'
'Compare a length prefixed string to the string in search buffer for an exact match
'On Entry:
'  T1 = address of string to compare with search string
'On Exit:
'  Z = 0 if no match, Z = 1 string matches

CmpStr                  rdbyte  T2, T1                  'read string length
                        cmp     T2, SearchByteCnt    wz 'see if length same as search string
              if_nz     jmp     #CmpStr_ret
                        'String is same length, see if it matches
                        'Limit compare to c_MaxSearchLen bytes
                        max     T2, #c_MaxSearchLen
                        'Start compare at first character after length                                                
                        movd    CmpStrCompare, #SearchBuf + 1    
                        'Loop and compare string bytes
CmpStrNext              add     T1, #1          'point to next char
                        rdbyte  T3, T1
CmpStrCompare           cmp     (SearchBuf), T3     wz
                        add     CmpStrCompare, D0
              if_z      djnz    T2, #CmpStrNext
CmpStr_ret              ret
'
'
'Read the next forth word NFA (name field address)
'On Entry:
'  LatestAdr = address of current forth word name field
'On Exit:
'  LatestAdr = address of next forth word name field or zero if at last word
GetNextNFA              tjz     LatestAdr, #GetNextNFA_ret
                        'Convert NFA to LFA
                        sub     LatestAdr, #4
                        'read word at link field address to get prior NFA
                        rdword  LatestAdr, LatestAdr
GetNextNFA_ret          ret
'
'
'Perform forth FastFind function
' Foth stack has search string address, and value of forth LATEST variable
'  FastFind ( c-addr uLatest -- c-addr 0 ) or ( c-addr uLatest -- xt 1 | -1 )
F_CMD_FIND              call    #F_PopStack     'LatestAdr = address of last forth word name field
                        mov     LatestAdr, T1
                        call    #F_PopStack     'SearchStrAdr = search string memory address
                        mov     SearchStrAdr, T1
                        'Copy search string to buffer so we only read once
                        call    #LoadStr
                        'If search string is blank no match is possible
                        tjz     SearchByteCnt, #F_CMD_FIND_Failed
                        'see if we can match any dictionary name
                        tjz     LatestAdr, #F_CMD_FIND_Failed
F_CMD_FIND_Next         mov     T1, LatestAdr
                        call    #CmpStr
              if_z      jmp     #F_CMD_FIND_Ok
                        'no match try next forth word
                        call    #GetNextNFA
                        tjnz    LatestAdr, #F_CMD_FIND_Next
                        'failed to match, return c-addr and 0 on stack
F_CMD_FIND_Failed       mov     T1, SearchStrAdr
                        call    #F_PushStack
                        mov     T1, #0
                        call    #F_PushStack
                        jmp     #F_CMD_FIND_ret
                        'We have a match, LatestAdr = NFA of matching forth word
                        'Convert the NFA to the CFA (code field address) of forth word
                        'Return a 1 if word is immediate, return a -1 if normal    
F_CMD_FIND_Ok           mov     T1, LatestAdr   'Calculate forth word FFA (flag field address)
                        sub     T1, #2
                        'T1 = matching word FFA
                        rdword  T1, T1          'T1 = matching forth word flags
                        test    T1, #c_ImmediateFlag    wc
                        'Immediate flag saved in carry flag
                        'C = 0 word is normal, C = 1 word is immediate
                        'Convert matcing forth word NFA to CFA (code field address)
                        rdbyte  T1, LatestAdr   'Read forth word name length
                        add     T1, #1          'Add one to length for string length byte
                        'if the string length is odd word align to next word boundry
                        test    T1, #1          wz
              if_nz     add     T1, #1                        
                        'Add string size to NFA to get CFA
                        add     T1, LatestAdr
                        call    #F_PushStack
                        mov     T1, DataWordMask        'set to -1 (16 bit)
              if_c      mov     T1, #1                                                
                        call    #F_PushStack
F_CMD_FIND_ret          ret



{ ************* FUNCTION COG-EX MOVED TO MAIN FORTH COG **********************
 
'26) COG-EX ( n0..nx uCmdBufAdr uCmdIdx -- ? )
'Request a command be executed by an external COG
'that uses the same mehtod to pass apameters as the code in this COG .
'uCmdBufAdr is the address of a long 32 bit variable used to execute
'commands on another COG. uCmdIdx is the command index to be executed by the COG.
'n0..nx represend any number of parameters passed on the stack to the COG.
F_CMD_COGEX             call    #F_PopStack     'T1 = command index (16 bit)
                        mov     T3, T1          'T3 = command index
                        call    #F_PopStack     'T1 = command buffer address (long aligned)
                        mov     T2, T1          'T2 = command buffer address (16 bits)
                        'Build command for othaer COG.
                        mov     T1, ForthStackPtr       'Current forth stack pointer address
                        shl     T1, #16                 'move to high word of long
                        or      T1, T3                  'add in command index (low 16 bits)
                        'Set command buffer to execute command on second COG
                        wrlong  T1, T2
                        'Command is set, now wait for it to be completed (set = 0)
F_CMD_COGEX_WAIT        rdlong  ForthStackPtr, T2
                        test    ForthStackPtr, DataWordMask     wz
        if_nz           jmp     #F_CMD_COGEX_WAIT
                        'Command completed, update the data stack pointer                        
                        shr     ForthStackPtr, #16
F_CMD_COGEX_ret         ret
'
}
'                        
'Wait for a system command to be requested by forth engine
F_WaitForCmd            rdlong  ForthCmd, par
                        test    ForthCmd, DataWordMask  wz
        if_z            jmp     #F_WaitForCmd
                        'Seperate stack address from command number
                        mov     ForthStackPtr, ForthCmd
                        shr     ForthStackPtr, #16
                        and     ForthCmd, DataWordMask
                        'see if we have a valid command
                        cmp     ForthCmd, #c_ForthCmdCount                      wc
        if_c            jmp     #F_WaitForCmd_ret
                        ' we have an invalid command
                        call    #F_CompleteCommand                        
                        jmp     #F_WaitForCmd
                        ' we have a valid command, get the stack pointer address
F_WaitForCmd_Ret        ret

'
'Update stack pointer then exit
F_SaveStackPtrExit      shl     ForthStackPtr, #16
                        wrlong  ForthStackPtr, PAR
                        mov     ForthStackPtr, #0
F_SaveStackPtrExit_ret  ret
'
'Indicate system command requested by forth engine is completed or invalid
F_CompleteCommand       mov     ForthCmd, #0                          
                        wrword  ForthCmd, par
                        mov     ForthStackPtr, #0
F_CompleteCommand_ret   ret
'
'
F_Do_Cmds               call    #F_WaitForCmd
                        'we have a cmd index 1..25
                        mov     T1, #F_JUMP_TABLE
                        add     T1, ForthCmd
                        movs    :F_Set_Cmd_Jump, T1
                        nop
:F_Set_Cmd_Jump         mov     :F_Do_Cmd_Jump, (T1)
                        nop
:F_Do_Cmd_Jump          nop
                        'The called function will return here
F_Do_Cmds_Return        call    #F_SaveStackPtrExit
                        jmp     #F_Do_Cmds
'
'   
'Table of Jump opcodes for each of the base I/O functions
F_JUMP_TABLE            nop                             '(0)
                        call    #F_CMD_EMIT             '(1)
                        call    #F_CMD_KEY              '(2)
                        call    #F_CMD_CAN_KEY          '(3)
                        call    #F_CMD_CAN_EMIT         '(4)
                        call    #F_CMD_EMPTYKEYS        '(5)
                        call    #F_CMD_READ_BLK         '(6)
                        call    #F_CMD_WRITE_BLK        '(7)
                        call    #F_CMD_BLK_COUNT        '(8)
                        call    #F_CMD_COG_NEW          '(9)
                        call    #F_CMD_COG_INIT         '(10)
                        call    #F_CMP_COG_STOP         '(11)
                        call    #F_CMD_LOCK_NEW         '(12)
                        call    #F_CMD_LOCK_RET         '(13)
                        call    #F_CMD_FILL             '(14)
                        call    #F_CMD_CMOVE            '(15)
                        call    #F_CMD_CMOVE_REV        '(16)
                        call    #F_CMD_BAUD             '(17)
                        call    #F_CMD_I2C_RESET        '(18)
                        call    #F_CMD_I2C_READ         '(19)
                        call    #F_CMD_I2C_WRITE        '(20)
                        call    #F_CMD_I2C_START        '(21)
                        call    #F_CMD_I2C_STOP         '(22)
                        call    #F_CMD_I2C_SPEED        '(23)
                        jmp     #F_CMD_REBOOT           '(24)
                        call    #F_CMD_FIND             '(25)
'
'                        
'Pre defined constants
OneMillion              long    1_000_000 
I2C_SCL_Bit             long    0
I2C_SDA_Bit             long    0
DataWordMask            long    $FFFF
D0
Const_512               long    512
'
I2C_HALF_BIT_DELAY      long    1
'
'Forth command variables ForthCmdAdr (16 bits) followed directly by StackPtrAdr (16 bits)
ForthCmd                long    0
ForthStackPtr           long    0
'
ComBitTicksAdr          long    c_bit_ticks_adr
'Com port rx fifo addresses in global memory
ComRxHeadAdr            long    0
ComRxTailAdr            long    0
ComRxBufferSize         long    0
ComRxBufferAdr          long    0
'Com port tx fifo addresses in global memory
ComTxHeadAdr            long    0
ComTxTailAdr            long    0
ComTxBufferSize         long    0
ComTxBufferAdr          long    0
'SPI command buffer address
SD_SpiCmdBufAdr         long    0
'Fist 512 byte SD block index of forth block file on SD card
SD_FileStartIdx         long    0
'Number of forth blocks available in forth block file
SD_FileBlockCount       long    0
'Fast Find Variables
SearchStrAdr            long    0
SearchByteCnt           long    0

SearchBuf               res     c_MaxSearchLen + 1
LatestAdr               res     1

' ******** END OF COG CODE ********
'